package ru.crp.lvms.cluster;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestClientException;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.cmls.calculations.solvers.xml.config.SolverType;
import ru.crp.lvms.cluster.jobs.CheckSlurmClustersJob;
import ru.crp.lvms.cluster.jobs.CheckSpaceClustersJob;
import ru.crp.lvms.cluster.model.*;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.tasks.slurm.exceptions.SlurmConnectionException;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.*;

@Service
public class ClustersManager {
    public static final int BENCH_LOCAL_CLUSTER_ID = 1;
    public static final String CLUSTER_INFO_QUEUE = "/queue/clusters";

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private NodesStatisticsDao nodesStatisticsDao;

    @Autowired
    private ClusterDao clusterDao;

    @Autowired
    private CheckSpaceClustersJob checkSpaceClustersJob;

    @Autowired
    private CheckSlurmClustersJob checkSlurmClustersJob;

    @Autowired
    private SlurmClusterConfigDao slurmClusterConfigDao;


    public List<ClusterInfo> getFreshClustersInfo() {
        Map<Integer, ClusterEntity> allClusters = getAllClusters().stream()
                .collect(toMap(ClusterEntity::getId, Function.identity()));

        Map<ClusterEntity, List<NodeForClusterInfoEntity>> nodesByCluster = nodesStatisticsDao.getAllNodes().stream()
                .collect(groupingBy(node -> allClusters.get(node.getClusterId())));
        getAllClusters().forEach(cluster -> nodesByCluster.putIfAbsent(cluster, emptyList()));

        List<ClusterInfo> clusterInfos = new ArrayList<>(nodesByCluster.size());

        List<String> allSolverNames = solverConfigManager.getAllConfigs().stream()
                .filter(config -> config.getSolverType() == SolverType.CAE)
                .map(SolverConfig::getName)
                .collect(toList());

        nodesByCluster.forEach((cluster, nodes) -> {
            ClusterInfo clusterInfo = new ClusterInfo(cluster.getId(), cluster.getName(), cluster.getType(), allSolverNames);
            nodes.forEach(clusterInfo::addNode);
            clusterInfos.add(clusterInfo);
        });
        return clusterInfos;
    }


    @Nullable
    public NodeStatus getNodeStatus(NodeEntity node) {
        ClusterEntity cluster = clusterDao.findOne(node.getClusterId());
        ClusterType clusterType = ClusterType.valueOf(cluster.getTypeId());
        try {
            if (clusterType == ClusterType.Space) {
                return checkSpaceClustersJob.getMostRecentNodeStatus(node);
            }
            if (clusterType == ClusterType.Slurm) {
                return checkSlurmClustersJob.getMostRecentNodeStatus(node);
            }
        } catch (RestClientException | SlurmConnectionException ex) {
            return null;
        }
        throw new IllegalArgumentException(String.format("Unexpected cluster type - %s", clusterType));
    }

    public boolean isSlurmTask(TaskEntity task) {
        ClusterEntity cluster = getClusterForTask(task);
        ClusterType clusterType = ClusterType.valueOf(cluster.getTypeId());
        return clusterType == ClusterType.Slurm;
    }

    @Transactional(readOnly = true)
    public SlurmClusterConfigEntity getSlurmClusterConfig(int clusterId) {
        return slurmClusterConfigDao.getByClusterId(clusterId);
    }

    public List<ClusterEntity> getAllClusters() {
        return clusterDao.findAll();
    }

    public ClusterEntity getClusterForTask(TaskEntity task) {
        return clusterDao.findOne(task.getClusterId());
    }

    public List<ClusterEntity> getClustersByType(ClusterType clusterType) {
        return clusterDao.getClustersByType(clusterType);
    }

    public List<ClusterEntity> suggestClusters(SuggestRequest suggestRequest, UserEntity currentUser) {
        return clusterDao.suggest(suggestRequest, currentUser.getId());
    }

    public ClusterType getClusterType(int clusterId) {
        return clusterDao.getClusterType(clusterId);
    }
}
