package ru.crp.lvms.cluster.jobs;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.calculations.solvers.dto.SolverPresenceRequest;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.cluster.model.SlurmClusterConfigEntity;
import ru.crp.lvms.events.NodeStatusUpdatedEvent;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.tasks.slurm.SlurmConnectionManager;
import ru.crp.lvms.tasks.slurm.SlurmNodeInfo;
import ru.crp.lvms.tasks.slurm.SlurmNodeInfoMapper;
import ru.crp.lvms.tasks.slurm.SlurmTaskInfoDao;
import ru.crp.lvms.tasks.slurm.exceptions.SlurmConnectionException;
import ru.crp.lvms.tasks.slurm.requests.PartitionNodesRequest;

import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static java.util.stream.Collectors.toSet;

@Component
public class CheckSlurmClustersJob extends BaseCheckClusterJob {
    private static final Logger LOGGER = LoggerFactory.getLogger(CheckSlurmClustersJob.class);

    @Autowired
    private SlurmConnectionManager slurmConnectionManager;

    @Autowired
    private SlurmNodeInfoMapper slurmNodeInfoMapper;

    @Autowired
    private SlurmTaskInfoDao slurmTaskInfoDao;

    @Override
    public ClusterType getClusterType() {
        return ClusterType.Slurm;
    }

    @Scheduled(fixedDelay = 60000)
    @Override
    public void checkClusters() {
        super.checkClusters();
    }


    public void checkCluster(ClusterEntity clusterToCheck, List<SolverPresenceRequest> solversToCheck) {
        int clusterId = clusterToCheck.getId();
        SlurmClusterConfigEntity clusterConfig = clustersManager.getSlurmClusterConfig(clusterId);
        try {
            List<SlurmNodeInfo> nodeNamesList = slurmConnectionManager.execCommand(clusterConfig, new PartitionNodesRequest(clusterConfig.getPartitionName()));
            List<String> busySlurmNodeNames = slurmTaskInfoDao.getBusySlurmNodeNames(clusterId);
            for (SlurmNodeInfo nodeInfo : nodeNamesList) {
                refreshNode(clusterConfig, nodeInfo, solversToCheck, busySlurmNodeNames);
            }
        } catch (SlurmConnectionException ex) {
            // intentionally skip ex
            LOGGER.warn("Slurm connection is not alive", ex);
            slurmConnectionManager.setSlurmConnectionAlive(clusterId, false);
        } catch (Exception ex) {
            LOGGER.error("Unexpected exception while updating slurm nodes status", ex);
            slurmConnectionManager.setSlurmConnectionAlive(clusterId, false);
        }
    }


    @Override
    protected NodeStatus doGetMostRecentNodeStatus(NodeEntity nodeToCheck, List<SolverPresenceRequest> solversToCheck) {
        SlurmClusterConfigEntity config = clustersManager.getSlurmClusterConfig(nodeToCheck.getClusterId());
        List<SlurmNodeInfo> nodeNamesList = slurmConnectionManager.execCommand(config, new PartitionNodesRequest(config.getPartitionName()));
        for (SlurmNodeInfo nodeInfo : nodeNamesList) {
            if (nodeInfo.getNodeName().equals(nodeToCheck.getName())) {
                return slurmNodeInfoMapper.mapToNodeStatus(nodeInfo, solversToCheck);
            }
        }
        return null;
    }


    private NodeEntity refreshNode(SlurmClusterConfigEntity config, SlurmNodeInfo slurmNodeInfo,
                                   List<SolverPresenceRequest> solversToCheck, List<String> busySlurmNodeNames) throws BadRequestException {
        NodeEntity node = slurmNodeInfoMapper.mapToNode(config, slurmNodeInfo, busySlurmNodeNames);
        NodeEntity oldNodeState = null;
        if (node.getId() == 0) {
            node = nodeManager.createAndRead(node, null);
        } else {
            oldNodeState = nodeManager.findOne(node.getId(), null);
            nodeManager.update(node, null);
        }
        Set<String> solverNames = solversToCheck.stream().map(SolverPresenceRequest::getSolverName).collect(toSet());
        Set<String> nodeSolvers = slurmNodeInfoMapper.extractNodeSolvers(slurmNodeInfo, solverNames);
        Set<String> previousNodeSolvers = node.getAvailableSolvers().stream().collect(Collectors.toSet());
        if (!nodeSolvers.equals(previousNodeSolvers) && !previousNodeSolvers.isEmpty()) {
            nodeManager.updateNodeSolvers(node, nodeSolvers);
        }
        if (!NodeEntity.equalsIgnoringModificationDate(oldNodeState, node)) {
            applicationEventPublisher.publishEvent(new NodeStatusUpdatedEvent(oldNodeState, node));
        }
        return node;
    }
}
