package ru.crp.lvms.cluster.jobs;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.ResourceAccessException;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.calculations.solvers.dto.SolverPresenceRequest;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.events.NodeStatusUpdatedEvent;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.solvers.SolverConnectionManager;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.utils.Utils;

import java.util.List;

@Component
public class CheckSpaceClustersJob extends BaseCheckClusterJob {
    private static final Logger LOGGER = LoggerFactory.getLogger(CheckSpaceClustersJob.class);

    @Autowired
    private SolverConnectionManager solverConnectionManager;


    @Override
    public ClusterType getClusterType() {
        return ClusterType.Space;
    }


    @Scheduled(fixedDelay = 60000)
    @Override
    public void checkClusters() {
        super.checkClusters();
    }


    @Override
    public void checkCluster(ClusterEntity clusterToCheck, List<SolverPresenceRequest> solversToCheck) throws BadRequestException {
        for (NodeEntity node : nodeManager.getAllForCluster(clusterToCheck.getId())) {
            NodeEntity oldNodeState = null;
            try {
                oldNodeState = (NodeEntity) node.clone();
                updateNodeStatus(node, solversToCheck);
            } catch (ResourceAccessException | HttpClientErrorException ex) {
                node.setIsAlive(false);
                nodeManager.update(node, null);
            } catch (Exception e) {
                LOGGER.error("Unexpected exception while updating node status {}", node.getId(), e);
                node.setIsAlive(false);
                nodeManager.update(node, null);
            } finally {
                publishEventIfNeed(oldNodeState, node);
            }
        }
    }


    public void updateNodeStatus(NodeEntity nodeToUpdate, List<SolverPresenceRequest> solversToCheck) throws BadRequestException {
        NodeStatus nodeStatus = solverConnectionManager.getNodeStatus(nodeToUpdate.getUrl(), solversToCheck);
        nodeToUpdate.setNumOfCores(nodeStatus.getAvailableProcessors());
        nodeToUpdate.setMemory(Utils.memoryInMegabytes(nodeStatus.getTotalPhysicalMemorySize()));
        nodeToUpdate.setIsAlive(true);
        nodeManager.update(nodeToUpdate, null);
        nodeManager.updateNodeSolvers(nodeToUpdate, nodeStatus.getSolvers());
    }


    @Override
    protected NodeStatus doGetMostRecentNodeStatus(NodeEntity nodeToCheck, List<SolverPresenceRequest> solversToCheck) {
        return solverConnectionManager.getNodeStatus(nodeToCheck.getUrl(), solversToCheck);
    }

    private void publishEventIfNeed(NodeEntity oldNodeState, NodeEntity newNodeState) {
        if (oldNodeState == null || newNodeState == null) {
            return;
        }
        if (!NodeEntity.equalsIgnoringModificationDate(oldNodeState, newNodeState)) {
            try {
                applicationEventPublisher.publishEvent(new NodeStatusUpdatedEvent(oldNodeState, (NodeEntity) newNodeState.clone()));
            } catch (CloneNotSupportedException ex) {
                LOGGER.error("Can't clone NodeEntity object", ex);
            }
        }
    }
}
