package ru.crp.lvms.managers;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.stereotype.Service;
import ru.crp.lvms.exceptions.EntityUpdateException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.ParameterValuesDao;
import ru.crp.lvms.model.dao.ParametersDao;
import ru.crp.lvms.model.dao.TreeDao;
import ru.crp.lvms.model.entities.*;
import ru.crp.lvms.utils.ValidationHelper;
import ru.crp.lvms.web.dto.DefaultParameterValue;

import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.List;
import java.util.Map;

import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;
import static ru.crp.lvms.model.AccessRights.*;
import static ru.crp.lvms.utils.MessageUtils.*;

@Service
public class ParametersManager extends BaseManagerSuggest<ParametersEntityTable> {
    public static final BigDecimal DEFAULT_TEMPERATURE = BigDecimal.valueOf(20d);

    @Autowired
    private ParametersDao parametersDao;

    @Autowired
    private TreeDao treeDao;

    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private ParameterValuesDao parameterValuesDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private ValidationHelper validationHelper;


    @Override
    protected BaseSpaceDao<ParametersEntityTable> aliveDao() {
        return parametersDao;
    }


    public List<DefaultParameterValue> getDefaultMaterialParametersValues(boolean isComposite, int materialId,
                                                                          UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        List<ParameterValues> allParameterValues = parameterValuesDao.findByMaterialId(materialId);
        //key - param id, value -> param value at default temperature if exists
        Map<Integer, BigDecimal> existedDefaultParameterValues = allParameterValues.stream()
                .map(pv -> Pair.of(pv.getId(),
                        DefaultParameterValue.extractValueForTemp(DEFAULT_TEMPERATURE, pv)))
                .filter(pair -> pair.getValue() != null)
                .collect(toMap(Pair::getKey, Pair::getValue));
        return findAllForMaterialType(isComposite, currentUser).stream()
                .map(paramEntity -> DefaultParameterValue
                        .extractFromParameterEntity(paramEntity, DEFAULT_TEMPERATURE,
                                existedDefaultParameterValues.get(paramEntity.getId())))
                .collect(toList());
    }


    private List<ParametersEntityTable> findAllForMaterialType(boolean isComposite, UserEntity currentUser) {
        return parametersDao.findAllForMaterialType(isComposite);
    }


    public List<ParameterSample> getMaterialParameterValues(boolean isComposite, int materialId, int parameterId,
                                                            UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        ParameterValues paramValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        return paramValues == null ? emptyList() : paramValues.getValuesList();
    }


    public List<ParameterSample> setMaterialParameterValues(boolean isComposite, int materialId, int parameterId,
                                                            List<ParameterSample> newParameterValues, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        checkParameterExists(parameterId, isComposite);

        // delete old values if they exists
        parameterValuesDao.deleteParameterValuesForMaterial(materialId, parameterId);
        deleteParameterFileForMaterial(materialId, parameterId);
        // and add new
        parameterValuesDao.insertParameterValuesForMaterial(materialId, parameterId, newParameterValues);
        ParameterValues paramValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        return paramValues == null ? emptyList() : paramValues.getValuesList();
    }


    public List<ParameterSample> upsertMaterialParameterValue(boolean isComposite, int materialId, int parameterId,
                                                              ParameterSample newParameterValue, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        checkParameterExists(parameterId, isComposite);

        parameterValuesDao.upsertParameterValue(materialId, parameterId, newParameterValue);
        ParameterValues paramValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        return paramValues == null ? emptyList() : paramValues.getValuesList();
    }


    public List<ParameterSample> updateMaterialParameterValue(boolean isComposite, int materialId, int parameterId,
                                                              ParameterSample newParameterValue, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        checkParameterExists(parameterId, isComposite);
        checkPointExists(newParameterValue.getId());
        parameterValuesDao.updatePoint(newParameterValue);
        ParameterValues paramValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        return paramValues == null ? emptyList() : paramValues.getValuesList();
    }


    public void deleteParameterValue(boolean isComposite, int materialId, int parameterId,
                                     int pointId, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, delete, currentUser);
        checkParameterExists(parameterId, isComposite);
        parameterValuesDao.deletePoint(pointId);
    }


    private void deleteParameterFileForMaterial(int materialId, int parameterId) {
        ParameterValues parameterValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        if (parameterValues == null || !parameterValues.hasFile()) {
            return;
        }
        FileEntity parameterFile = fileManager.findOne(parameterValues.getParameterFileId(), null);
        if (parameterFile == null) {
            throw new EntityUpdateException(msgSrc.getMessage(PARAMETER_FILE_NOT_FOUND,
                    new Object[]{parameterValues.getParameterFileId()}, null, null));
        }
        fileManager.deleteParameterFile(parameterValues.getParameterFileId(), null);
    }


    public List<ParameterSample> deleteMaterialParameterValues(boolean isComposite, int materialId, int parameterId,
                                                               UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        checkParameterExists(parameterId, isComposite);
        ParameterValues deletedParameterValues = parameterValuesDao.findByMaterialAndParameterIds(materialId, parameterId);
        parameterValuesDao.deleteParameterValuesForMaterial(materialId, parameterId);
        deleteParameterFileForMaterial(materialId, parameterId);
        return deletedParameterValues == null ? emptyList() : deletedParameterValues.getValuesList();
    }


    private void checkMaterialExistsAndAccessible(boolean isComposite, int materialId, AccessRights desiredAccess,
                                                  UserEntity currentUser) {
        if (isComposite) {
            compositeMaterialManager.checkEntityExistsAndAccesible(materialId, desiredAccess, currentUser);
        } else {
            materialManager.checkEntityExistsAndAccesible(materialId, desiredAccess, currentUser);
        }
    }


    private void checkParameterExists(Integer paramId, boolean isComposite) {
        if (!parametersDao.exists(paramId, isComposite)) {
            throw new NotFoundException(msgSrc.getMessage(PARAMETER_NOT_FOUND, new Object[]{paramId}, null));
        }
    }


    private void checkPointExists(int pointId) {
        if (!parameterValuesDao.pointExists(pointId)) {
            throw new NotFoundException(msgSrc.getMessage(PARAMETER_POINT_NOT_FOUND, new Object[]{pointId}, null));
        }
    }


    @Override
    public void checkDelete(ParametersEntityTable entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), PARAMETER_USED_IN_MATERIALS,
                treeDao::findMaterialsByParameterId, currentUser);
        super.checkDelete(entity, currentUser);
    }
}
