package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.StandardDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.StandardEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nullable;

import static ru.crp.lvms.utils.MessageUtils.STANDARD_HAS_RELATION;

@Service
public class StandardManager extends BaseManagerSuggest<StandardEntity> {
    @Autowired
    private FileManager fileManager;

    @Autowired
    private StandardDao standardDao;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private ValidationHelper validationHelper;

    @Override
    protected BaseSpaceDao<StandardEntity> aliveDao() {
        return standardDao;
    }

    public boolean exists(Integer standardId) {
        return standardDao.exists(standardId);
    }


    public int createStandardIfNotExists(@Nullable Integer standardId, String standardName) {
        if (standardId != null) {
            return standardId;
        }
        StandardEntity existedEntity = standardDao.findByName(standardName);
        if (existedEntity != null) {
            return existedEntity.getId();
        }
        StandardEntity newEntity = new StandardEntity();
        newEntity.setName(standardName);
        return standardDao.insertAndRead(newEntity).getId();
    }


    @Override
    public void checkDelete(StandardEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), STANDARD_HAS_RELATION,
                simulationManager::getSimulationsByStandardId, currentUser);
        super.checkDelete(entity, currentUser);
    }


    @Override
    public int create(StandardEntity entity, @Nullable UserEntity currentUser) {
        if (currentUser != null) {
            entity.setOwnerId(currentUser.getId());
        }
        return super.create(entity, currentUser);
    }


    public StandardEntity createAndRead(StandardEntity newStandard,
                                        @Nullable MultipartFile standardFileContent,
                                        @Nullable MultipartFile standardLoadFileContent,
                                        UserEntity currentUser) {
        newStandard.setOwnerId(currentUser.getId());
        StandardEntity createdStandard = super.createAndRead(newStandard, currentUser);
        if (standardFileContent != null) {
            createFile(standardFileContent, FileType.StandardFile, createdStandard, currentUser);
        }
        if (standardLoadFileContent != null) {
            createFile(standardLoadFileContent, FileType.StandardLoadFile, createdStandard, currentUser);
        }
        if (standardFileContent != null || standardLoadFileContent != null) {
            return standardDao.updateAndRead(createdStandard);
        }
        return createdStandard;
    }


    private void createFile(MultipartFile fileContent, FileType type, StandardEntity standard, UserEntity currentUser) {
        FileEntity newFile = fileManager.createFile(fileContent, standard.getId(),
                type, currentUser);
        if (type == FileType.StandardFile) {
            standard.setStandardFileId(newFile.getId());
        } else {
            standard.setLoadFileId(newFile.getId());
        }
    }

    public FileEntity checkAndGetFile(Integer standardId, FileType fileType, UserEntity currentUser) {
        StandardEntity standardEntity = checkEntityExistsAndAccesible(standardId, AccessRights.view, currentUser);
        Integer fileId = getFileByType(standardEntity, fileType);
        if (fileId == null) {
            return null;
        }
        return fileManager.findOne(fileId, currentUser);
    }

    public void updateFile(int standardId, FileType fileType, MultipartFile newFileContent,
                           UserEntity currentUser) {
        StandardEntity standardEntity = checkEntityExistsAndAccesible(standardId, AccessRights.update, currentUser);
        Integer fileId = getFileByType(standardEntity, fileType);
        if (fileId == null) {
            createFile(newFileContent, fileType, standardEntity, currentUser);
            standardDao.update(standardEntity);
        } else {
            fileManager.updateFileContent(fileId, newFileContent, currentUser);
        }
    }


    private Integer getFileByType(StandardEntity standardEntity, FileType fileType) {
        if (fileType == FileType.StandardFile) {
            return standardEntity.getStandardFileId();
        } else if (fileType == FileType.StandardLoadFile) {
            return standardEntity.getLoadFileId();
        } else {
            throw new IllegalArgumentException("Unexpected file type: " + fileType);
        }
    }

    private void setFileIdByType(@Nullable Integer newFileId, StandardEntity standardEntity, FileType fileType) {
        if (fileType == FileType.StandardFile) {
            standardEntity.setStandardFileId(newFileId);
        } else if (fileType == FileType.StandardLoadFile) {
            standardEntity.setLoadFileId(newFileId);
        } else {
            throw new IllegalArgumentException("Unexpected file type: " + fileType);
        }
    }


    public void deleteFile(Integer standardId, FileType fileType, @Nullable UserEntity currentUser) {
        StandardEntity standardEntity = checkEntityExistsAndAccesible(standardId, AccessRights.update, currentUser);
        deleteFile(standardEntity, fileType, currentUser);
    }


    private void deleteFile(StandardEntity standardEntity, FileType fileType, @Nullable UserEntity currentUser) {
        Integer fileId = getFileByType(standardEntity, fileType);
        if (fileId == null) {
            return;
        }
        setFileIdByType(null, standardEntity, fileType);
        update(standardEntity, null);
        fileManager.deleteFile(fileId, currentUser);
    }


    @Override
    public void delete(StandardEntity entity, @Nullable UserEntity currentUser) {
        checkDelete(entity, currentUser);
        if (entity.getStandardFileId() != null) {
            deleteFile(entity, FileType.StandardFile, currentUser);
        }
        if (entity.getLoadFileId() != null) {
            deleteFile(entity, FileType.StandardLoadFile, currentUser);
        }
        aliveDao().delete(entity.getId());
    }
}
