package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.TechnologyDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import static ru.crp.lvms.utils.MessageUtils.TECHNOLOGY_USED_IN_MATERIAL;

@Service
public class TechnologyManager extends BaseManagerSuggest<TechnologyEntity> implements ImageHandler<TechnologyEntity> {
    @Autowired
    private TechnologyDao technologyDao;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private SimulationManager simulationManager;


    public TechnologyEntity createTechnology(TechnologyEntity newTechnology,
                                             @Nullable MultipartFile imageFileContent,
                                             @Nullable MultipartFile descriptionFileContent,
                                             UserEntity currentUser) {
        newTechnology.setOwnerId(currentUser.getId());
        TechnologyEntity createdTechnology = createAndRead(newTechnology, currentUser);
        if (imageFileContent != null) {
            checkIsImage(imageFileContent);
            createFile(imageFileContent, FileType.ImageFile, createdTechnology, currentUser);
        }
        if (descriptionFileContent != null) {
            createFile(descriptionFileContent, FileType.DescriptionFile, createdTechnology, currentUser);
        }
        if (imageFileContent != null || descriptionFileContent != null) {
            return updateAndRead(newTechnology, currentUser);
        }
        return createdTechnology;
    }


    private void createFile(MultipartFile fileContent, FileType type, TechnologyEntity technology, UserEntity currentUser) {
        FileEntity newFile = fileManager.createFile(fileContent, technology.getId(), type, currentUser);
        if (type == FileType.ImageFile) {
            technology.setImageFileId(newFile.getId());
        } else {
            technology.setDescriptionFileId(newFile.getId());
        }
    }


    public FileEntity checkAndGetDescriptionFile(Integer technologyId, UserEntity currentUser) {
        TechnologyEntity technology = checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser);
        Integer fileId = technology.getDescriptionFileId();
        if (fileId == null) {
            return null;
        }
        return fileManager.findOne(fileId, currentUser);
    }


    public void updateDescriptionFile(int technologyId, MultipartFile newFileContent,
                                      UserEntity currentUser) {
        TechnologyEntity technology = checkEntityExistsAndAccesible(technologyId, AccessRights.update, currentUser);
        Integer fileId = technology.getDescriptionFileId();
        if (fileId == null) {
            createFile(newFileContent, FileType.DescriptionFile, technology, currentUser);
            aliveDao().update(technology);
        } else {
            fileManager.updateFileContent(fileId, newFileContent, currentUser);
        }
    }


    public void deleteDescriptionFile(Integer technologyId, UserEntity currentUser) {
        TechnologyEntity technology = checkEntityExistsAndAccesible(technologyId, AccessRights.update, currentUser);
        Integer descriptionFileId = technology.getDescriptionFileId();
        if (descriptionFileId == null) {
            return;
        }
        technology.setDescriptionFileId(null);
        update(technology, currentUser);
        fileManager.deleteFile(descriptionFileId, currentUser);
    }


    @Override
    public void delete(TechnologyEntity entity, @Nullable UserEntity currentUser) {
        checkDelete(entity, currentUser);

        Optional<Integer> imageFileId = Optional.ofNullable(entity.getImageFileId());
        Optional<Integer> descriptionFileId = Optional.ofNullable(entity.getDescriptionFileId());
        entity.setImageFileId(null);
        entity.setDescriptionFileId(null);

        update(entity, currentUser);

        imageFileId.ifPresent(id -> fileManager.deleteFile(id, currentUser));
        descriptionFileId.ifPresent(id -> fileManager.deleteFile(id, currentUser));

        aliveDao().delete(entity.getId());
    }


    @Override
    public void checkDelete(TechnologyEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), TECHNOLOGY_USED_IN_MATERIAL,
                compositeMaterialManager::getCompositeMaterialsUsingTechnology, currentUser);
        super.checkDelete(entity, currentUser);
    }


    @Override
    public BaseSpaceDao<TechnologyEntity> aliveDao() {
        return technologyDao;
    }


    @Override
    public FileManager fileManager() {
        return fileManager;
    }


    public List<SimulationEntity> getSimulationssForTechnology(Integer technologyId, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser);
        return simulationManager.getSimulationsForTechnology(technologyId, currentUser);
    }

    public PageImpl<SimulationEntity> getSimulationsPageForTechnology(Integer technologyId, Pageable page,
                                                                      Map<String, Object> filters, UserEntity currentUser) {
        checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser);
        return simulationManager.getSimulationsPageForTechnology(technologyId, page, filters, currentUser);
    }

    public void addLinkToEquipment(int technologyId, int equipmentId) {
        technologyDao.addLinkToEquipment(technologyId, equipmentId);
    }

    public void removeLinkToEquipment(int technologyId, int equipmentId) {
        compositeMaterialManager.removeTechnologyEquipment(technologyId, equipmentId);
        technologyDao.removeLinkToEquipment(technologyId, equipmentId);
    }
}

