package ru.crp.lvms.managers.effective.characteristics;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.dao.DefectNameDao;
import ru.crp.lvms.model.dao.effective.characteristics.ECDao;
import ru.crp.lvms.model.dao.effective.characteristics.ECDefectDao;
import ru.crp.lvms.model.dao.effective.characteristics.ECParamDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.pdf.reports.ECReportInput;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.tasks.fileGenerators.ECParameters;
import ru.crp.lvms.tasks.fileGenerators.ThermalSimulationParameters;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Date;
import java.util.List;
import java.util.Map;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.MessageUtils.EC_USED_IN_SIMULATION;

@Service
public class ECManager extends BaseManagerSuggest<ECEntity> implements ImageHandler<ECEntity> {
    @Autowired
    private ECDao ECDao;

    @Autowired
    private ECDefectDao defectDao;

    @Autowired
    private ECParamDao paramDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private DefectNameDao defectNameDao;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private ResultsManager resultsManager;

    @Nonnull
    @Override
    public ECEntity createAndRead(ECEntity newEc, @Nullable UserEntity currentUser) {
        ECEntity createdEc = super.createAndRead(newEc, currentUser);
        // adding all parameters to effected characteristic using default values
        paramDao.getAllParamIds().stream()
                .map(paramId -> new ECParamEntity(createdEc.getId(), paramId))
                .forEach(paramDao::insert);

        // adding all defects to effected characteristic using default values
        defectNameDao.findAllowedForEC().stream()
                .map(defectName -> new ECDefectEntity(defectName.getId(), createdEc.getId()))
                .forEach(defectDao::insert);
        return extend(createdEc, currentUser);
    }

    public ECEntity extend(ECEntity entity, UserEntity currentUser) {
        entity.setDefects(defectDao.getByCharacteristicId(entity.getId()));
        entity.setParams(paramDao.getByCharacteristicId(entity.getId()));
        return entity;
    }

    public List<ECEntity> extend(List<ECEntity> entities, UserEntity currentUser) {
        return entities.stream()
                .map(entity -> extend(entity, currentUser))
                .collect(toList());
    }

    public List<ECEntity> getMaterialCharacteristics(int materialId, UserEntity currentUser) {
        return aliveDao().findByMaterialId(materialId);
    }

    public PageImpl<ECEntity> getMaterialCharacteristicsPage(int materialId, Map<String, Object> filters,
                                                             Pageable pageable, UserEntity user) {
        filters.put("materialId", materialId);
        return aliveDao().find(filters, pageable);
    }

    @Override
    public ECDao aliveDao() {
        return ECDao;
    }

    @Override
    public FileManager fileManager() {
        return fileManager;
    }

    public List<ECDefectEntity> getECDefects(Integer ecId) {
        return defectDao.getByCharacteristicId(ecId);
    }


    public PageImpl<ECDefectEntity> getECDefectsPage(Pageable pageable, Map<String, Object> filters) {
        return defectDao.find(filters, pageable);
    }

    public List<ECParamEntity> getECParameters(Integer ecId,
                                               ECParamEntity.Type type) {
        return paramDao.getByCharacteristicId(ecId, type);
    }

    public ECDefectEntity getECDefect(Integer ecId, Integer defectId, PortalUser currentUser) {
        return defectDao.findOne(defectId, ecId);
    }

    public ECDefectEntity createECDefect(ECDefectEntity newDefectDTO) {
        ECDefectEntity existedDefect = defectDao.getByCharacteristicIdAndDefectNameId(newDefectDTO.getDefectNameId(), newDefectDTO.getEcId());
        if (existedDefect != null) {
            throw new BadRequestException("Дефект '" + existedDefect.getDefectName() + "' уже добавлен к эффективной характеристике");
        }
        return defectDao.insertAndRead(newDefectDTO);
    }

    public ECDefectEntity updateECDefect(ECDefectEntity existedDefect, ECDefectEntity updateDTO) {
        existedDefect.setLength(updateDTO.getLength());
        existedDefect.setWidth(updateDTO.getWidth());
        existedDefect.setThickness(updateDTO.getThickness());
        existedDefect.setPercentage(updateDTO.getPercentage());
        return defectDao.updateAndRead(existedDefect);
    }

    public void deleteECDefect(ECDefectEntity defect) {
        defectDao.delete(defect.getId());
    }

    public ECParamEntity createECParam(ECParamEntity newParamDTO) {
        ECParamEntity existedParam = paramDao.getByCharacteristicIdAndParamId(newParamDTO.getEcId(), newParamDTO.getParameterId());
        if (existedParam != null) {
            throw new BadRequestException("Параметр '" + existedParam.getParameterName() + "' уже добавлен к эффективной характеристике");
        }
        return paramDao.insertAndRead(newParamDTO);
    }

    public ECParamEntity getECParameter(Integer ecId, Integer id, PortalUser currentUser) {
        return paramDao.findOne(id, ecId);
    }

    public ECParamEntity updateECParam(ECParamEntity existedParam, ECParamEntity updateDTO) {
        existedParam.setParameterValue(updateDTO.getParameterValue());
        return paramDao.updateAndRead(existedParam);
    }

    public void deleteECParam(ECParamEntity param) {
        paramDao.delete(param.getId());
    }

    public ECParameters getEffectiveCharacteristicsParameters(Integer effectiveCharacteristicId) {
        return ECDao.getEffectiveCharacteristicsParameters(effectiveCharacteristicId);
    }

    public ThermalSimulationParameters getThermalSimulationParameters(Integer effectiveCharacteristicId, boolean isEpoxidePresent) {
        return ECDao.getThermalSimulationParameters(effectiveCharacteristicId, isEpoxidePresent);
    }

    @Override
    public void checkDelete(ECEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), EC_USED_IN_SIMULATION,
                simulationManager::getSimulationsByECId, currentUser);
        super.checkDelete(entity, currentUser);
    }

    @Override
    public void delete(ECEntity ecToDelete, @Nullable UserEntity currentUser) {
        checkDelete(ecToDelete, currentUser);
        resultsManager.deleteResults(Mapper.map(ecToDelete, ResultsParent.class), true, CreatedBy.Auto, CreatedBy.Manual);
        aliveDao().delete(ecToDelete.getId());
    }

    @NotNull
    public ECReportInput formEcReportInput(ECEntity ecEntity) throws IOException {
        ECReportInput reportInfo = new ECReportInput();
        reportInfo.setEcName(ecEntity.getCompositeMaterialName());
        reportInfo.setCreationDate(new Date());
        reportInfo.setUserName(ecEntity.getOwner());
        reportInfo.setMesoParams(getECParameters(ecEntity.getId(), ECParamEntity.Type.Mesostructure));
        reportInfo.setMicroParams(getECParameters(ecEntity.getId(), ECParamEntity.Type.Microstructure));
        reportInfo.setDefects(getECDefects(ecEntity.getId()));
        reportInfo.setEcMezostructure(resultsManager.getECResultValues(ecEntity.getId(), ResultType.mezostrParams));
        reportInfo.setEcMicrostructure(resultsManager.getECResultValues(ecEntity.getId(), ResultType.microstrParams));
        Path microPic = resultsManager.getResultPictureByName(ecEntity.getId(),
                TaskParentType.effectiveCharacteristic, ResultsManager.MICRO_PICTURE_NAME);
        reportInfo.setPathToMicrostructurePicture(microPic);
        Path mezoPic = resultsManager.getResultPictureByName(ecEntity.getId(),
                TaskParentType.effectiveCharacteristic, ResultsManager.MEZO_PICTURE_NAME);
        reportInfo.setPathToMesostructurePicture(mezoPic);
        return reportInfo;
    }

    public List<ECEntity> getEffectiveCharacteristicsByCompositeMaterialId(Integer compositeMaterialId, UserEntity userEntity) {
        return ECDao.getByCompositeMaterialId(compositeMaterialId);
    }
}
