package ru.crp.lvms.managers.tasks;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.tosca.enums.ToscaJsonParametersParents;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import static java.util.Collections.singletonList;
import static ru.crp.lvms.managers.solvers.SolverConfigManager.*;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.ANIMATION_NAME;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.SMOOTHING_NAME;

@Service
public class OptimizationTaskManager {

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ClustersManager clustersManager;

    public List<TaskEntity> createChainOfTasksForOptimization(TaskEntity newTask, UserEntity user) {
        OptimizationEntity parentOptimization = optimizationManager.findOne(newTask.getParentId(), user);
        if (parentOptimization.getSolverCaoId() != null) {
            return createCaoTasks(newTask, parentOptimization);
        } else if (parentOptimization.getSolverCaeId() != null) {
            return createCaeTasks(newTask, parentOptimization);
        } else {
            throw new BadRequestException(String.format("Тип решателя %s не поддерживается", OptimizationCao.valueOf(parentOptimization.getSolverCaoId())));
        }
    }


    private List<TaskEntity> createCaoTasks(TaskEntity newTask, OptimizationEntity parentOptimization) {
        switch (OptimizationCao.valueOf(parentOptimization.getSolverCaoId())) {
            case Tosca:
                try {
                    return createToscaTasks(newTask, parentOptimization);
                } catch (CloneNotSupportedException e) {
                    throw new InternalError("Clone not supported", e);
                }
            case ModeFrontier:
                newTask.setType(Type.modFrontierSolving);
                newTask.setSolverName(OptimizationCao.ModeFrontier.getSolverConfigName());
                break;
            case Optimus:
                newTask.setType(Type.optimusSolving);
                newTask.setSolverName(OptimizationCao.Optimus.getSolverConfigName());
                break;
            case Optistruct:
                newTask.setSolverName(OptimizationCao.Optistruct.getSolverConfigName());
                List<FileEntity> generatedFEMFiles = optimizationManager.getOptimizationFemFiles(parentOptimization);
                if (generatedFEMFiles.isEmpty()) {
                    newTask.setType(Type.optistructGenerateFEM);
                } else {
                    newTask.setType(Type.optistructSolving);
                }
                break;
            default:
                throw new BadRequestException(String.format("Тип решателя %s не поддерживается",
                        OptimizationCao.valueOf(parentOptimization.getSolverCaoId())));
        }

        return singletonList(newTask);
    }

    private List<TaskEntity> createToscaTasks(TaskEntity newTask, OptimizationEntity parentOptimization) throws CloneNotSupportedException {
        List<OptimizationJsonSetting> settings =
                optimizationJsonSettingsDao.findSettingsByParent(parentOptimization.getId(),
                        ToscaJsonParametersParents.AnimationParameters.getId());
        boolean isSmoothingEnabled = settings.stream()
                .filter(ojs -> Objects.equals(ojs.getXmlName(), SMOOTHING_NAME))
                .map(OptimizationJsonSetting::getBoolValue).findFirst().orElse(false);
        boolean isAnimationEnabled = settings.stream()
                .filter(ojs -> Objects.equals(ojs.getXmlName(), ANIMATION_NAME))
                .map(OptimizationJsonSetting::getBoolValue).findFirst().orElse(false);

        List<TaskEntity> tasks = new ArrayList<>();

        if (parentOptimization.getStatus() != SolvingStatus.Completed) {
            newTask.setType(Type.toscaSolving);
            newTask.setSolverName(OptimizationCao.Tosca.getSolverConfigName());
            tasks.add(newTask);
        } else if (!isSmoothingEnabled && !isAnimationEnabled) {
            throw new BadRequestException("Не выбран метод постпроцессинга (animation, smoothing)");
        }

        if (isSmoothingEnabled) {
            TaskEntity smoothing = newTask.clone();
            smoothing.setStatusId(CalculationStatus.Draft.getId());
            smoothing.setType(Type.toscaSmoothing);
            smoothing.setSolverName(SolverConfigManager.TOSCA_POSTPROC_SMOOTH_NAME);
            tasks.add(smoothing);
        }

        if (isAnimationEnabled) {
            TaskEntity animation = newTask.clone();
            animation.setStatusId(CalculationStatus.Draft.getId());
            animation.setType(Type.toscaAnimation);
            animation.setSolverName(SolverConfigManager.TOSCA_POSTPROC_ANIM_NAME);
            tasks.add(animation);
        }
        return tasks;
    }


    private List<TaskEntity> createCaeTasks(TaskEntity newTask, OptimizationEntity parentOptimization) {
        ClusterType clusterType = clustersManager.getClusterForTask(newTask).getType();
        if (Objects.equals(ClusterType.valueOf(clusterType.getId()), ClusterType.Slurm)) {
            if (Objects.equals(parentOptimization.getSolverCaeId(), OptimizationCae.ANSYS.getId())) {
                newTask.setType(Type.solving);
                newTask.setSolverName(ANSYS_FOR_SLURM_SOLVER_NAME);
            } else if (Objects.equals(parentOptimization.getSolverCaeId(), OptimizationCae.LSDYNA.getId())) {
                newTask.setType(Type.solving);
                newTask.setSolverName(LS_DYNA_FOR_SLURM_SOLVER_NAME);
            } else {
                throw new BadRequestException(String.format("Тип решателя %s не поддерживается", OptimizationCae.valueOf(parentOptimization.getSolverCaeId())));
            }
        }
        else if (Objects.equals(ClusterType.valueOf(clusterType.getId()), ClusterType.Space)) {
            if (Objects.equals(parentOptimization.getSolverCaeId(), OptimizationCae.ANSYS.getId())) {
                newTask.setType(Type.solving);
                newTask.setSolverName(OptimizationCae.ANSYS.getSolverConfigName());
            } else  if (Objects.equals(parentOptimization.getSolverCaeId(), OptimizationCae.NXNASTRAN.getId())) {
                newTask.setType(Type.solving);
                newTask.setSolverName(OptimizationCae.NXNASTRAN.getSolverConfigName());
            } else {
                    throw new BadRequestException(String.format("Тип решателя %s не поддерживается", OptimizationCae.valueOf(parentOptimization.getSolverCaeId())));
                }
            }
        else {
            throw new BadRequestException(String.format("Тип кластера %s не поддерживается",ClusterType.valueOf(clusterType.getId())));
        }
        return singletonList(newTask);
    }
}
