package ru.crp.lvms.managers.tasks;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.lvms.cluster.ClustersManager;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.cluster.model.ClusterType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.solvers.SolverConnectionManager;
import ru.crp.lvms.model.dao.NodesDao;
import ru.crp.lvms.model.dao.UserDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.tasks.misc.TaskCommandsFactory;
import ru.crp.lvms.utils.AuthUtils;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

@Service
public class StopTaskManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(StopTaskManager.class);

    @Autowired
    private TaskManager taskManager;

    @Autowired
    private ClustersManager clustersManager;

    @Autowired
    private TaskCommandsFactory taskCommandsFactory;

    @Autowired
    private UserDao userDao;

    @Autowired
    private NodesDao nodesDao;

    @Autowired
    private TaskStatusUpdater taskStatusUpdater;

    @Autowired
    private SolverConnectionManager solverConnectionManager;

    public void removeTaskFromQueue(TaskEntity taskToRemove, UserEntity user) {
        if (!taskManager.isWaiting(taskToRemove)) {
            throw new BadRequestException("Из очереди можно удалять только ожидающие задачи");
        }
        try {
            ClusterEntity taskCluster = clustersManager.getClusterForTask(taskToRemove);
            taskCommandsFactory.getClusterCommandManager(taskCluster.getType()).cancelTask(taskToRemove, null);
        } catch (Exception ex) {
            throw new InternalException(String.format("Internal exception while cancelling task %d", taskToRemove.getId()));
        }
    }

    public Map<String, Object> stopAndUpdateTask(int taskId, UserEntity currentUser) throws Exception {
        TaskEntity taskToStop = taskManager.findOne(taskId, currentUser);
        if (taskToStop == null) {
            throw new NotFoundException(String.format("Задача с номером %d не найдена", taskId));
        }
        // todo change bench tasks to use same mechanism as slurm
        if (clustersManager.isSlurmTask(taskToStop)) {
            taskCommandsFactory.getClusterCommandManager(ClusterType.Slurm).cancelTask(taskToStop, null);
            HashMap<String, Object> result = new HashMap<>(2);
            result.put("message", "Started stopping task");
            result.put("status", 0);
            return result;
        }
        Map<String, Object> stopTaskResponse = stopTask(taskToStop, currentUser);
        Object status = stopTaskResponse.get("status");
        String message = (String) stopTaskResponse.get("message");
        if (status != null) {
            if ((Integer) status == 0) {
                taskStatusUpdater.changeTaskStatus(taskToStop, CalculationStatus.Terminating, currentUser);
            } else if (message != null) {
                LOGGER.warn(String.format("Error while stopping task %d: %s", taskId, message));
                throw new BadRequestException(message);
            }
        }
        return stopTaskResponse;
    }


    private Map<String, Object> stopTask(TaskEntity taskToStop, UserEntity currentUser) throws IOException {
        AuthUtils.ownerOrGreaterOrAdmin(userDao.findOne(taskToStop.getOwnerId()), currentUser);
        if (taskToStop.getNodeId() == null) {
            LOGGER.error("Trying to stop task which is not started");
            throw new BadRequestException("Задача не была запущена");
        }

        checkTaskIsStoppable(taskToStop);

        NodeEntity nodeEntity = nodesDao.findOne(taskToStop.getNodeId());
        return solverConnectionManager.sendStopTaskRequest(nodeEntity.getUrl(), taskToStop.getId());
    }


    private void checkTaskIsStoppable(TaskEntity taskToStop) {
        CalculationStatus currentTaskStatus = CalculationStatus.valueOf(taskToStop.getStatusId());
        if (currentTaskStatus.isTerminalState()) {
            throw new BadRequestException("Задача уже завершена");
        }
        if (currentTaskStatus == CalculationStatus.Terminating) {
            throw new BadRequestException("Завершение задачи уже начато");
        }
        boolean taskCanBeStopped = CalculationStatus.getStoppableStatuses().contains(CalculationStatus.valueOf(taskToStop.getStatusId()));
        if (!taskCanBeStopped) {
            throw new BadRequestException("Задача находится в непрырываемом состоянии или недостаточно прав на ее остановку");
        }
    }
}
