package ru.crp.lvms.model.dao;

import com.fasterxml.classmate.TypeResolver;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.ProtectedMarker;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmlb.model.dao.JdbcCrudDaoSuggest;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.cmlb.model.utils.EntityUtils;
import ru.crp.lvms.managers.image.ImageFileContainer;
import ru.crp.lvms.model.mappers.CustomBeanPropertyRowMapper;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

@Transactional(propagation = Propagation.REQUIRED)
public class BaseSpaceDao<T extends Entity> extends JdbcCrudDaoSuggest<T> {

    @Override
    public void setBeanClass(Class<T> beanClass) {
        setRowMapper(new CustomBeanPropertyRowMapper<>(new ObjectMapper(), beanClass));
    }

    public List<T> findByColumnName(String columnName, String value) {
        return find(" WHERE " + getAlias() + "." + columnName + "= ?", value);
    }

    @Override
    protected String getColumnNameForCustomFilters(String filterName) {
        //is used with suggest
        if ("pathNames".equals(filterName)) {
            return "path_names";
        }
        return super.getColumnNameForCustomFilters(filterName);
    }

    // should be used instead of final method getJdbcTemplate() which can not be overriden in proxy
    public JdbcTemplate jdbcTemplate() {
        return super.getJdbcTemplate();
    }


    protected int insertAndRead(String insertSql, Map<String, Object> params) {
        MapSqlParameterSource paramSource = new MapSqlParameterSource(params);
        GeneratedKeyHolder keyHolder = new GeneratedKeyHolder();
        this.getNamedParameterJdbcTemplate().update(insertSql, paramSource, keyHolder, new String[]{"id"});
        return keyHolder.getKey().intValue();
    }


    public List<T> suggest(String additionalWhere, SuggestRequest suggestRequest, int currentUserId) {
        Class<?> entityClass = getEntityClass();
        StringBuilder sql = new StringBuilder(additionalWhere);
        boolean isProtectedObject = false;
        if (ProtectedMarker.class.isAssignableFrom(entityClass)) {
            sql.append(" AND user_can_view(" + getAlias() + ".path_names, TRUE, ?) ");
            isProtectedObject = true;
        } else {
            sql.append(" AND true");
        }
        for (SuggestRequest.Entry query : suggestRequest.getQueries()) {
            String queryColumn = query.getColumn();
            String customFilterName = getColumnNameForCustomFilters(queryColumn);
            if (customFilterName == null
                    && !entityMappingInfo.containsMappingForColumn(queryColumn)
                    && !entityMappingInfo.containsMappingForField(queryColumn)) {
                continue;
            }
            if (customFilterName != null) {
                queryColumn = customFilterName;
            } else if (entityMappingInfo.containsMappingForField(queryColumn)) {
                queryColumn = entityMappingInfo.getColumnName(queryColumn);
            }
            String operation = "=";
            if (query.getValue().contains("%")) {
                operation = "LIKE";
                sql.append(" and LOWER(" + getAlias() + "." + queryColumn + ") " + operation + " '"
                        + query.getValue().toLowerCase() + "'");
            } else {
                sql.append(" and " + getAlias() + "." + queryColumn + " " + operation + " '"
                        + query.getValue().toLowerCase() + "'");
            }
        }
        List<Integer> excludedIds = suggestRequest.getExcludedIds();
        if (!excludedIds.isEmpty()) {
            sql.append(" and " + getAlias() + ".id NOT IN" + EntityUtils.sqlForListParams(excludedIds));
        }
        sql.append(" limit " + suggestRequest.getPageLimit());

        List<Object> argsList = new ArrayList<>(excludedIds);
        if (isProtectedObject) {
            argsList.add(0, currentUserId);
        }
        return find(sql.toString(), argsList.toArray());
    }

    private Class<?> getEntityClass() {
        return new TypeResolver().resolve(getClass())
                .typeParametersFor(JdbcCrudDaoSuggest.class)
                .get(0).getErasedType();
    }


    // is not used in this project. Todo think about refactoring
    @Override
    protected boolean isCAETree() {
        return false;
    }

    public void updateImageFile(ImageFileContainer imageFileContainer, @Nullable Integer newImageFileId) {
        getJdbcTemplate()
                .update(String.format("UPDATE %s SET image_file_id = ? WHERE id = ?", getTableName()), newImageFileId, imageFileContainer.getId());
    }
}
