package ru.crp.lvms.model.dao.effective.characteristics;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import ru.crp.lvms.managers.ParametersManager;
import ru.crp.lvms.model.dao.BaseCompositeDao;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntityTable;
import ru.crp.lvms.tasks.fileGenerators.ECParameters;
import ru.crp.lvms.tasks.fileGenerators.ThermalSimulationParameters;

import java.math.BigDecimal;
import java.util.List;

@Repository
public class ECDao extends BaseCompositeDao<ECEntityTable, ECEntity> {

    private static final int ELASTICITY_MODULUS_ID = 1;
    private static final int POISSON_RATIO_ID = 2;
    private static final int STRENGTH_LIMIT_ID = 3;
    private static final int THERMAL_CONDUCTIVITY_ID = 5;
    private static final int LINEAR_THERMAL_EXPANSION_ID = 6;
    private static final int FIBER_DIAMETER_ID = 1;
    private static final int FIBER_CONCENTRATION_ID = 2;
    private static final int VALUE_A_ID = 4;
    private static final int VALUE_B_ID = 5;
    private static final int VALUE_L_ID = 6;
    private static final int AVERAGE_LAYER_THICKNESS_ID = 13;
    private static final BigDecimal TEMPERATURE = ParametersManager.DEFAULT_TEMPERATURE;

    @Repository
    public static class TableDao extends BaseSpaceDao<ECEntityTable> {

    }

    @Autowired
    private TableDao tableDao;

    public List<ECEntity> findByMaterialId(int materialId) {
        return find("WHERE composite_material_id = ?", materialId);
    }

    @Override
    protected BaseSpaceDao<ECEntityTable> tableDao() {
        return tableDao;
    }

    private String getEffectiveCharacteristicsParameters(Integer id, Integer paramId) {
        List<String> resp = jdbcTemplate().query("SELECT param_value " +
                        "FROM   effective_characteristics_parameters_view AS ecpv " +
                        "WHERE  ecpv.effective_characteristic_id = ? " +
                        "AND    ecpv.param_id = ?;",
                new Object[]{id, paramId},
                (rs, rowNum) -> rs.getString("param_value")
        );
        return resp.isEmpty() ? null : resp.get(0);
    }

    private String getMaterialParameters(Integer id, Integer paramId, String field) {
        String sql = "SELECT value " +
                "FROM   effective_characteristics " +
                "JOIN   composite_materials  " +
                "ON     effective_characteristics.composite_material_id = composite_materials.id " +
                "JOIN   materials            ON composite_materials.<field> = materials.id " +
                "JOIN   parameters_materials ON materials.id = parameters_materials.material_id " +
                "WHERE  effective_characteristics.id = ? " +
                "AND    parameters_materials.parameter_id = ? " +
                "AND    parameters_materials.temp = ?;";
        List<String> resp = jdbcTemplate().query(sql.replace("<field>", field),
                new Object[]{id, paramId, TEMPERATURE},
                (rs, rowNum) -> rs.getString("value")
        );
        return resp.isEmpty() ? null : resp.get(0);
    }

    public ECParameters getEffectiveCharacteristicsParameters(Integer id) {
        ECParameters parameters = new ECParameters();

        parameters.setFiberDiameter(getEffectiveCharacteristicsParameters(id, FIBER_DIAMETER_ID));
        parameters.setFiberConcentration(getEffectiveCharacteristicsParameters(id, FIBER_CONCENTRATION_ID));
        parameters.setAverageLayerThickness(getEffectiveCharacteristicsParameters(id, AVERAGE_LAYER_THICKNESS_ID));
        parameters.setValueA(getEffectiveCharacteristicsParameters(id, VALUE_A_ID));
        parameters.setValueB(getEffectiveCharacteristicsParameters(id, VALUE_B_ID));
        parameters.setValueL(getEffectiveCharacteristicsParameters(id, VALUE_L_ID));

        parameters.setEpoxideElacticityModulus(getMaterialParameters(id, ELASTICITY_MODULUS_ID, "epoxide_id"));
        parameters.setEpoxidePoissonRatio(getMaterialParameters(id, POISSON_RATIO_ID, "epoxide_id"));
        parameters.setEpoxideStrengthLimit(getMaterialParameters(id, STRENGTH_LIMIT_ID, "epoxide_id"));

        parameters.setFiberElacticityModulus(getMaterialParameters(id, ELASTICITY_MODULUS_ID, "fiber_id"));
        parameters.setFiberPoissonRatio(getMaterialParameters(id, POISSON_RATIO_ID, "fiber_id"));
        parameters.setFiberStrengthLimit(getMaterialParameters(id, STRENGTH_LIMIT_ID, "fiber_id"));

        parameters.setMatrixElacticityModulus(getMaterialParameters(id, ELASTICITY_MODULUS_ID, "matrix_id"));
        parameters.setMatrixPoissonRatio(getMaterialParameters(id, POISSON_RATIO_ID, "matrix_id"));
        parameters.setMatrixStrengthLimit(getMaterialParameters(id, STRENGTH_LIMIT_ID, "matrix_id"));
        return parameters;
    }

    public ThermalSimulationParameters getThermalSimulationParameters(Integer id, boolean isEpoxidePresent) {
        String fiberTC = getMaterialParameters(id, THERMAL_CONDUCTIVITY_ID, "fiber_id");
        String fiberLTE = getMaterialParameters(id, LINEAR_THERMAL_EXPANSION_ID, "fiber_id");
        String matrixTC = getMaterialParameters(id, THERMAL_CONDUCTIVITY_ID, "matrix_id");
        String matrixLTE = getMaterialParameters(id, LINEAR_THERMAL_EXPANSION_ID, "matrix_id");

        ThermalSimulationParameters parameters = new ThermalSimulationParameters();
        parameters.setFiberLinearThermalExpansion(fiberLTE);
        parameters.setFiberThermalConductivity(fiberTC);
        parameters.setMatrixLinearThermalExpansion(matrixLTE);
        parameters.setMatrixThermalConductivity(matrixTC);
        parameters.setEpoxidePresent(isEpoxidePresent);

        if (isEpoxidePresent) {
            String epoxideTC = getMaterialParameters(id, THERMAL_CONDUCTIVITY_ID, "epoxide_id");
            String epoxideLTE = getMaterialParameters(id, LINEAR_THERMAL_EXPANSION_ID, "epoxide_id");
            parameters.setEpoxideLinearThermalExpansion(epoxideLTE);
            parameters.setEpoxideThermalConductivity(epoxideTC);
        }

        return parameters;
    }

    public List<ECEntity> getByCompositeMaterialId(Integer compositeMaterialId) {
        return find("WHERE composite_material_id = ?", compositeMaterialId);
    }
}
