package ru.crp.lvms.optimizations.base.controllers.files;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.validation.Valid;
import java.util.List;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.URLs.OPTIMIZATIONS_URL;

@RestController
@RequestMapping(value = OPTIMIZATIONS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptimizationCAEandCaDFilesController {
    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ConstructionElementManager constructionElementManager;

    @Autowired
    private FileManager fileManager;

    /**
     * CAE FILES
     **/
    @RequestMapping(value = "/{optimizationId}/CAEFiles/available/suggest", method = RequestMethod.POST)
    public List<FileEntity> suggestAvailableCAEFiles(@PathVariable("optimizationId") int optimizationId,
                                                     @RequestBody @Valid SuggestRequest suggestRequest, @CurrentUser PortalUser currentUser) {
        return suggestAvailableFiles(FileType.CAEFile, optimizationId, suggestRequest, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CAEFiles", method = RequestMethod.POST)
    public FileEntity addNewCAEModel(@PathVariable("optimizationId") Integer optimizationId,
                                     @RequestParam(name = "file") MultipartFile newModel, @CurrentUser PortalUser currentUser) {
        return addNewModel(FileType.CAEFile, optimizationId, newModel, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CAEFiles", method = RequestMethod.POST, params = {"entityId"})
    public FileEntity addLinkToParentCAEModel(@PathVariable("optimizationId") Integer optimizationId,
                                              @RequestParam(name = "entityId") Integer existedFileId,
                                              @CurrentUser PortalUser currentUser) {
        return addLinkToParentModel(FileType.CAEFile, optimizationId, existedFileId, currentUser.getUser());
    }

    @RequestMapping(value = "/{optimizationId}/CAEFiles", method = RequestMethod.GET)
    public List<FileEntity> getAllCAEFiles(@PathVariable("optimizationId") Integer optimizationId,
                                           @CurrentUser PortalUser currentUser) {
        return getAllFiles(FileType.CAEFile, optimizationId, currentUser.getUser());
    }

    @RequestMapping(value = "/{optimizationId}/CAEFiles/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getCAEFilesPage(@PathVariable("optimizationId") Integer optimizationId,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) {
        return getFilesPage(FileType.CAEFile, optimizationId, dataRequest, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CAEFiles/{id}", method = RequestMethod.DELETE)
    public void deleteCAEFile(@PathVariable("optimizationId") Integer optimizationId,
                              @PathVariable("id") Integer fileId,
                              @CurrentUser PortalUser currentUser) {
        deleteFile(optimizationId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAEFiles", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteCAEFiles(@PathVariable("id") int optimizationId,
                               @RequestBody List<Integer> entitiesToDelete, @CurrentUser PortalUser currentUser) {
        entitiesToDelete.forEach(id -> deleteCAEFile(optimizationId, id, currentUser));
    }


    /**
     * CAD FILES
     **/
    @RequestMapping(value = "/{optimizationId}/CADFiles/available/suggest", method = RequestMethod.POST)
    public List<FileEntity> suggestAvailableCADFiles(@PathVariable("optimizationId") int optimizationId,
                                                     @RequestBody @Valid SuggestRequest suggestRequest, @CurrentUser PortalUser currentUser) {
        return suggestAvailableFiles(FileType.CADFile, optimizationId, suggestRequest, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CADFiles", method = RequestMethod.POST)
    public FileEntity addNewCADModel(@PathVariable("optimizationId") Integer optimizationId,
                                     @RequestParam(name = "file") MultipartFile newModel, @CurrentUser PortalUser currentUser) {
        return addNewModel(FileType.CADFile, optimizationId, newModel, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CADFiles", method = RequestMethod.POST, params = {"entityId"})
    public FileEntity addLinkToParentCADModel(@PathVariable("optimizationId") Integer optimizationId,
                                              @RequestParam(name = "entityId") Integer existedFileId,
                                              @CurrentUser PortalUser currentUser) {
        return addLinkToParentModel(FileType.CADFile, optimizationId, existedFileId, currentUser.getUser());
    }

    @RequestMapping(value = "/{optimizationId}/CADFiles", method = RequestMethod.GET)
    public List<FileEntity> getAllCADFiles(@PathVariable("optimizationId") Integer optimizationId,
                                           @CurrentUser PortalUser currentUser) {
        return getAllFiles(FileType.CADFile, optimizationId, currentUser.getUser());
    }

    @RequestMapping(value = "/{optimizationId}/CADFiles/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getCADFilesPage(@PathVariable("optimizationId") Integer optimizationId,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) {
        return getFilesPage(FileType.CADFile, optimizationId, dataRequest, currentUser.getUser());
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{optimizationId}/CADFiles/{id}", method = RequestMethod.DELETE)
    public void deleteCADFile(@PathVariable("optimizationId") Integer optimizationId,
                              @PathVariable("id") Integer fileId,
                              @CurrentUser PortalUser currentUser) {
        deleteFile(optimizationId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CADFiles", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteCADFiles(@PathVariable("id") int optimizationId,
                               @RequestBody List<Integer> entitiesToDelete, @CurrentUser PortalUser currentUser) {
        entitiesToDelete.forEach(id -> deleteCADFile(optimizationId, id, currentUser));
    }


    private List<FileEntity> suggestAvailableFiles(FileType fileType, int optimizationId, SuggestRequest suggestRequest, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.update, currentUser);
        List<Integer> excludedFiles = optimizationManager.getOptimizationFiles(optimization, currentUser).stream()
                .filter(file -> file.getFileTypeId() == fileType.getId())
                .map(FileEntity::getId)
                .collect(toList());
        suggestRequest.setExcludedIds(excludedFiles);

        Integer parentCEId = optimization.getParentCEId();
        return constructionElementManager.suggestFiles(parentCEId, fileType, suggestRequest, currentUser);
    }

    private FileEntity addNewModel(FileType fileType, Integer optimizationId, MultipartFile newModel, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser);
        return optimizationManager.addModel(optimization, newModel, fileType, currentUser);
    }

    private FileEntity addLinkToParentModel(FileType fileType, Integer optimizationId, Integer existedFileId, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser);
        return optimizationManager.addLinkToCEFile(optimization, existedFileId, fileType, currentUser);
    }

    private List<FileEntity> getAllFiles(FileType fileType, Integer optimizationId, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser);
        return optimizationManager.getOptimizationFiles(optimization, currentUser).stream()
                .filter(file -> file.getFileTypeId() == fileType.getId())
                .collect(toList());
    }

    private PageingResponse<FileEntity> getFilesPage(FileType fileType, Integer optimizationId, DataRequest dataRequest, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser);
        dataRequest.getFilters().put("fileTypeId", fileType.getId());
        Pageable pageRequest = dataRequest.getPageable().toSpringPageable();
        PageImpl<FileEntity> page = optimizationManager.getOptimizationFilesPage(optimization, dataRequest.getFilters(),
                pageRequest, currentUser);
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageRequest, page.getTotalElements());
    }

    private void deleteFile(Integer optimizationId, Integer fileId, UserEntity currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser);
        FileEntity fileToDelete = fileManager.findOne(fileId, currentUser);
        if (fileToDelete == null) {
            throw new NotFoundException(fileId, FileEntity.class);
        }
        optimizationManager.deleteModel(optimization, fileToDelete, currentUser);
    }
}
