package ru.crp.lvms.optimizations.base.controllers.files;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import java.util.List;

import static ru.crp.lvms.utils.URLs.OPTIMIZATION_FILES_URL;
import static ru.crp.lvms.web.controllers.misc.CrudController.deleteSeveralEntities;

@RestController
@RequestMapping(value = OPTIMIZATION_FILES_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptimizationFilesController {
    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private MFOptimizationSettingsManager MFOptimizationSettingsManager;

    @Autowired
    private FileManager fileManager;

    @RequestMapping(method = RequestMethod.GET)
    public List<FileEntity> getAll(@PathVariable("optimizationId") Integer optimizationId,
                                   @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        return optimizationManager.getOptimizationFiles(optimization, currentUser.getUser());
    }


    @RequestMapping(value = "/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getPage(@PathVariable("optimizationId") Integer optimizationId,
                                               @RequestBody DataRequest dataRequest,
                                               @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        Pageable pageRequest = dataRequest.getPageable().toSpringPageable();
        PageImpl<FileEntity> page = optimizationManager.getOptimizationFilesPage(optimization, dataRequest.getFilters(),
                pageRequest, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageRequest, page.getTotalElements());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/replace", method = RequestMethod.POST)
    public List<FileEntity> replaceFiles(@PathVariable("optimizationId") Integer optimizationId,
                                         @RequestBody List<Integer> fileIds,
                                         @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        optimizationManager.replaceFiles(optimization, fileIds, currentUser.getUser());
        return optimizationManager.getOptimizationFiles(optimization, currentUser.getUser());
    }


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/ModFrontierScript", method = RequestMethod.POST)
    public FileEntity replaceScript(@PathVariable("optimizationId") Integer optimizationId,
                                    @RequestPart("file") MultipartFile newModel, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        List<FileEntity> scripts = optimizationManager.getOptimizationFiles(optimization, FileType.ModFrontierScript, currentUser.getUser());
        scripts.forEach(script -> optimizationManager.deleteModel(optimization, script, currentUser.getUser()));
        MFOptimizationSettingsManager.unselectVaryingParams(optimizationId);
        return optimizationManager.addModel(optimization, newModel, FileType.ModFrontierScript, currentUser.getUser());
    }


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public void deleteFile(@PathVariable("optimizationId") Integer optimizationId,
                           @PathVariable("id") Integer fileId,
                           @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        FileEntity fileToDelete = fileManager.findOne(fileId, currentUser.getUser());
        if (fileToDelete == null) {
            throw new NotFoundException(fileId, FileEntity.class);
        }
        optimizationManager.deleteModel(optimization, fileToDelete, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveralFiles(@PathVariable("optimizationId") Integer optimizationId,
                                   @RequestParam("ids") List<Integer> fileIds,
                                   @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, fileIds, (id) -> deleteFile(optimizationId, id, currentUser));
    }
}
