package ru.crp.lvms.optimizations.base.dao;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.base.enums.JsonSettingsType;
import ru.crp.lvms.optimizations.mf.dao.MFOptimizationBaseSettingsDao;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.mf.managers.MFDefaultParamsCreator;
import ru.crp.lvms.optimizations.optimus.dao.OptimusOptimizationBaseSettingsDao;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.io.IOException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static ru.crp.lvms.optimizations.mf.managers.MFDefaultParamsCreator.MOGA_METHOD_ID;
import static ru.crp.lvms.optimizations.mf.managers.MFDefaultParamsCreator.SIMPLEX_METHOD_ID;

@Repository
@Transactional(propagation = Propagation.REQUIRED)
public class OptimizationJsonSettingsDao extends BaseSpaceDao<OptimizationJsonSetting> {
    @Autowired
    private MFOptimizationBaseSettingsDao MFOptimizationBaseSettingsDao;

    @Autowired
    private OptimusOptimizationBaseSettingsDao optimusOptimizationBaseSettingsDao;

    private Map<Integer, Integer> defaultDoeNumPoints;

    @PostConstruct
    public void init() {
        defaultDoeNumPoints = new HashMap<>();
        defaultDoeNumPoints.put(MFDefaultParamsCreator.RANDOM_DOE_ID, 9);
        defaultDoeNumPoints.put(MFDefaultParamsCreator.SOBOL_DOE_ID, 11);
        defaultDoeNumPoints.put(MFDefaultParamsCreator.ULH_DOE_ID, 10);
    }

    private static class JsonSettingsRowMapper implements RowMapper<OptimizationJsonSetting> {
        private ObjectMapper jacksonMapper;

        public JsonSettingsRowMapper(ObjectMapper jacksonMapper) {
            this.jacksonMapper = jacksonMapper;
        }

        @Override
        public OptimizationJsonSetting mapRow(ResultSet rs, int rowNum) throws SQLException {
            try {
                OptimizationJsonSetting setting = new OptimizationJsonSetting();
                setting.setId(rs.getInt("id"));
                setting.setOptimizationId(rs.getInt("optimization_id"));
                setting.setParentId(rs.getInt("parent_id"));
                if (rs.wasNull()) {
                    setting.setParentId(null);
                }

                JsonNode jsonSettings = jacksonMapper.readValue(rs.getString("settings"), JsonNode.class);

                setting.setType(jacksonMapper.treeToValue(jsonSettings.get("type"), JsonSettingsType.class));

                setting.setName(jsonSettings.get("name").asText());
                if (jsonSettings.get("xmlName").isNull()) {
                    setting.setXmlName(null);
                } else {
                    setting.setXmlName(jsonSettings.get("xmlName").asText());
                }

                setting.setBoolValue(jsonSettings.get("boolValue").isNull() ? null : jsonSettings.get("boolValue").asBoolean());
                setting.setPlainValue(jsonSettings.get("plainValue").isNull() ? null : jsonSettings.get("plainValue").asText());
                List<OptimizationListSetting> listSettings = new ArrayList<>();
                jsonSettings.get("listValue").forEach(listEntry -> {
                    try {
                        OptimizationListSetting optimizationListSetting = jacksonMapper.treeToValue(listEntry, OptimizationListSetting.class);
                        listSettings.add(optimizationListSetting);
                    } catch (JsonProcessingException ex) {
                        throw new InternalException(ex);
                    }
                });
                setting.setListValue(listSettings);
                List<OptimizationJsonSetting> childJsonSettings = new ArrayList<>();
                JsonNode compositeValue = jsonSettings.get("compositeValue");
                if (compositeValue != null && !compositeValue.isNull()) {
                    compositeValue
                            .forEach(listEntry -> {
                                try {
                                    OptimizationJsonSetting optimizationListSetting = jacksonMapper.treeToValue(listEntry, OptimizationJsonSetting.class);
                                    childJsonSettings.add(optimizationListSetting);
                                } catch (JsonProcessingException ex) {
                                    throw new InternalException(ex);
                                }
                            });
                }
                setting.setCompositeValue(childJsonSettings);

                JsonNode defaultValue = jsonSettings.get("defaultValue");
                if (defaultValue != null) {
                    switch (setting.getType()) {
                        case plain:
                            setting.setDefaultValue(defaultValue.isNull() ? null : defaultValue.asText());
                            break;
                        case bool:
                            setting.setDefaultValue(defaultValue.isNull() ? null : defaultValue.asBoolean());
                            break;
                        case list:
                            setting.setDefaultValue(jacksonMapper.treeToValue(defaultValue, OptimizationListSetting.class));
                            break;
                        case composite:
                            setting.setDefaultValue(jacksonMapper.treeToValue(defaultValue, OptimizationJsonSetting.class));
                            break;
                        default:
                            throw new IllegalArgumentException(String.format("Unexpected json settings type '%s'", setting.getType()));
                    }
                }


                return setting;
            } catch (IOException ex) {
                throw new InternalException("Can't deserialize json", ex);
            }
        }
    }

    @Autowired
    private ObjectMapper jacksonMapper;

    public List<OptimizationJsonSetting> findRunSettings(int optimizationId) {
        return jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? AND parent_id IS NULL ORDER BY id",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId);
    }

    public List<OptimizationJsonSetting> findSettingsByParent(int optimizationId, int parentId) {
        return jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? AND parent_id = ? ORDER BY id",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId, parentId);
    }

    public void createSettings(OptimizationJsonSetting setting) {
        Map<String, Object> jsonSettings = extractPropertiesToSave(setting);
        try {
            jdbcTemplate().update("INSERT INTO json_settings(parent_id, optimization_id, settings) VALUES(?, ?, ?::jsonb)",
                    setting.getParentId(), setting.getOptimizationId(), jacksonMapper.writeValueAsString(jsonSettings));
        } catch (JsonProcessingException ex) {
            throw new InternalException("Can't serialize json", ex);
        }
    }

    private Map<String, Object> extractPropertiesToSave(OptimizationJsonSetting setting) {
        Map<String, Object> jsonSettings = new HashMap<>();
        jsonSettings.put("type", setting.getType());
        jsonSettings.put("name", setting.getName());
        jsonSettings.put("xmlName", setting.getXmlName());
        jsonSettings.put("boolValue", setting.getBoolValue());
        jsonSettings.put("plainValue", setting.getPlainValue());
        jsonSettings.put("listValue", setting.getListValue());
        jsonSettings.put("compositeValue", setting.getCompositeValue());
        jsonSettings.put("defaultValue", setting.getDefaultValue());
        return jsonSettings;
    }


    @Nullable
    public OptimizationJsonSetting findById(int optimizationId, int paramId) {
        List<OptimizationJsonSetting> params = jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? AND id = ?",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId, paramId);
        return params.isEmpty() ? null : params.get(0);
    }

    @Override
    public OptimizationJsonSetting updateAndRead(OptimizationJsonSetting setting) {
        Map<String, Object> jsonSettings = extractPropertiesToSave(setting);
        try {
            jdbcTemplate().update("UPDATE json_settings SET settings = ?::jsonb WHERE id = ?",
                    jacksonMapper.writeValueAsString(jsonSettings), setting.getId());
        } catch (JsonProcessingException ex) {
            throw new InternalException("Can't serialize json", ex);
        }
        return findById(setting.getOptimizationId(), setting.getId());
    }

    public void resetDefaultDoeParams(int optimizationId, int selectedAlgorithmId) {
        Integer numTargetFuncs = MFOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, MFBaseSettingType.NUM_TARGET_FUNCS.getId()).getCurrentValue();
        Integer numVaryingParams = MFOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, MFBaseSettingType.NUM_CAO_VARYING_PARAMS.getId()).getCurrentValue();

        final int numPoints;
        if (selectedAlgorithmId == MOGA_METHOD_ID) {
            numPoints = 2 * numTargetFuncs * numVaryingParams;
        } else if (selectedAlgorithmId == SIMPLEX_METHOD_ID) {
            numPoints = numVaryingParams + 1;
        } else {
            numPoints = 0;
        }

        List<OptimizationJsonSetting> paramsToUpdate = jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? AND settings->>'xmlName' IN" +
                        "('doerand.parameter.num.rows', 'sobol.parameter.num.rows', 'ulh.parameter.designs')",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId);
        paramsToUpdate.forEach(p -> {
            if (selectedAlgorithmId != MOGA_METHOD_ID && selectedAlgorithmId != SIMPLEX_METHOD_ID) {
                p.setPlainValue("" + defaultDoeNumPoints.get(p.getParentId()));
            } else {
                p.setPlainValue("" + numPoints);
            }
            updateAndRead(p);
        });
    }

    public void resetOptimusDefaultDoeParams(int optimizationId, int selectedAlgorithmId) {
        Integer numTargetFuncs = optimusOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, OptimusBaseSettingType.NUM_TARGET_FUNCS.getId()).getCurrentValue();
        Integer numVaryingParams = optimusOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, OptimusBaseSettingType.NUM_CAO_VARYING_PARAMS.getId()).getCurrentValue();

        final int numPoints;
        if (selectedAlgorithmId == MOGA_METHOD_ID) {
            numPoints = 2 * numTargetFuncs * numVaryingParams;
        } else if (selectedAlgorithmId == SIMPLEX_METHOD_ID) {
            numPoints = numVaryingParams + 1;
        } else {
            numPoints = 0;
        }

        List<OptimizationJsonSetting> paramsToUpdate = jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? AND settings->>'xmlName' IN" +
                        "('doerand.parameter.num.rows', 'sobol.parameter.num.rows', 'ulh.parameter.designs')",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId);
        paramsToUpdate.forEach(p -> {
            if (selectedAlgorithmId != MOGA_METHOD_ID && selectedAlgorithmId != SIMPLEX_METHOD_ID) {
                p.setPlainValue("" + defaultDoeNumPoints.get(p.getParentId()));
            } else {
                p.setPlainValue("" + numPoints);
            }
            updateAndRead(p);
        });
    }

    @Nullable
    public List<OptimizationJsonSetting> findAllOptimizationSettings(int optimizationId) {
        List<OptimizationJsonSetting> params = jdbcTemplate().query("SELECT * FROM json_settings WHERE optimization_id = ? ORDER BY id",
                new JsonSettingsRowMapper(jacksonMapper), optimizationId);
        return params;
    }

}
