package ru.crp.lvms.optimizations.mf.config;

import org.modelmapper.TypeToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.AlgorithmType;
import ru.crp.lvms.model.entities.modfrontier.DoeType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.mf.config.xml.*;
import ru.crp.lvms.optimizations.mf.dao.MFOptimizationCADVaryingParamsDao;
import ru.crp.lvms.optimizations.mf.dao.MFOptimizationConstraintsDao;
import ru.crp.lvms.optimizations.mf.dao.MFOptimizationTargetFuncsDao;
import ru.crp.lvms.optimizations.mf.dao.MFOptimizationVaryingParamsDao;
import ru.crp.lvms.optimizations.mf.entities.*;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.utils.Mapper;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.Mapper.map;
import static ru.crp.lvms.utils.Utils.mapWithIndex;

@Component
public class MFConfigGenerator {

    @Autowired
    private MFOptimizationVaryingParamsDao mfOptimizationVaryingParamsDao;

    @Autowired
    private MFOptimizationCADVaryingParamsDao mfOptimizationCADVaryingParamsDao;

    @Autowired
    private MFOptimizationConstraintsDao mfOptimizationConstraintsDao;

    @Autowired
    private MFOptimizationTargetFuncsDao mfOptimizationTargetFuncsDao;

    @Autowired
    private MFOptimizationSettingsManager mfOptimizationSettingsManager;

    @Autowired
    private OptimizationJsonSettingsDao jsonSettingsDao;

    public Path generateConfig(OptimizationEntity optimization) throws IOException, JAXBException {
        ConfigRoot configRoot = composeXmlStructure(optimization);

        Path tempConfigFile = Files.createTempFile(null, null);

        JAXBContext jaxbContext = JAXBContext.newInstance(ConfigRoot.class);
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        jaxbMarshaller.marshal(configRoot, tempConfigFile.toFile());

        return tempConfigFile;
    }

    private ConfigRoot composeXmlStructure(OptimizationEntity optimization) {
        ConfigRoot configRoot = new ConfigRoot();

        List<OptimizationJsonSetting> runSettings = jsonSettingsDao.findRunSettings(optimization.getId()).stream()
                // some settings (like 'timeout') should not be used in settings.xml but on solver
                .filter(s -> s.getXmlName() != null)
                .collect(toList());
        RunType runType = new RunType();
        PropertiesType properties = new PropertiesType();
        properties.setProperty(Mapper.map(runSettings, new TypeToken<List<PropertyType>>() {
        }));
        runType.setProperties(properties);
        configRoot.setRun(runType);

        Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings = mfOptimizationSettingsManager.findBaseSettings(optimization.getId(), null);
        configRoot.setDoeDB(formDoeDb(optimization, baseSettings));
        configRoot.setDesignsDB(formDesignDb(optimization, baseSettings));

        configRoot.setDoe(formDoe(optimization));
        configRoot.setScheduler(formScheduler(optimization));

        return configRoot;
    }

    private DoeDBType formDoeDb(OptimizationEntity optimization, Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings) {
        List<VariableType> variables = getVaryingParams(optimization, baseSettings);
        DoeDBType doeDBType = new DoeDBType();
        doeDBType.setVariables(variables);
        return doeDBType;
    }

    // returns used varying and cad varying params in single list
    private List<VariableType> getVaryingParams(OptimizationEntity optimization, Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings) {
        Integer varyingParamsCount = baseSettings.get(MFBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue();
        List<MFOptimizationVaryingParam> varyingParams = mfOptimizationVaryingParamsDao
                .findForOptimization(optimization.getId()).stream()
                .limit(varyingParamsCount)
                .collect(toList());

        List<VariableType> variables = mapWithIndex(varyingParams, 1, i -> {
            VariableType vt = map(varyingParams.get(i - 1), VariableType.class);
            vt.setName("par" + i);
            return vt;
        }).collect(toList());
        if (optimization.getSolverCadId() != null) {
            Integer cadVaryingParamsCount = baseSettings.get(MFBaseSettingType.NUM_CAD_VARYING_PARAMS).getCurrentValue();
            List<MFOptimizationCADVaryingParam> cadVaryingParams = mfOptimizationCADVaryingParamsDao
                    .findForOptimization(optimization.getId()).stream()
                    .limit(cadVaryingParamsCount)
                    .collect(toList());

            variables.addAll(mapWithIndex(cadVaryingParams, 1, i -> {
                VariableType vt = map(cadVaryingParams.get(i - 1), VariableType.class);
                vt.setName("gpar" + i);
                return vt;
            }).collect(toList()));
        }
        return variables;
    }


    private DesignsDBType formDesignDb(OptimizationEntity optimization, Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings) {
        List<VariableType> variables = getVaryingParams(optimization, baseSettings);
        Integer usedVaryingParamsCount = baseSettings.get(MFBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue();
        //Purely generated constants to fill variables up to the max value
        List<VariableType> constants = IntStream
                .rangeClosed(usedVaryingParamsCount + 1, MFBaseSettingType.NUM_CAO_VARYING_PARAMS.getMaxValue())
                .mapToObj(i -> {
                    VariableType var = new VariableType();
                    var.setName("par" + i);
                    var.setClazz("it.esteco.procint.design.DesignInputVariable");
                    var.setValueAttribute("0.0");
                    var.setKind("CONSTANT");
                    return var;
                })
                .collect(toList());
        if (optimization.getSolverCadId() != null) {
            Integer usedCADVaryingParamsCount = baseSettings.get(MFBaseSettingType.NUM_CAD_VARYING_PARAMS).getCurrentValue();
            constants.addAll(IntStream
                    .rangeClosed(usedCADVaryingParamsCount + 1, MFBaseSettingType.NUM_CAO_VARYING_PARAMS.getMaxValue())
                    .mapToObj(i -> {
                        VariableType var = new VariableType();
                        var.setName("gpar" + i);
                        var.setClazz("it.esteco.procint.design.DesignInputVariable");
                        var.setValueAttribute("0.0");
                        var.setKind("CONSTANT");
                        return var;
                    })
                    .collect(toList()));
        }

        //Purely generated outputs
        List<VariableType> outputs = IntStream
                .rangeClosed(1, MFBaseSettingType.NUM_OUTPUT_PARAMS.getMaxValue())
                .mapToObj(i -> {
                    VariableType var = new VariableType();
                    var.setClazz("it.esteco.procint.design.DesignOutputVariable");
                    var.setName("out" + i);
                    return var;
                })
                .collect(toList());

        // target functions and constraints must have sequentially numbered 'expression' fields
        Integer targetFuncsCount = baseSettings.get(MFBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue();
        List<MFOptimizationTargetFunc> targetFuncs = mfOptimizationTargetFuncsDao.findForOptimization(optimization.getId()).stream().limit(targetFuncsCount).collect(toList());
        List<VariableType> targetFunctionsWithoutExpression = mapWithIndex(targetFuncs, 1, i -> {
            VariableType var = map(targetFuncs.get(i - 1), VariableType.class);
            var.setName("obj" + i);
            return var;
        }).collect(toList());
        List<VariableType> targetFunctionsWithExpressions = mapWithIndex(targetFunctionsWithoutExpression, 1, i -> {
            VariableType var = targetFunctionsWithoutExpression.get(i - 1);
            var.setExpression("out" + i);
            return var;
        }).collect(toList());

        Integer constraintsCount = baseSettings.get(MFBaseSettingType.NUM_CONSTRAINTS).getCurrentValue();
        List<MFOptimizationConstraint> optimizationConstrains = mfOptimizationConstraintsDao.findForOptimization(optimization.getId()).stream().limit(constraintsCount).collect(toList());

        List<VariableType> constraintsWithoutExpression = mapWithIndex(optimizationConstrains, 1, i -> {
            VariableType var = map(optimizationConstrains.get(i - 1), VariableType.class);
            var.setName("c" + i);
            return var;
        }).collect(toList());

        List<VariableType> constraintsWithExpressions = mapWithIndex(constraintsWithoutExpression, 1, i -> {
            VariableType var = constraintsWithoutExpression.get(i - 1);
            var.setExpression("out" + (i + targetFunctionsWithExpressions.size()));
            return var;
        }).collect(toList());

        DesignsDBType designsDBType = new DesignsDBType();
        designsDBType.setVariables(asList(variables, constants, outputs, targetFunctionsWithExpressions, constraintsWithExpressions).stream()
                .flatMap(List::stream).collect(toList()));
        return designsDBType;
    }

    private DOEType formDoe(OptimizationEntity optimization) {
        DOEType doeType = new DOEType();
        Integer selectedDoeId = mfOptimizationSettingsManager.findSelectedDoeId(optimization.getId());
        ParametersDictionary selectedDoe = mfOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.DoEType, selectedDoeId);
        DoeType doe = DoeType.map(selectedDoe);
        PlugInType plugInType = PlugInType.create(doe);
        List<OptimizationJsonSetting> doeParams = jsonSettingsDao.findSettingsByParent(optimization.getId(), selectedDoeId).stream()
                // some settings (like 'timeout') should not be used in settings.xml but on solver
                .filter(s -> s.getXmlName() != null)
                .collect(toList());
        plugInType.setProperty(Mapper.map(doeParams, new TypeToken<List<PropertyType>>() {
        }));
        if (doe == DoeType.ULH) {
            PropertyType property = new PropertyType("ulh.parameter.multithreading.policy", "ulh.parameter.multithreading.policy.maximum");
            property.setKeys("[ulh.parameter.multithreading.policy.maximum,ulh.parameter.multithreading.policy.userdef]");
            property.setOptions("[Use Maximum Number of Available Processors,User Defined]");
            plugInType.getProperty().add(property);
            plugInType.getProperty().add(new PropertyType("ulh.parameter.threads.number", "NUM_OF_THREADS"));
        }
        doeType.setPlugin(plugInType);
        return doeType;
    }

    private SchedulerType formScheduler(OptimizationEntity optimization) {
        SchedulerType schedulerType = new SchedulerType();
        Integer selectedAlgorithmId = mfOptimizationSettingsManager.findSelectedAlgorithmId(optimization.getId());
        ParametersDictionary selectedAlgorithm = mfOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.OptimizationMethod, selectedAlgorithmId);
        PlugInType plugInType = PlugInType.create(AlgorithmType.map(selectedAlgorithm));
        List<OptimizationJsonSetting> algorithmParams = jsonSettingsDao.findSettingsByParent(optimization.getId(), selectedAlgorithmId).stream()
                // some settings (like 'timeout') should not be used in settings.xml but on solver
                .filter(s -> s.getXmlName() != null)
                .collect(toList());
        plugInType.setProperty(Mapper.map(algorithmParams, new TypeToken<List<PropertyType>>() {
        }));
        schedulerType.setPlugin(plugInType);
        return schedulerType;
    }
}
