package ru.crp.lvms.optimizations.mf.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.dao.ParametersDictionaryDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.AlgorithmType;
import ru.crp.lvms.model.entities.modfrontier.DoeType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.dao.*;
import ru.crp.lvms.optimizations.mf.entities.*;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.mf.enums.MFSettingsType;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettignsDao;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettingsManager;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toMap;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class MFOptimizationSettingsManager extends AbstractOptimizationSettingsManager<MFSettingsType> {
    @Autowired
    private MFOptimizationConstraintsDao MFOptimizationConstraintsDao;

    @Autowired
    private MFOptimizationTargetFuncsDao MFOptimizationTargetFuncsDao;

    @Autowired
    private MFOptimizationBaseSettingsDao MFOptimizationBaseSettingsDao;

    @Autowired
    private MFOptimizationVaryingParamsDao MFOptimizationVaryingParamsDao;

    @Autowired
    private MFOptimizationCADVaryingParamsDao mfOptimizationCADVaryingParamsDao;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ParametersDictionaryDao parametersDictionaryDao;

    @Autowired
    private OptimizationDao optimizationDao;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    private Map<MFSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> settingsDaos;

    @Override
    protected Map<MFSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> getSettingsDaos() {
        Map<MFSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>>
                settingsDaos = new EnumMap<>(MFSettingsType.class);
        settingsDaos.put(MFSettingsType.Constraint, MFOptimizationConstraintsDao);
        settingsDaos.put(MFSettingsType.TargetFunc, MFOptimizationTargetFuncsDao);
        settingsDaos.put(MFSettingsType.BaseSetting, MFOptimizationBaseSettingsDao);
        settingsDaos.put(MFSettingsType.VaryingParam, MFOptimizationVaryingParamsDao);
        settingsDaos.put(MFSettingsType.CADVaryingParam, mfOptimizationCADVaryingParamsDao);
        return settingsDaos;
    }

    // key - algorithm id, value - default doe id for this algorithm
    private Map<Integer, Integer> algorithmDefaultDoes;

    @PostConstruct
    @Override
    public void setup() {
        super.setup();
        algorithmDefaultDoes = new HashMap<>();
        algorithmDefaultDoes.put(AlgorithmType.SIMPLEX.getId(), DoeType.ULH.getId());
        algorithmDefaultDoes.put(AlgorithmType.MOGA_II.getId(), DoeType.ULH.getId());
        algorithmDefaultDoes.put(AlgorithmType.DOESequence.getId(), DoeType.Random.getId());

    }

    public Map<MFBaseSettingType, MFOptimizationBaseSetting> findBaseSettings(int optimizationId,
                                                                              UserEntity currentUser) {
        Map<MFBaseSettingType, MFOptimizationBaseSetting> settingsByType = MFOptimizationBaseSettingsDao
                .findForOptimization(optimizationId).stream()
                .collect(toMap(s -> MFBaseSettingType.valueOf(s.getTypeId()), Function.identity()));
        Stream.of(MFBaseSettingType.values()).forEach(type -> settingsByType.putIfAbsent(type, null));
        return settingsByType;
    }


    public void selectDefaultAlgorithm(int optimizationId, Integer numTargetFuncs) {
        if (numTargetFuncs == null) {
            throw new InternalException("Initial number of target functions must be set");
        }
        int selectedAlgorithmId;
        if (numTargetFuncs == 1) {
            selectedAlgorithmId = AlgorithmType.SIMPLEX.getId();
        } else {
            selectedAlgorithmId = AlgorithmType.MOGA_II.getId();
        }
        selectAlgorithm(optimizationId, selectedAlgorithmId);
        optimizationJsonSettingsDao.resetDefaultDoeParams(optimizationId, selectedAlgorithmId);
    }


    public MFOptimizationBaseSetting getBaseSettingsByType(int optimizationId, MFBaseSettingType type) {
        MFOptimizationBaseSetting baseSetting = MFOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, type.getId());
        if (baseSetting == null) {
            throw new BadRequestException("Базовые настройки с типом '" + type.getDisplayName() + "' для оптимизации " + optimizationId + " не найдены");
        }
        return baseSetting;
    }

    public void unselectVaryingParams(Integer optimizationId) {
        MFOptimizationVaryingParamsDao.unselectParams(optimizationId);
    }

    public void selectVaryingParam(Integer optimizationId, Integer paramId) {
        MFOptimizationVaryingParamsDao.selectVaryingParam(optimizationId, paramId);
    }

    public List<ParametersDictionary> findByTypeName(ParametersDictionaryType parametersDictionaryType) {
        return parametersDictionaryDao.findByTypeName(parametersDictionaryType);
    }

    @Nullable
    public ParametersDictionary findByTypeNameAndId(ParametersDictionaryType parametersDictionaryType, int id) {
        return parametersDictionaryDao.findByTypeNameAndId(parametersDictionaryType, id);
    }

    @Nullable
    public Integer findSelectedAlgorithmId(int optimizationId) {
        return optimizationDao.findSelectedAlgorithmId(optimizationId);
    }

    @Nullable
    public Integer findSelectedDoeId(int optimizationId) {
        return optimizationDao.findSelectedDoeId(optimizationId);
    }


    public void selectDoe(int optimizationId, @Nullable Integer selectedDoeId) {
        optimizationDao.selectDoe(optimizationId, selectedDoeId);
    }

    public void selectAlgorithm(int optimizationId, @Nullable Integer selectedAlgorithmId) {
        optimizationDao.selectAlgorithm(optimizationId, selectedAlgorithmId);
        if (selectedAlgorithmId != null) {
            Integer selectedDoeId = algorithmDefaultDoes.get(selectedAlgorithmId);
            selectDoe(optimizationId, selectedDoeId);
        } else {
            selectDoe(optimizationId, null);
        }
    }

    public void copySettings(int referenceOptimizationId, int newOptimizationId) {
        List<MFOptimizationBaseSetting> mfBaseSettings=MFOptimizationBaseSettingsDao.findForOptimization(referenceOptimizationId);
        mfBaseSettings.forEach(tbs ->{
            tbs.setOptimizationId(newOptimizationId);
            MFOptimizationBaseSettingsDao.insert(tbs);
        });
        List<MFOptimizationConstraint> MFConstraints=MFOptimizationConstraintsDao.findForOptimization(referenceOptimizationId);
        MFConstraints.forEach(tc ->{
            tc.setOptimizationId(newOptimizationId);
            MFOptimizationConstraintsDao.insert(tc);
        });
        List<MFOptimizationTargetFunc> MFTargetFunctions=MFOptimizationTargetFuncsDao.findForOptimization(referenceOptimizationId);
        MFTargetFunctions.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            MFOptimizationTargetFuncsDao.insert(ttf);
        });
        List<MFOptimizationVaryingParam> MFVaryingParams=MFOptimizationVaryingParamsDao.findForOptimization(referenceOptimizationId);
        MFVaryingParams.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            MFOptimizationVaryingParam newParam = MFOptimizationVaryingParamsDao.insertAndRead(ttf);
            if(ttf.getIsUsedInScript()==true) {
                selectVaryingParam(newOptimizationId, newParam.getId());
            }
            ;
        });
        if (optimizationManager.findOne(referenceOptimizationId,null).getSolverCadId() != null) {
            List<MFOptimizationCADVaryingParam> MFCadVaryingParams = mfOptimizationCADVaryingParamsDao.findForOptimization(referenceOptimizationId);
            MFCadVaryingParams.forEach(ttf -> {
                ttf.setOptimizationId(newOptimizationId);
                mfOptimizationCADVaryingParamsDao.insert(ttf);
            });
        }
    }
}
