package ru.crp.lvms.optimizations.optimus.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.dao.ParametersDictionaryDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.AlgorithmType;
import ru.crp.lvms.model.entities.modfrontier.DoeType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettignsDao;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optimus.dao.*;
import ru.crp.lvms.optimizations.optimus.entities.*;
import ru.crp.lvms.optimizations.optimus.enums.OptimusAlgorithmType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusDoeType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusSettingsType;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toMap;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class OptimusOptimizationSettingsManager extends AbstractOptimizationSettingsManager<OptimusSettingsType> {
    @Autowired
    private OptimusOptimizationConstraintsDao optimusOptimizationConstraintsDao;

    @Autowired
    private OptimusOptimizationTargetFuncsDao optimusOptimizationTargetFuncsDao;

    @Autowired
    private OptimusOptimizationBaseSettingsDao optimusOptimizationBaseSettingsDao;

    @Autowired
    private OptimusOptimizationVaryingParamsDao optimusOptimizationVaryingParamsDao;

    @Autowired
    private OptimusOptimizationCADVaryingParamsDao optimusOptimizationCADVaryingParamsDao;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ParametersDictionaryDao parametersDictionaryDao;

    @Autowired
    private OptimizationDao optimizationDao;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    private Map<OptimusSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> settingsDaos;

    @Override
    protected Map<OptimusSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>> getSettingsDaos() {
        Map<OptimusSettingsType, AbstractOptimizationSettignsDao<? extends AbstractOptimizationSettings>>
                settingsDaos = new EnumMap<>(OptimusSettingsType.class);
        settingsDaos.put(OptimusSettingsType.Constraint, optimusOptimizationConstraintsDao);
        settingsDaos.put(OptimusSettingsType.TargetFunc, optimusOptimizationTargetFuncsDao);
        settingsDaos.put(OptimusSettingsType.BaseSetting, optimusOptimizationBaseSettingsDao);
        settingsDaos.put(OptimusSettingsType.VaryingParam, optimusOptimizationVaryingParamsDao);
        settingsDaos.put(OptimusSettingsType.CADVaryingParam, optimusOptimizationCADVaryingParamsDao);
        return settingsDaos;
    }

    // key - algorithm id, value - default doe id for this algorithm
    private Map<Integer, Integer> algorithmDefaultDoes;

    @PostConstruct
    @Override
    public void setup() {
        super.setup();
        algorithmDefaultDoes = new HashMap<>();
        algorithmDefaultDoes.put(OptimusAlgorithmType.EGO.getId(), OptimusDoeType.RANDOM.getId());
        algorithmDefaultDoes.put(OptimusAlgorithmType.SIMAN.getId(), OptimusDoeType.RANDOM.getId());
        algorithmDefaultDoes.put(OptimusAlgorithmType.DOE.getId(), OptimusDoeType.RANDOM.getId());

    }

    public Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> findBaseSettings(int optimizationId,
                                                                                        UserEntity currentUser) {
        Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> settingsByType = optimusOptimizationBaseSettingsDao
                .findForOptimization(optimizationId).stream()
                .collect(toMap(s -> OptimusBaseSettingType.valueOf(s.getTypeId()), Function.identity()));
        Stream.of(OptimusBaseSettingType.values()).forEach(type -> settingsByType.putIfAbsent(type, null));
        return settingsByType;
    }


    public void selectDefaultAlgorithm(int optimizationId, Integer numTargetFuncs) {
        if (numTargetFuncs == null) {
            throw new InternalException("Initial number of target functions must be set");
        }
        int selectedAlgorithmId;
        if (numTargetFuncs == 1) {
            selectedAlgorithmId = OptimusAlgorithmType.EGO.getId();
        } else {
            selectedAlgorithmId = OptimusAlgorithmType.SIMAN.getId();
        }
        selectAlgorithm(optimizationId, selectedAlgorithmId);
        optimizationJsonSettingsDao.resetOptimusDefaultDoeParams(optimizationId, selectedAlgorithmId);
    }


    public OptimusOptimizationBaseSetting getBaseSettingsByType(int optimizationId, OptimusBaseSettingType type) {
        OptimusOptimizationBaseSetting baseSetting = optimusOptimizationBaseSettingsDao.getBaseSettingsByType(optimizationId, type.getId());
        if (baseSetting == null) {
            throw new BadRequestException("Базовые настройки с типом '" + type.getDisplayName() + "' для оптимизации " + optimizationId + " не найдены");
        }
        return baseSetting;
    }

    public void unselectVaryingParams(Integer optimizationId) {
        optimusOptimizationVaryingParamsDao.unselectParams(optimizationId);
    }

    public void selectVaryingParam(Integer optimizationId, Integer paramId) {
        optimusOptimizationVaryingParamsDao.selectVaryingParam(optimizationId, paramId);
    }

    public List<ParametersDictionary> findByTypeName(ParametersDictionaryType parametersDictionaryType) {
        return parametersDictionaryDao.findByTypeName(parametersDictionaryType);
    }

    @Nullable
    public ParametersDictionary findByTypeNameAndId(ParametersDictionaryType parametersDictionaryType, int id) {
        return parametersDictionaryDao.findByTypeNameAndId(parametersDictionaryType, id);
    }

    @Nullable
    public Integer findSelectedAlgorithmId(int optimizationId) {
        return optimizationDao.findSelectedAlgorithmId(optimizationId);
    }

    @Nullable
    public Integer findSelectedDoeId(int optimizationId) {
        return optimizationDao.findSelectedDoeId(optimizationId);
    }


    public void selectDoe(int optimizationId, @Nullable Integer selectedDoeId) {
        optimizationDao.selectDoe(optimizationId, selectedDoeId);
    }

    public void selectAlgorithm(int optimizationId, @Nullable Integer selectedAlgorithmId) {
        optimizationDao.selectAlgorithm(optimizationId, selectedAlgorithmId);
        if (selectedAlgorithmId != null) {
            Integer selectedDoeId = algorithmDefaultDoes.get(selectedAlgorithmId);
            selectDoe(optimizationId, selectedDoeId);
        } else {
            selectDoe(optimizationId, null);
        }
    }

    public void copySettings(int referenceOptimizationId, int newOptimizationId) {
        List<OptimusOptimizationBaseSetting> optimusBaseSettings=optimusOptimizationBaseSettingsDao.findForOptimization(referenceOptimizationId);
        optimusBaseSettings.forEach(tbs ->{
            tbs.setOptimizationId(newOptimizationId);
            optimusOptimizationBaseSettingsDao.insert(tbs);
        });
        List<OptimusOptimizationConstraint> optimusConstraints=optimusOptimizationConstraintsDao.findForOptimization(referenceOptimizationId);
        optimusConstraints.forEach(tc ->{
            tc.setOptimizationId(newOptimizationId);
            optimusOptimizationConstraintsDao.insert(tc);
        });
        List<OptimusOptimizationTargetFunc> optimusTargetFunctions=optimusOptimizationTargetFuncsDao.findForOptimization(referenceOptimizationId);
        optimusTargetFunctions.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            optimusOptimizationTargetFuncsDao.insert(ttf);
        });
        List<OptimusOptimizationVaryingParam> optimusVaryingParams=optimusOptimizationVaryingParamsDao.findForOptimization(referenceOptimizationId);
        optimusVaryingParams.forEach(ttf ->{
            ttf.setOptimizationId(newOptimizationId);
            OptimusOptimizationVaryingParam newParam = optimusOptimizationVaryingParamsDao.insertAndRead(ttf);
            if(ttf.getIsUsedInScript()==true) {
                selectVaryingParam(newOptimizationId, newParam.getId());
            };
            ;
        });
        if (optimizationManager.findOne(referenceOptimizationId,null).getSolverCadId() != null) {
            List<OptimusOptimizationCADVaryingParam> optimusCadVaryingParams = optimusOptimizationCADVaryingParamsDao.findForOptimization(referenceOptimizationId);
            optimusCadVaryingParams.forEach(ttf -> {
                ttf.setOptimizationId(newOptimizationId);
                optimusOptimizationCADVaryingParamsDao.insert(ttf);
            });
        }
    }
}
