package ru.crp.lvms.optimizations.optimus.results;

import ru.crp.lvms.chart.*;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFVaryingParamInterval;
import ru.crp.lvms.optimizations.mf.enums.MFTargetFuncsType;
import ru.crp.lvms.optimizations.mf.results.ModFrontierParamsRow;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationConstraint;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationTargetFunc;
import ru.crp.lvms.optimizations.optimus.entities.OptimusVaryingParamInterval;
import ru.crp.lvms.optimizations.optimus.enums.OptimusTargetFuncsType;
import ru.crp.lvms.optimizations.optimus.results.OptimusParamsRow;

import java.util.*;
import java.util.stream.IntStream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;

public class OptimusParams {
        private static final int NUM_UTIL_FIELDS = 8;

    private Map<Integer, Integer> varyingParamsIndices = new HashMap<>();
    private Map<Integer, Integer> outputParamsIndices = new HashMap<>();
    private Map<Integer, Integer> targetFunctionsIndices = new HashMap<>();
    private Map<Integer, Integer> constraintsIndices = new HashMap<>();
    private List<OptimusParamsRow> rows = new ArrayList<>();

    public OptimusParams(String[] headerRow) {
        for (String header : headerRow) {
            if (header.startsWith("par")) {
                varyingParamsIndices.put(getZeroBasedIndexFromHeader(header, "par"), varyingParamsIndices.size());
            } else if (header.startsWith("out")) {
                outputParamsIndices.put(getZeroBasedIndexFromHeader(header, "out"), outputParamsIndices.size());
            } else if (header.startsWith("obj")) {
                targetFunctionsIndices.put(getZeroBasedIndexFromHeader(header, "obj"), targetFunctionsIndices.size());
            } else if (header.startsWith("c")) {
                constraintsIndices.put(getZeroBasedIndexFromHeader(header, "c"), constraintsIndices.size());
            }
        }
    }

    private int getZeroBasedIndexFromHeader(String header, String prefix) {
        String indexStr = header.substring(prefix.length());
        return Integer.parseInt(indexStr) - 1;
    }

    public void addResultRow(String[] resultRow, Boolean optTypeKey) {
        int expectedColumnsCount = getExpectedColumnsCount();
        if (optTypeKey && resultRow.length != expectedColumnsCount) {
            throw new IllegalArgumentException("Expected columns count = " + expectedColumnsCount + ". Actual = " + resultRow.length);
        }
        if (!optTypeKey && (resultRow.length != (expectedColumnsCount-2))) {
            throw new IllegalArgumentException("Expected columns count = " + (expectedColumnsCount-2) + ". Actual = " + resultRow.length);
        }
        if (optTypeKey && resultRow[6].endsWith("3")) {
            rows.add(new OptimusParamsRow(resultRow, varyingParamsIndices, outputParamsIndices, targetFunctionsIndices, constraintsIndices,optTypeKey));
        }
        else if (!optTypeKey && (resultRow[5].endsWith("0")||resultRow[5].endsWith("2"))) {
            rows.add(new OptimusParamsRow(resultRow, varyingParamsIndices, outputParamsIndices, targetFunctionsIndices, constraintsIndices,optTypeKey));
        }
    }

    private int getExpectedColumnsCount() {
        return IntStream
                .of(varyingParamsIndices.size(), outputParamsIndices.size(), NUM_UTIL_FIELDS).sum();
    }

    @Override
    public String toString() {
        return "OptimusParams{" +
                "numConstraints=" + constraintsIndices.size() +
                ", numTargetFunctions=" + targetFunctionsIndices.size() +
                ", numOutputParams=" + outputParamsIndices.size() +
                ", numVaryingParams=" + varyingParamsIndices.size() +
                '}';
    }

    public List<OptimusParamsRow> getRows() {
        return rows;
    }

    public Chart createTargetFunctionChart(int targetFunctionIndex, OptimusOptimizationTargetFunc targetFunc, Integer numCaoVarPar) {
        OptimusTargetFuncsType optimusTargetFuncsType = OptimusTargetFuncsType.valueOf(targetFunc.getFuncId());
        String chartTitle = String.format("Целевая функция %s", targetFunc.getName());
        Chart targetFunctionChart = new LineChart(Chart.Type.CURVE, chartTitle, "Количество итераций", "", AxisDomainType.INTEGER, targetFunc.getName(), "", AxisDomainType.REAL);
        Optional<OptimusParamsRow> mostRationalDecisionOpt = getMostRationalDecision(optimusTargetFuncsType == OptimusTargetFuncsType.MAX, targetFunctionIndex);
        mostRationalDecisionOpt
                .ifPresent(mostRationalDecision -> targetFunctionChart.setDescription(createDescription(mostRationalDecision,numCaoVarPar)));

        rows.stream()
                .map(row -> row.toTargetFunctionPoint(0))
                .peek(point -> mostRationalDecisionOpt
                        .map(mostRationalDecision -> mostRationalDecision.toTargetFunctionPoint(targetFunctionIndex))
                        .filter(superiorPoint -> superiorPoint.getX() == point.getX() && superiorPoint.getY() == point.getY())
                        .ifPresent(superiorPoint -> point.setType(PointType.SUPERIOR)))
                .forEach(targetFunctionChart::addPoint);
        return targetFunctionChart;
    }

    public LineChart createSliceChart(String xAxisName, String yAxisName,
                                  Integer targetVaryingParamIdx, Integer targetOutputVariableIdx,
                                  List<OptimusVaryingParamInterval> intervals) {
        String chartTitle = "График зависимости " + yAxisName + " от " + xAxisName;

        LineChart sliceChart = new LineChart(Chart.Type.DOTS, chartTitle,
                xAxisName, "", AxisDomainType.REAL,
                yAxisName, "", AxisDomainType.REAL);

        rows.stream()
                .filter(row -> !row.isError())
                .filter(row -> intervals.stream().allMatch(interval -> {
                    Double value = row.getVaryingParams().get(interval.getVaryingParamIndex());
                    return interval.includes(value);
                }))
                .map(row -> row.toVaryingParamPoint(targetVaryingParamIdx, targetOutputVariableIdx))
                .forEach(sliceChart::addPoint);

        return sliceChart;
    }

    public Chart createConstraintChart(int constraintIndex, OptimusOptimizationConstraint constraint) {
        String chartTitle = String.format("Ограничение %s", constraint.getName());
        LineChart constraintChart = new LineChart(Chart.Type.CURVE, chartTitle, "Количество итераций", "", AxisDomainType.INTEGER, constraint.getName(), "", AxisDomainType.REAL);
        rows.stream()
                .map(row -> row.toConstraintPoint(constraintIndex))
                .forEach(constraintChart::addPoint);
        constraintChart.addYLineBorder(new LineDescription(constraint.getValue(), constraint.getName()));
        return constraintChart;
    }

    private List<String> createDescription(OptimusParamsRow mostRationalDecision, Integer numCaoVarPar) {
        List<Double> varyingParams = mostRationalDecision.getVaryingParams();
        String varyingParamsList = IntStream.range(0, numCaoVarPar)
                .mapToObj(paramIndex -> String.format("par%d=%.3f", (paramIndex + 1), varyingParams.get(paramIndex)))
                .collect(joining(","));

        List<Double> constraints = mostRationalDecision.getConstraints();
        String constraintsList = IntStream.range(0, constraints.size())
                .mapToObj(constraintIndex -> String.format("c%d=%.3f", (constraintIndex + 1), constraints.get(constraintIndex)))
                .collect(joining(","));
        String targetFuncDescription = String.format("Значение целевой функции: %s. Итерация: %s", mostRationalDecision.getTargetFunctionValue(0), mostRationalDecision.getIteration());
        return asList("Наиболее рациональное решение:", "Варьируемые параметры:", varyingParamsList, targetFuncDescription, "Ограничения:", constraintsList);
    }


    public Optional<OptimusParamsRow> getMostRationalDecision(boolean isFindMax, int targetFunctionIndex) {
        Comparator<OptimusParamsRow> comparator = getFuncComparator(isFindMax, targetFunctionIndex);

        return rows.stream()
                .filter(OptimusParamsRow::isAcceptable)
                .sorted(comparator)
                .findFirst();
    }


    private Comparator<OptimusParamsRow> getFuncComparator(boolean isFindMax, int targetFunctionIndex) {
        Comparator<OptimusParamsRow> comparator = Comparator.comparingDouble(row -> row.getTargetFunctionValue(targetFunctionIndex));
        // This condition was inversed by Anton Lukashin because default sorting order is ascending.
        if (isFindMax) {
            comparator = comparator.reversed();
        }
        return comparator;
    }

    public List<OptimusParamsRow> getParetoSet(List<Comparator<Double>> funcComparators) {
        if (getNumTargetFunctions() <= 1) {
            throw new IllegalArgumentException("Unexpected target functions count");
        }
        List<OptimusParamsRow> acceptableRows = rows.stream().filter(OptimusParamsRow::isAcceptable).collect(toList());

        List<Double> maxTargetFunctionValues = IntStream.range(0, funcComparators.size())
                .mapToObj(comparatorIndex -> getMaxTargetFuncValue(comparatorIndex, funcComparators.get(comparatorIndex), acceptableRows))
                .collect(toList());

        List<OptimusParamsRow> borders = acceptableRows.stream()
                .filter(row -> IntStream.range(0, maxTargetFunctionValues.size()).anyMatch(i -> maxTargetFunctionValues.get(i).equals(row.getTargetFunctionValue(i))))
                .collect(toList());

        return borders.stream()
                .filter(row -> borders.stream().noneMatch(potentialDominator -> row.isDominated(potentialDominator, funcComparators)))
                .collect(toList());
    }

    private Double getMaxTargetFuncValue(int targetFunctionIndex, Comparator<Double> comparator, List<OptimusParamsRow> rows) {
        return rows.stream()
                .map(row -> row.getTargetFunctionValue(targetFunctionIndex))
                .max(comparator).orElseThrow(() -> new IllegalArgumentException("No data"));
    }

    public int getNumTargetFunctions() {
        return targetFunctionsIndices.size();
    }

    public int getNumConstraints() {
        return constraintsIndices.size();
    }
}
