package ru.crp.lvms.optimizations.optimus.results;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FilenameUtils;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.Chart;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.resultsHandlers.KeyResultHandler;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.mf.results.ModFrontierParams;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationConstraint;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationTargetFunc;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationVaryingParam;
import ru.crp.lvms.optimizations.optimus.results.OptimusParams;

import javax.persistence.criteria.CriteriaBuilder;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

public class OptimusParamsHandler extends KeyResultHandler {

    private ObjectMapper objectMapper;
    private final Function<Integer, List<OptimusOptimizationTargetFunc>> targetFunctionsSupplier;
    private final Function<Integer, List<OptimusOptimizationConstraint>> constraintsSupplier;
    private final Function<Integer, List<OptimusOptimizationVaryingParam>> varParamsSupplier;

    public OptimusParamsHandler(Path pathToStorage, ObjectMapper objectMapper,
                                Function<Integer, List<OptimusOptimizationTargetFunc>> targetFunctionsSupplier,
                                Function<Integer, List<OptimusOptimizationConstraint>> constraintsSupplier,
                                Function<Integer, List<OptimusOptimizationVaryingParam>> varParamsSupplier) {
        super(pathToStorage);
        this.objectMapper = objectMapper;
        this.targetFunctionsSupplier = targetFunctionsSupplier;
        this.constraintsSupplier = constraintsSupplier;
        this.varParamsSupplier = varParamsSupplier;
    }


    @Override
    public boolean canHandle(ResultType resultType, Path pathToFile) {
        return resultType == ResultType.optimusParams && Files.isRegularFile(pathToFile);
    }


    @Override
    protected List<ResultEntity> doHandleResultsFile(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        List<ResultEntity> extractedResults = new ArrayList<>();

        ResultEntity rawParams = initCommonResultProperties(pathToFile, resultsParent, createdBy);
        rawParams.setName(pathToFile.getFileName().toString());
        rawParams.setType(ResultType.optimusParams);
        extractedResults.add(rawParams);

        List<OptimusOptimizationTargetFunc> targetFuncs = targetFunctionsSupplier.apply(resultsParent.getId());
        List<OptimusOptimizationConstraint> constraintsSettings = constraintsSupplier.apply(resultsParent.getId());
        List<OptimusOptimizationVaryingParam> optimusOptimizationVaryingParam = varParamsSupplier.apply(resultsParent.getId());
        try {
            if (targetFuncs.size() == 1) {
                extractedResults.addAll(handleSingleTargetFunction(pathToFile, resultsParent, createdBy, targetFuncs, constraintsSettings, optimusOptimizationVaryingParam));
            }
            return extractedResults;
        } catch (IOException ex) {
            throw new ResultProcessingException(ex);
        }
    }


    private List<ResultEntity> handleSingleTargetFunction(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy,
                                                          List<OptimusOptimizationTargetFunc> optimusOptimizationTargetFunc, List<OptimusOptimizationConstraint> constraintsSettings,
                                                          List<OptimusOptimizationVaryingParam> optimusOptimizationVaryingParam) throws IOException {
        OptimusParams optimusParams = parseOptimusParams(pathToFile);
        List<ResultEntity> extractedResults = new ArrayList<>();
        for (int i = 0; i < optimusParams.getNumTargetFunctions(); ++i) {
            OptimusOptimizationTargetFunc currentTargetFunc = optimusOptimizationTargetFunc.get(i);
            Chart targetFunctionChart = optimusParams.createTargetFunctionChart(i, currentTargetFunc, optimusOptimizationVaryingParam.size());
            String curveFileSuffix = String.format("%s-target-function", currentTargetFunc.getName());
            Path pathToJson = saveCurve(targetFunctionChart.toChartsSet(), curveFileSuffix, pathToFile);
            ResultEntity resultCurve = initCommonResultProperties(pathToJson, resultsParent, createdBy);
            resultCurve.setName(String.format("Целевая функция %s", currentTargetFunc.getName()));
            resultCurve.setType(ResultType.curves);
            extractedResults.add(resultCurve);
        }

        for (int i = 0; i < optimusParams.getNumConstraints(); ++i) {
            OptimusOptimizationConstraint currentConstraint = constraintsSettings.get(i);
            Chart constraintChart = optimusParams.createConstraintChart(i, currentConstraint);

            String curveFileSuffix = String.format("%s-constraint", currentConstraint.getName());
            Path pathToJson = saveCurve(constraintChart.toChartsSet(), curveFileSuffix, pathToFile);
            ResultEntity resultCurve = initCommonResultProperties(pathToJson, resultsParent, createdBy);
            resultCurve.setName(String.format("Ограничение %s", currentConstraint.getName()));
            resultCurve.setType(ResultType.curves);
            extractedResults.add(resultCurve);
        }
        return extractedResults;
    }


    public static OptimusParams parseOptimusParams(Path pathToFile) throws IOException {
        final boolean[] resultsStarted = new boolean[]{false};
        List<String[]> headerData = Files.lines(pathToFile, Charset.forName("CP1251"))
                .skip(2)
                .limit(3)
                .map(str -> str.split("\\s+"))
                .collect(toList());
        Boolean optTypeKey = headerData.get(2)[0].equals("OPT");
        List<String[]> resultData = Files.lines(pathToFile, Charset.forName("CP1251")).filter(str -> {
            if (resultsStarted[0]) return true;
            if (optTypeKey && str.startsWith("\"GOAL\"")) {
                resultsStarted[0] = true;
                return false;
            }
            else if (!optTypeKey && str.startsWith("1 0 0 0")) {
                resultsStarted[0] = true;
                return true;
            }
            return false;
        })
                .map(str -> str.split("\\s+"))
                .collect(toList());
        List<String> tempArray = new ArrayList<String>();
        if (optTypeKey) {
            for (int i = 0; i < Integer.parseInt(headerData.get(0)[0]); ++i) {
                tempArray.add("par" + Integer.toString(i + 1));
            }
            for (int i = 0; i < Integer.parseInt(headerData.get(1)[0]) - 1; ++i) {
                tempArray.add("out" + Integer.toString(i + 1));
            }
            for (int i = 0; i < 1; ++i) {
                tempArray.add("obj" + Integer.toString(i + 1));
            }
            for (int i = 0; i < Integer.parseInt(headerData.get(1)[0]) - 2; ++i) {
                tempArray.add("c" + Integer.toString(i + 1));
            }
        }
        else {
            for (int i = 0; i < Integer.parseInt(headerData.get(0)[0]); ++i) {
                tempArray.add("par" + Integer.toString(i + 1));
            }
            for (int i = 0; i < Integer.parseInt(headerData.get(1)[0]); ++i) {
                tempArray.add("out" + Integer.toString(i + 1));
            }
            for (int i = 0; i < 1; ++i) {
                tempArray.add("obj" + Integer.toString(i + 1));
            }
            for (int i = 0; i < Integer.parseInt(headerData.get(1)[0])-1; ++i) {
                tempArray.add("c" + Integer.toString(i + 1));
            }
        }
        String[] optimusParamsHeader = new String[tempArray.size()];
        OptimusParams optimusParams = new OptimusParams(tempArray.toArray(optimusParamsHeader));
        resultData.forEach(s-> optimusParams.addResultRow(s, optTypeKey));
        return optimusParams;
    }


    private Path saveCurve(ChartsSet targetFunctionCurve, String fileNameSuffix, Path pathToRawFile) throws IOException {
        String jsonFileName = FilenameUtils.removeExtension(pathToRawFile.getFileName().toString());
        jsonFileName = String.format("%s-%s.json", jsonFileName, fileNameSuffix);
        Path pathToJsonFile = pathToRawFile.resolveSibling(jsonFileName);
        objectMapper.writeValue(pathToJsonFile.toFile(), targetFunctionCurve);
        return pathToJsonFile;
    }
}
