package ru.crp.lvms.optimizations.optistruct.results;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FilenameUtils;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.Chart;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.chart.Point;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.resultsHandlers.KeyResultHandler;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructOptimizationBaseSetting;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.function.Function;

import static java.util.stream.Collectors.toList;

public class OptistructParamsHandler extends KeyResultHandler {
    private ObjectMapper objectMapper;

    private Function<Integer, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting>> baseSettingSupplier;

    public OptistructParamsHandler(Path pathToStorage, ObjectMapper objectMapper,
                                   Function<Integer, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting>> baseSettingSupplier) {
        super(pathToStorage);
        this.objectMapper = objectMapper;
        this.baseSettingSupplier = baseSettingSupplier;
    }


    @Override
    public boolean canHandle(ResultType resultType, Path pathToFile) {
        return resultType == ResultType.optistructRawData && Files.isRegularFile(pathToFile);
    }


    @Override
    protected List<ResultEntity> doHandleResultsFile(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        List<ResultEntity> extractedResults = new ArrayList<>();

        ResultEntity rawParams = initCommonResultProperties(pathToFile, resultsParent, createdBy);
        rawParams.setName(pathToFile.getFileName().toString());
        rawParams.setType(ResultType.optistructRawData);
        extractedResults.add(rawParams);

        try {
            Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> settings = baseSettingSupplier.apply(resultsParent.getId());
            Integer numConstraints = settings.get(OptistructBaseSettingType.NUM_CONSTRAINTS).getCurrentValue();
            extractedResults.addAll(extractCurves(pathToFile, resultsParent, createdBy, numConstraints));
            return extractedResults;
        } catch (IOException | NumberFormatException ex) {
            throw new ResultProcessingException(ex);
        }
    }


    private List<ResultEntity> extractCurves(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy, int numOfConstraints) throws IOException {
        List<List<Double>> curveData = Files.lines(pathToFile)
                .map(str -> Arrays.stream(str.split("\\s"))
                        .filter(s -> !s.trim().isEmpty())
                        .map(Double::parseDouble)
                        .collect(toList()))
                .collect(toList());
        if (curveData.isEmpty()) {
            throw new ResultProcessingException(String.format("Empty curve data file: %s", pathToFile));
        }
        int numCurves = curveData.get(0).size() - 1;
        if (numCurves < 1) {
            throw new ResultProcessingException(String.format("Wrong curve data file format: %s", pathToFile));
        }
        // key - file suffix
        Map<String, LineChart> charts = new LinkedHashMap<>(numCurves);
        charts.put("target-function-curve.json", new LineChart(Chart.Type.CURVE, "Изменение целевой функции", "Номер итерации", "", "Целевая функция", ""));
        for (int i = 0; i < numOfConstraints; ++i) {
            charts.put("constraint-" + i + "-violation.json", new LineChart(Chart.Type.CURVE, "Нарушение ограничения " + (i + 1), "Номер итерации", "", "Ограничение " + (i + 1), ""));
        }
        int numResponses = numCurves - charts.size();
        for (int i = 0; i < numResponses; ++i) {
            charts.put("response-" + i + ".json", new LineChart(Chart.Type.CURVE, "Изменение отклика № " + (i + 1), "Номер итерации", "", "Отклик № " + (i + 1), ""));
        }
        curveData.forEach(row -> {
            List<Chart> resultCurves = new ArrayList<>(charts.values());
            for (int i = 0; i < resultCurves.size(); ++i) {
                resultCurves.get(i).addPoint(Point.commonPoint(row.get(0), row.get(i + 1)));
            }
        });

        return charts.entrySet().stream()
                .map(entry -> saveCurveToFile(entry.getKey(), entry.getValue().toChartsSet(), pathToFile))
                .map(pathToJson -> {
                    ResultEntity resultCurve = initCommonResultProperties(pathToJson, resultsParent, createdBy);
                    resultCurve.setName(pathToJson.getFileName().toString());
                    resultCurve.setType(ResultType.curves);
                    return resultCurve;
                })
                .collect(toList());
    }


    private Path saveCurveToFile(String fileSuffix, ChartsSet chartsSet, Path rawDataFile) {
        String jsonFileName = FilenameUtils.removeExtension(rawDataFile.getFileName().toString());
        jsonFileName = jsonFileName + "-" + fileSuffix;
        Path pathToJsonFile = rawDataFile.resolveSibling(jsonFileName);
        try {
            objectMapper.writeValue(pathToJsonFile.toFile(), chartsSet);
        } catch (IOException ex) {
            throw new ResultProcessingException(ex);
        }
        return pathToJsonFile;
    }
}
