package ru.crp.lvms.optimizations.optistruct.tcl;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructConvertDirection;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.List;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Collections.singletonList;
import static java.util.stream.Collectors.toList;

@Component
public class OptistructConvertScriptGenerator {
    public static final String INCLUDE_CDB_STR = "*feinputwithdata2 \"#ansys/ansys\" \"OUTPUT_FILES_PATH/file_name.cdb\" 0 0 0 0 0 1 4 1 0";
    public static final String INCLUDE_LC_PATTERN = "*feinputwithdata2 \"#ansys/ansys\" \"OUTPUT_FILES_PATH/%s\" 0 0 0 0 0 1 4 1 0";
    public static final String INCLUDE_LC_CHANGENAME = "*setvalue loadcols id=%d name=\"loadcol%s_force\"";
    public static final String INCLUDE_LC_CHANGEID ="*setvalue loadcols id=%d id={loadcols %s}";

    @Value("${optistuctTcl.abaqus2optistruct}")
    private String abaqusToOptistruct;

    private Path abaqusToOptistructPath;

    @Value("${optistuctTcl.ansys2optistruct}")
    private String ansysToOptistruct;

    private Path ansysToOptistructPath;

    @Value("${optistuctTcl.optistruct2ansys}")
    private String optistructToAnsys;

    private Path optistructToAnsysPath;

    @Value("${optistuctTcl.optistruct2abaqus}")
    private String optistructToAbaqus;

    private Path optistructToAbaqusPath;

    @Autowired
    private OptimizationManager optimizationManager;

    @PostConstruct
    private void init() throws IOException {
        abaqusToOptistructPath = Paths.get(abaqusToOptistruct);
        ansysToOptistructPath = Paths.get(ansysToOptistruct);
        optistructToAnsysPath = Paths.get(optistructToAnsys);
        optistructToAbaqusPath = Paths.get(optistructToAbaqus);
        if (Files.notExists(abaqusToOptistructPath) || Files.notExists(ansysToOptistructPath) ||
                Files.notExists(optistructToAnsysPath) || Files.notExists(optistructToAbaqusPath)) {
            throw new InternalException("Convert скрипты не найдены");
        }
    }

    public Path generateConfig(OptimizationEntity optimization, OptistructConvertDirection optistructConvertDirection) throws IOException {
        Path tempConfigFile = Files.createTempFile(null, null);
        Path sourceFile = chooseSolver(optistructConvertDirection);
        // we need to add loadconditions for ansys to optistruct converter
        if (optistructConvertDirection == OptistructConvertDirection.ansysToOptistruct) {
            List<String> loadConditionNames = optimizationManager.getInputFilesForOptimization(optimization, singletonList(FileType.LoadCondition)).stream()
                    .map(FileData::getName)
                    .collect(toList());
            List<String> transformedFile = Files.readAllLines(sourceFile).stream()
                    .flatMap(str -> {
                        if (!str.trim().equals(INCLUDE_CDB_STR)) {
                            return Stream.of(str);
                        }
                        return Stream.concat(
                                Stream.concat(
                                 Stream.concat(
                                Stream.of(str),IntStream.range(0, loadConditionNames.size())
                                                 .mapToObj(i -> String.format(INCLUDE_LC_PATTERN, loadConditionNames.get(i)))),
                                        IntStream.range(0, loadConditionNames.size()+1)
                                                .mapToObj(i ->String.format(INCLUDE_LC_CHANGENAME + "\n" + INCLUDE_LC_CHANGEID,
                                                        i+1,Integer.toString(1000000+i), i+1, Integer.toString(1000000+i)))),
                                        IntStream.range(0, loadConditionNames.size())
                                                .mapToObj(i ->{if (loadConditionNames.get(i).lastIndexOf(".s0")!=-1) {
                                                    return String.format(INCLUDE_LC_CHANGENAME + "\n" + INCLUDE_LC_CHANGEID, 1000000+i+1,
                                                            loadConditionNames.get(i).substring(loadConditionNames.get(i).lastIndexOf(".s0") + 3),
                                                            1000000+i+1, loadConditionNames.get(i).substring(loadConditionNames.get(i).lastIndexOf(".s0") + 3));
                                                }
                                                return String.format(INCLUDE_LC_CHANGENAME + "\n" + INCLUDE_LC_CHANGEID, 1000000+i+1,
                                                        loadConditionNames.get(i).substring(loadConditionNames.get(i).lastIndexOf(".s") + 2),
                                                        1000000+i+1, loadConditionNames.get(i).substring(loadConditionNames.get(i).lastIndexOf(".s") + 2));
                            })
                        );
                    })
                    .collect(toList());
            Files.write(tempConfigFile, transformedFile);
        } else {
            Files.copy(sourceFile, tempConfigFile, StandardCopyOption.REPLACE_EXISTING);
        }
        return tempConfigFile;
    }

    private Path chooseSolver(OptistructConvertDirection optistructConvertDirection) {
        switch (optistructConvertDirection) {
            case abaqusToOptistruct:
                return abaqusToOptistructPath;
            case ansysToOptistruct:
                return ansysToOptistructPath;
            case optistructToAnsys:
                return optistructToAnsysPath;
            case optistructToAbaqus:
            default:
                return optistructToAbaqusPath;
        }
    }
}
