package ru.crp.lvms.optimizations.optistruct.tcl;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructOptimizationAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructOptimizationAdditionalParameter;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

@Component
public class OptistructOssScriptGenerator {

    private final static int MESH_TYPE_STL = 5;
    private final static int MESH_TYPE_SOLID = 3;
    private final static int DENSITY_THRESHOLD = 8;
    private final static int LAPITERATION = 9;
    private final static int LAP_FEATURE_ANGLE = 10;
    private final static int NUM_ARGS = 11;
    private final static int REMESH_SIZE = 12;
    private final static int MESH_TYPE = 13;
    private final static int REMESH_FEATURE_ANGLE = 14;
    public static final String OS_SMOOTH_LINE_PREFIX = "*ossmooth_12";

    @Value("${optistuctTcl.oss}")
    private String ossTcl;

    private Path ossTclPath;

    @Autowired
    private OptistructOptimizationAdditionalParametersDao optistructOptimizationAdditionalParametersDao;

    @PostConstruct
    private void init() throws IOException {
        ossTclPath = Paths.get(ossTcl);
        if (Files.notExists(ossTclPath)) {
            throw new InternalException("Oss скрипт не найден");
        }
    }

    public Path generateConfig(OptimizationEntity optimization) throws IOException {
        Path tempConfigFile = Files.createTempFile(null, null);
        Files.write(tempConfigFile, formData(optimization));
        return tempConfigFile;
    }

    private List<String> formData(OptimizationEntity optimization) throws IOException {
        List<OptistructOptimizationAdditionalParameter> params = optistructOptimizationAdditionalParametersDao.findForOptimization(optimization.getId());
        String stlLine = createStringArrayBlock(params,MESH_TYPE_STL);
        String cdbLine = createStringArrayBlock(params,MESH_TYPE_SOLID);
        String numArgs = getNumStringArgs(params);
        try (Stream<String> lines = Files.lines(ossTclPath)) {
            List<String> fileContent = lines.map(line -> {
                if (line.trim().startsWith(OS_SMOOTH_LINE_PREFIX)) {
                    return line.substring(0, line.length() - 1) + numArgs;
                } else {
                    return line;
                }
            })
                    .collect(toList());
            fileContent.add(0, stlLine);
            fileContent.add(5, cdbLine);
            return fileContent;
        }
    }

    private String createStringArrayBlock(List<OptistructOptimizationAdditionalParameter> params,Integer meshType) {
        String numArgs = getNumStringArgs(params);
        return "*createstringarray " + numArgs
                + " " + createIsoSurf(params,meshType)
                + " " + (("3".equals(numArgs)) ? createRemesh(params) : "")
                + " \"other_params: 1 0 0 0 10\"";
    }


    private String getNumStringArgs(List<OptistructOptimizationAdditionalParameter> params) {
        return params.stream().filter(parameter -> parameter.getParameterId() == NUM_ARGS)
                .map(OptistructOptimizationAdditionalParameter::getParameterTypeCodeName)
                .findAny().orElse("3");
    }

    private String createIsoSurf(List<OptistructOptimizationAdditionalParameter> params,Integer meshType) {
        String densityThreshold = selectFromParamsWithId(params, DENSITY_THRESHOLD).orElse("0.3");
        String lapiteration = selectFromParamsWithId(params, LAPITERATION).orElse("0");
        String lapFeatureAngle = selectFromParamsWithId(params, LAP_FEATURE_ANGLE).orElse("0");
        return "\"isosurf: 3 "+meshType+" "+densityThreshold + " 0 -1 0 0 " + lapiteration + " " + lapFeatureAngle + " 1 0\"";
    }


    private String createRemesh(List<OptistructOptimizationAdditionalParameter> params) {
        String remeshSize = selectFromParamsWithId(params, REMESH_SIZE).orElse("0.05");
        String meshType = params.stream().filter(parameter -> parameter.getParameterId() == MESH_TYPE)
                .map(OptistructOptimizationAdditionalParameter::getParameterTypeCodeName)
                .findAny().orElse("0");
        String remeshFeatureAngle = selectFromParamsWithId(params, REMESH_FEATURE_ANGLE).orElse("30");
        return "\"remesh: " + remeshSize + " " + meshType + " 1 2 " + remeshFeatureAngle + "\"";
    }

    private Optional<String> selectFromParamsWithId(List<OptistructOptimizationAdditionalParameter> params, Integer id) {
        return params.stream().filter(parameter -> parameter.getParameterId() == id).findAny()
                .map(OptistructOptimizationAdditionalParameter::getValue);
    }
}
