package ru.crp.lvms.optimizations.tosca.config.commandConverters;

import org.apache.commons.lang3.tuple.Pair;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.tosca.config.Command;
import ru.crp.lvms.optimizations.tosca.config.OptimizationTypeParameters;
import ru.crp.lvms.optimizations.tosca.config.argumentConverters.*;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationTechnologicalConstraint;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationTechnologicalConstraintManager;

import java.util.*;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toList;

abstract public class CommandConverter {
    public static final String DVCON_ID_PREFIX = "TechnologicalConstraint";
    protected static final String LINK_ID_PREFIX = "Link";
    public static final String UTIL_TC_NAME_PREFIX = "TC_";

    public static final int CAST_TC_ID = 5;
    public static final int SYMMETRY_CONDITION_TC_ID = 6;
    public static final int TOPOGRAPHIC_TC_ID = 8;
    public static final int SHAPE_TC_ID = 7;
    public static final int MAX_DIAMETER_TC_ID = 2;

    private static final Map<Integer, List<Pair<String, String>>> DEFAULT_PARAMETERS = new HashMap<>();

    static {
        DEFAULT_PARAMETERS.put(CAST_TC_ID, new ArrayList<>(asList(Pair.of("ANGLE", "0"), Pair.of("MID_PLANE", "NONE"), Pair.of("FILTER_MAT", "NO"))));
        DEFAULT_PARAMETERS.put(SYMMETRY_CONDITION_TC_ID, new ArrayList<>(Collections.singletonList(Pair.of("MASTER", "AUTO"))));
        DEFAULT_PARAMETERS.put(SHAPE_TC_ID, new ArrayList<>(Collections.singletonList(Pair.of("CHECK_BC", "NO"))));
    }

    private List<CommandArgumentConverter> argumentConverters;

    public CommandConverter(ToscaOptimizationTechnologicalConstraintManager toscaOptimizationTechnologicalConstraintManager) {
        argumentConverters = new ArrayList<>();
//        argumentConverters.add(new CSArgumentConverter(toscaTechnologicalConstraintTreeValueManager));
        // order is important
        argumentConverters.add(new IgnoreArgumentsConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new SymmetryConditionTypeArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new CheckDofArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new MasterArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new MidPlaneArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new PullDirArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new MultipleChildrenArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        argumentConverters.add(new TreeLeafArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
        // should always be last
        argumentConverters.add(new SimpleArgumentConverter(toscaOptimizationTechnologicalConstraintManager));
    }


    abstract public Stream<Command> convert(OptimizationTypeParameters typeParameters, ToscaOptimizationTechnologicalConstraint tc,
                                            Integer index, String space, List<TechnologicalConstraintTreeNodeValue> values);

    abstract public boolean canConvert(ToscaOptimizationTechnologicalConstraint tc);


    protected Command createDvConCommand(OptimizationTypeParameters typeParameters, ToscaOptimizationTechnologicalConstraint tc, int index) {
        Command command = new Command("DVCON" + typeParameters.getDvName())
                .addParameter("ID_NAME", DVCON_ID_PREFIX + index);
        if ((tc.getComandName() == null || !tc.getComandName().startsWith(UTIL_TC_NAME_PREFIX))) {
            command.addParameter("CHECK_TYPE", tc.getComandName());
        }
        return command;
    }


    protected List<Pair<String, String>> convertTCValues(ToscaOptimizationTechnologicalConstraint tc, List<TechnologicalConstraintTreeNodeValue> values) {
        values.sort(Comparator.comparing(TechnologicalConstraintTreeNodeValue::getTechnologicalConstraintTreeId));
        List<Pair<String, String>> convertedValues = new ArrayList<>();
        for (TechnologicalConstraintTreeNodeValue value : values) {
            List<Pair<String, String>> convertedValue = selectArgumentConverter(tc, value).convertValue(tc, value, values, convertedValues);
            convertedValues.addAll(convertedValue);
        }
        addDefaultValues(tc, convertedValues);
        return convertedValues;
    }


    private void addDefaultValues(ToscaOptimizationTechnologicalConstraint tc, List<Pair<String, String>> convertedParams) {
        List<Pair<String, String>> defaultParamsToUse = DEFAULT_PARAMETERS.getOrDefault(tc.getTechnologicalConstraintId(), Collections.emptyList()).stream()
                .filter(defaultParam -> convertedParams.stream().noneMatch(p -> p.getKey().equals(defaultParam.getKey())))
                .collect(toList());
        convertedParams.addAll(defaultParamsToUse);
    }


    private CommandArgumentConverter selectArgumentConverter(ToscaOptimizationTechnologicalConstraint constraint, TechnologicalConstraintTreeNodeValue value) {
        return argumentConverters.stream().filter(ac -> ac.canConvert(constraint, value)).findFirst().get();
    }
}
