package ru.crp.lvms.optimizations.tosca.controllers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.tosca.ToscaSetsParser;
import ru.crp.lvms.optimizations.tosca.entities.ToscaSet;
import ru.crp.lvms.optimizations.tosca.enums.SetType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaSettingsType;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.URLs;

import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.model.mappers.PgArrayPropertyEditor.LOGGER;
import static ru.crp.lvms.tasks.fileGenerators.ToscaInputFileGenerator.ABAQUS_CAE_FILE_EXT;
import static ru.crp.lvms.tasks.fileGenerators.ToscaInputFileGenerator.ANSYS_CAE_FILE_EXT;

@RestController
@RequestMapping(value = URLs.TOSCA_SETS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class ToscaSetsController extends ToscaAbstractOptimizationSettingsController<ToscaSet> {

    @Autowired
    private ToscaSetsParser toscaSetsParser;

    @Autowired
    private FileManager fileManager;

    @Override
    protected ToscaBaseSettingType getToskaOptimizationBaseSettingType() {
        return ToscaBaseSettingType.NUM_SETS;
    }

    @Override
    protected ToscaSettingsType getSettingsType() {
        return ToscaSettingsType.Set;
    }


    @Override
    public List<? extends AbstractOptimizationSettings> getAllSettings(@PathVariable("optimizationId") int optimizationId, @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager()
                .findOptimizationSettings(optimizationId, settingsType, currentUser.getUser());
    }

    @RequestMapping(value = "/selected", method = RequestMethod.GET)
    public List<ToscaSet> findSelectedSets(@PathVariable("optimizationId") Integer optimizationId,
                                           @Nullable @RequestParam(value = "type", required = false) SetType setType,
                                           @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().findSelectedSets(optimizationId, setType);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{setId}", method = RequestMethod.POST)
    public AbstractOptimizationSettings addToSelected(@PathVariable("optimizationId") Integer optimizationId,
                                                      @PathVariable("setId") Integer setId,
                                                      @RequestParam("isSelected") Boolean isSelected,
                                                      @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        ToscaSet set = (ToscaSet) getSettingsManager()
                .findOptimizationSettingsById(setId, ToscaSettingsType.Set, currentUser.getUser());
        set.setIsSelected(isSelected);
        return getSettingsManager().updateSettings(set, ToscaSettingsType.Set);
    }

    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<ToscaSet> suggestSets(@PathVariable("optimizationId") Integer optimizationId,
                                      @RequestBody SuggestRequest suggestRequest,
                                      @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().suggestSets(optimizationId, suggestRequest, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/extract", method = RequestMethod.POST)
    public void extractSetsFromFile(@PathVariable("optimizationId") Integer optimizationId,
                                    @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = checkOptimizationSolverType(optimizationId, currentUser);
        Integer CAESolverId = optimization.getSolverCaeId();
        if (CAESolverId == null) {
            throw new BadRequestException("CAE солвер должен быть задан");
        }
        OptimizationCae CAESolver = OptimizationCae.valueOf(CAESolverId);
        List<FileEntity> filesWithSets = optimizationManager.getOptimizationFiles(optimization, FileType.CAEFile, currentUser.getUser()).stream()
                .filter(file -> (CAESolver == OptimizationCae.ANSYS && file.getName().endsWith(ANSYS_CAE_FILE_EXT)) ||
                        CAESolver == OptimizationCae.ABAQUS && file.getName().endsWith(ABAQUS_CAE_FILE_EXT))
                .collect(toList());
        if (filesWithSets.isEmpty()) {
            throw new BadRequestException("Файл для извлечения наборов узлов и элементов не найден");
        }
        if (filesWithSets.size() > 1) {
            throw new BadRequestException("Должен быть только один файл модели");
        }
        FileEntity file = filesWithSets.get(0);
        Path pathToFile = fileManager.resolvePathToFile(file, currentUser.getUser());
        if (pathToFile == null || !Files.isRegularFile(pathToFile)) {
            LOGGER.error(String.format("File %d is present in db but not in file system", file.getId()));
            throw new InternalException(String.format("Файл '%s' не найден на диске", file.getName()));
        }

        try {
            List<ToscaSet> newSets = toscaSetsParser.extractSets(optimizationId, CAESolver, pathToFile);
            getSettingsManager().resetSets(optimizationId, newSets);
        } catch (IOException | IllegalArgumentException ex) {
            throw new InternalException(String.format("Не удалось извлечь наборы узлов и элементов из файла %s", file.getName()), ex);
        }
    }
}
