package ru.crp.lvms.optimizations.tosca.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.misc.BaseManager;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.tosca.dao.ToscaTechnologicalConstraintTreeDao;
import ru.crp.lvms.optimizations.tosca.dao.ToscaTechnologicalConstraintTreeValueDao;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNode;

import javax.annotation.Nullable;
import java.util.List;
import java.util.Objects;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNode.SELECT_TYPE;

@Service
public class ToscaTechnologicalConstraintTreeValueManager extends BaseManager<TechnologicalConstraintTreeNodeValue> {

    @Autowired
    private ToscaTechnologicalConstraintTreeValueDao toscaTechnologicalConstraintTreeValueDao;

    @Autowired
    private ToscaTechnologicalConstraintTreeDao toscaTechnologicalConstraintTreeDao;

    @Autowired
    private ToscaOptimizationTechnologicalConstraintManager toscaOptimizationTechnologicalConstraintManager;

    @Override
    protected ToscaTechnologicalConstraintTreeValueDao aliveDao() {
        return toscaTechnologicalConstraintTreeValueDao;
    }

    public List<TechnologicalConstraintTreeNodeValue> findValuesOfTC(Integer optimizationTCId) {
        return toscaTechnologicalConstraintTreeValueDao.findForConstraint(optimizationTCId);
    }

    @Override
    public int create(TechnologicalConstraintTreeNodeValue entity, @Nullable UserEntity currentUser) {
        checkCreate(entity, currentUser);
        // there may be not filled fields in 'entity' so we need to it's predefined properties from db
        TechnologicalConstraintTreeNode treeNode = toscaTechnologicalConstraintTreeDao.findOne(entity.getTechnologicalConstraintTreeId());
        deletePreviousSelectedNodes(entity.getTechnologicalConstraintId(), treeNode);
        boolean isValueEditable = Objects.equals(Boolean.TRUE, treeNode.getIsValueEditable());
        if (isValueEditable && entity.getValue() == null) {
            throw new BadRequestException(String.format("Значение поля '%s' должно быть задано", treeNode.getDisplayName()));
        } else {
            if (!isValueEditable && entity.getValue() != null) {
                entity.setValue(null);
            }
            return aliveDao().insert(entity);
        }
    }

    @Override
    public void update(TechnologicalConstraintTreeNodeValue entity, @Nullable UserEntity currentUser) {
        checkUpdate(entity, currentUser);
        // there may be not filled fields in 'entity' so we need to it's predefined properties from db
        TechnologicalConstraintTreeNode treeNode = toscaTechnologicalConstraintTreeDao.findOne(entity.getTechnologicalConstraintTreeId());
        Integer optimizationTcId = entity.getTechnologicalConstraintId();
        deletePreviousSelectedNodes(optimizationTcId, treeNode);
        boolean isValueEditable = Objects.equals(Boolean.TRUE, treeNode.getIsValueEditable());

        if (isValueEditable && entity.getValue() == null) {
            aliveDao().delete(entity.getId());
        } else {
            if (!isValueEditable && entity.getValue() != null) {
                entity.setValue(null);
            }
            aliveDao().updateValue(entity);
        }
    }

    private void deletePreviousSelectedNodes(int optimizationTcId, TechnologicalConstraintTreeNode entity) {
        if (SELECT_TYPE.equals(entity.getType())) {
            aliveDao().deleteByGroupAndParent(optimizationTcId, entity.getGroupName(), entity.getParentId(), entity.getId());
        }
    }


    public void deleteByTreeId(Integer treeId, int optimizationTCId) {
        aliveDao().deleteByTreeId(optimizationTCId, treeId);
        deleteDependencies(optimizationTCId);
    }

    public TechnologicalConstraintTreeNodeValue finByTreeId(Integer optimizationTCId, Integer technologicalConstraintTreeId) {
        return aliveDao().findByTreeId(optimizationTCId, technologicalConstraintTreeId);
    }

    public void deleteSeveralFieldsByFieldIds(List<Integer> treeIds, int optimizationTCId) {
        aliveDao().deleteSeveralFieldsByFieldIds(optimizationTCId, treeIds);
        deleteDependencies(optimizationTCId);
    }

    private void deleteDependencies(int optimizationTCId) {
        List<TechnologicalConstraintTreeNodeValue> presentFields = findValuesOfTC(optimizationTCId);
        List<Integer> presentFieldIds = presentFields.stream().map(TechnologicalConstraintTreeNodeValue::getTechnologicalConstraintTreeId).collect(toList());
        toscaOptimizationTechnologicalConstraintManager.setDependencies(presentFields);
        List<Integer> fieldsToDelete = presentFields.stream()
                .filter(f -> f.getDependencies() != null && !f.getDependencies().satisfies(f, presentFieldIds))
                .map(TechnologicalConstraintTreeNodeValue::getTechnologicalConstraintTreeId)
                .collect(toList());
        aliveDao().deleteWithChildrenByTreeId(optimizationTCId, fieldsToDelete);
    }
}
