package ru.crp.lvms.tasks.fileGenerators;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Component
public class AnsysConfigFileGenerator extends BaseInputFileGenerator {
    private static final Logger LOGGER = LoggerFactory.getLogger(AnsysConfigFileGenerator.class);

    public static final String GENERATED_FILE_NAME = "file.inp";

    private final static String INP_FILE_EXTENSION = "inp";
    private final static String DB_FILE_EXTENSION = "db";
    private final static String CDB_FILE_EXTENSION = "cdb";
    private final static Pattern S_FILE_PATTERN = Pattern.compile("^.*\\.s(\\d+)$", Pattern.CASE_INSENSITIVE);

    @Autowired
    private FileManager fileManager;

    public FileData generateInpFile(OptimizationEntity optimizationEntity) {
        List<FileEntity> optimizationFiles = optimizationManager.getOptimizationFiles(optimizationEntity, null);

        Stream<Function<List<FileEntity>, Optional<FileData>>> possibleFileSources =
                Stream.of(this::findAlreadyUploadedInpFile, this::formAccordinglyToDbModel, this::formAccordinglyToCdbModel);

        return possibleFileSources
                .map(source -> source.apply(optimizationFiles))
                .filter(Optional::isPresent)
                .map(Optional::get)
                .findFirst()
                .orElseThrow(() -> new InternalException(String.format("Не удалось сформировать файл '%s' для оптимизации %d", GENERATED_FILE_NAME, optimizationEntity.getId())));
    }

    private Optional<FileData> findAlreadyUploadedInpFile(List<FileEntity> optimizationFiles) {
        return optimizationFiles.stream()
                .filter(fe -> FilenameUtils.getExtension(fe.getName()).equalsIgnoreCase(INP_FILE_EXTENSION))
                .map(fe -> fileManager.resolvePathToFile(fe, null))
                .filter(Objects::nonNull)
                .filter(Files::exists)
                .map(file -> FileData.fromFile(AnsysConfigFileGenerator.GENERATED_FILE_NAME,
                        DataType.SESSION_FILE, file.toFile()))
                .findAny();
    }

    private Optional<FileData> formAccordinglyToDbModel(List<FileEntity> optimizationFiles) {
        return findFileByExtensionAndFormConfig(optimizationFiles, DB_FILE_EXTENSION);
    }

    private Optional<FileData> formAccordinglyToCdbModel(List<FileEntity> optimizationFiles) {
        return findFileByExtensionAndFormConfig(optimizationFiles, CDB_FILE_EXTENSION);
    }

    private Optional<FileData> findFileByExtensionAndFormConfig(List<FileEntity> optimizationFiles, String fileExtension) {
        return optimizationFiles.stream()
                .filter(fe -> FilenameUtils.getExtension(fe.getName()).equalsIgnoreCase(fileExtension))
                .findAny()
                .map(foundFile -> generateInpFileContent(optimizationFiles, FilenameUtils.getBaseName(foundFile.getName()), fileExtension))
                .flatMap(impFileContent -> {
                    try {
                        Path tmpFile = Files.createTempFile(null, null);
                        Files.write(tmpFile, impFileContent.getBytes());
                        return Optional.of(tmpFile);
                    } catch (IOException ex) {
                        LOGGER.error("Can't create temp file", ex);
                        return Optional.empty();
                    }
                })
                .map(file -> FileData.fromFile(AnsysConfigFileGenerator.GENERATED_FILE_NAME,
                        DataType.GENERATED, file.toFile()));
    }

    private String generateInpFileContent(List<FileEntity> optimizationFiles, String fileName, String fileExtension) {
        String fileContent = "";

        switch (fileExtension) {
            case DB_FILE_EXTENSION:
                fileContent += "resume,'" + fileName + "','db',,0,0\n/solu\n";
                break;
            case CDB_FILE_EXTENSION:
                fileContent += "/prep7\ncdread,db,'" + fileName + "','cdb',,,\n/solu\n";
                break;
            default:
                throw new InternalException("Unrecognized file extension: " + fileExtension);
        }

        OptionalInt indexOfLastSFile = findIndexOfLastSFile(optimizationFiles);
        OptionalInt indexOfFirstSFile = findIndexOfFirstSFile(optimizationFiles);

        if (indexOfLastSFile.isPresent()&&indexOfFirstSFile.isPresent()) {
            if(indexOfLastSFile.getAsInt()==indexOfFirstSFile.getAsInt()) {
                fileContent += "lssolve," + indexOfFirstSFile.getAsInt() + "," + indexOfLastSFile.getAsInt() + ",1";
            }
            else{
                Integer numLsFiles = getNumberOfSFiles(optimizationFiles);
                Double templsstep =(indexOfLastSFile.getAsInt()-indexOfFirstSFile.getAsInt())/(numLsFiles.doubleValue()-1);
                Integer lsstep = templsstep.intValue();
                fileContent += "lssolve," + indexOfFirstSFile.getAsInt() + "," + indexOfLastSFile.getAsInt() + ","+lsstep;
            }
        } else {
            fileContent += "solve";
        }
        String postProcString = formPostprocCommand(optimizationFiles);
        fileContent += "\n"+postProcString;
        return fileContent;
    }

    private OptionalInt findIndexOfLastSFile(List<FileEntity> optimizationFiles) {
        return optimizationFiles
                .stream()
                .filter(file -> file.getFileTypeId() == FileType.LoadCondition.getId())
                .map(file -> S_FILE_PATTERN.matcher(file.getName()))
                .filter(Matcher::matches)
                .map(matcher -> matcher.group(1))
                .mapToInt(Integer::valueOf)
                .max();
    }

    private OptionalInt findIndexOfFirstSFile(List<FileEntity> optimizationFiles) {
        return optimizationFiles
                .stream()
                .filter(file -> file.getFileTypeId() == FileType.LoadCondition.getId())
                .map(file -> S_FILE_PATTERN.matcher(file.getName()))
                .filter(Matcher::matches)
                .map(matcher -> matcher.group(1))
                .mapToInt(Integer::valueOf)
                .min();
    }

    private Integer getNumberOfSFiles(List<FileEntity> optimizationFiles) {
        List<FileEntity> optimizationSfiles = optimizationFiles
                .stream()
                .filter(file -> file.getFileTypeId() == FileType.LoadCondition.getId())
                .collect(Collectors.toList());

        return optimizationSfiles.size();
    }

    private String formPostprocCommand(List<FileEntity> optimizationFiles) {
        for (FileEntity f:optimizationFiles) {
            if (f.getName().contains("postproc.mac")) {
                return ("*use," + f.getName());
            }
        }
                    return("!No_postrpoc_file_detected");
    }
}
