package ru.crp.lvms.tasks.fileGenerators;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.CompositeMaterialManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.technology.TechnologyType;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import static ru.crp.cmls.api.TaskParentType.simulation;

@Component
public class SimulationInputFileGenerator extends BaseInputFileGenerator {

    @Value("${vsMacrosesDir}")
    private String pathToVSMacrosesStr;

    @Value("${vs3CMMacrosesDir}")
    private String pathToVSMacrosesFor3CMMaterialsStr;

    @Autowired
    private ECInputFileGenerator ecInputFileGenerator;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    public List<FileData> getInputFilesForTensionSimulation(TaskEntity task)
            throws IOException, ResourcePreparationException {

        SimulationEntity parent = checkAndGetParentSim(task);

        ArrayList<FileData> fileData = new ArrayList<>();

        addCommonSimulationFiles(fileData, parent);
        generateSimInputFile(fileData, parent);

        TechnologyType technologyType = compositeMaterialManager.findOne(parent.getCompositeMaterialId(), null).getTechnologyType();
        if (technologyType == TechnologyType.ADDITIVE_3CM) {
            addFilesFromDir(fileData, pathToVSMacrosesFor3CMMaterialsStr);
        } else {
            addFilesFromDir(fileData, pathToVSMacrosesStr);
        }

        return fileData;
    }


    public List<FileData> getInputFilesForBendSimulation(TaskEntity task)
            throws IOException, ResourcePreparationException {
        return getInputFilesForTensionSimulation(task);
    }

    public List<FileData> getInputFilesForCompressionSimulation(TaskEntity extendedTask)
            throws IOException, ResourcePreparationException {
        SimulationEntity simulation = checkAndGetParentSim(extendedTask);

        List<FileData> files = getInputFilesForTensionSimulation(extendedTask);

        ECEntity entity = ecManager.extend(ecManager.findOne(simulation.getEcId(), null), null);
        TechnologyType technologyType = compositeMaterialManager.findOne(entity.getCompositeMaterialId(), null).getTechnologyType();

        ecInputFileGenerator.generateECInputFile(files, entity, technologyType);

        return files;
    }

    public List<FileData> getInputFilesForThermalSimulation(TaskEntity task)
            throws IOException, ResourcePreparationException {
        SimulationEntity parent = checkAndGetParentSim(task);
        ArrayList<FileData> fileData = new ArrayList<>();

        generateSimThermalInputFile(fileData, parent);
        addCommonSimulationFiles(fileData, parent);
        TechnologyType technologyType = compositeMaterialManager.findOne(parent.getCompositeMaterialId(), null).getTechnologyType();
        if (technologyType == TechnologyType.ADDITIVE_3CM) {
            addFilesFromDir(fileData, pathToVSMacrosesFor3CMMaterialsStr);
        } else {
            addFilesFromDir(fileData, pathToVSMacrosesStr);
        }

        return fileData;
    }

    private void generateSimThermalInputFile(ArrayList<FileData> fileData, SimulationEntity parent) throws IOException {
        boolean isEpoxidePresent = compositeMaterialManager.findOne(parent.getCompositeMaterialId(), null).isEpoxidePresent();

        ThermalSimulationParameters parameters = ecManager.getThermalSimulationParameters(parent.getEcId(), isEpoxidePresent);
        if (!parameters.isAllParametersPresent()) {
            throw new TaskStartingException("Unable to find parameters for virtual simulation");
        }
        Path inputFile = Files.createTempFile(null, null);

        List<String> content = new ArrayList<>();
        content.add(commasToDots("1", "0"));
        content.add(commasToDots(parameters.getMatrixThermalConductivity(), "0"));
        content.add(commasToDots(parameters.getFiberThermalConductivity(), "0"));
        if (isEpoxidePresent) {
            content.add(commasToDots(parameters.getEpoxideThermalConductivity(), "0"));
        }
        content.add(commasToDots(parameters.getMatrixLinearThermalExpansion(), "0"));
        content.add(commasToDots(parameters.getFiberLinearThermalExpansion(), "0"));
        if (isEpoxidePresent) {
            content.add(commasToDots(parameters.getEpoxideLinearThermalExpansion(), "0"));
        }

        String inputData = StringUtils.collectionToDelimitedString(content, " ");

        Files.write(inputFile, inputData.getBytes(StandardCharsets.UTF_8));

        fileData.add(FileData.fromFile("t_input1.txt", DataType.SUBMODEL, inputFile.toFile()));
    }

    private void generateSimInputFile(ArrayList<FileData> fileData, SimulationEntity simulation) throws IOException {
        if (simulation.getManualSizeH() == null || simulation.getManualSizeB() == null || simulation.getManualSizeL() == null) {
            throw new TaskStartingException("All virtual simulation parameters must be present");
        }
        Path inputFile = Files.createTempFile(null, null);
        String inputData = commasToDots(simulation.getManualSizeB(), "0") + " " +
                commasToDots(simulation.getManualSizeH(), "0") + " " +
                commasToDots(simulation.getManualSizeL(), "0") + " " +
                commasToDots("1", "0");
        Files.write(inputFile, inputData.getBytes(StandardCharsets.UTF_8));
        fileData.add(FileData.fromFile("vi_input1.txt", DataType.SUBMODEL, inputFile.toFile()));
    }

    private SimulationEntity checkAndGetParentSim(TaskEntity task) {
        if (task.getParentType() != simulation) {
            throw new TaskStartingException("Unable to find parameters for virtual simulation");
        }
        SimulationEntity parent = simulationManager.findOne(task.getParentId(), null);
        if (parent == null) {
            throw new TaskStartingException("Unable to find virtual simulation");
        }
        return parent;
    }


    private void addCommonSimulationFiles(ArrayList<FileData> fileData, SimulationEntity parent) {
        ResultsParent resultParent = new ResultsParent(parent.getEcId(), TaskParentType.effectiveCharacteristic);
        List<ResultEntity> microstrParams = resultsManager.getResultByParent(resultParent, ResultType.microstrParams);
        if (microstrParams.isEmpty()) {
            throw new TaskStartingException("File with microstructure params is not found");
        }
        List<ResultEntity> mezostrParams = resultsManager.getResultByParent(resultParent, ResultType.mezostrParams);
        if (mezostrParams.isEmpty()) {
            throw new TaskStartingException("File with mezostructure params is not found");
        }
        List<ResultEntity> databases = resultsManager.getResultByParent(resultParent, ResultType.database);
        if (databases.size() != 2) {
            throw new BadRequestException("Database files is not found");
        }

        Path pathToMicrostrParams = resultsFilesManager.getPathToResult(microstrParams.get(0));
        fileData.add(FileData.fromFile(pathToMicrostrParams.getFileName().toString(), DataType.SUBMODEL, pathToMicrostrParams.toFile()));

        Path pathToMezostrParams = resultsFilesManager.getPathToResult(mezostrParams.get(0));
        fileData.add(FileData.fromFile(pathToMezostrParams.getFileName().toString(), DataType.SUBMODEL, pathToMezostrParams.toFile()));

        Path pathToDatabase1 = resultsFilesManager.getPathToResult(databases.get(0));
        fileData.add(FileData.fromFile(pathToDatabase1.getFileName().toString(), DataType.SUBMODEL, pathToDatabase1.toFile()));

        Path pathToDatabase2 = resultsFilesManager.getPathToResult(databases.get(1));
        fileData.add(FileData.fromFile(pathToDatabase2.getFileName().toString(), DataType.SUBMODEL, pathToDatabase2.toFile()));
    }
}
