package ru.crp.lvms.tasks.misc;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import ru.crp.lvms.cluster.ClusterDao;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.managers.NodesManager;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.managers.tasks.TaskStatusUpdater;
import ru.crp.lvms.model.dao.TaskDao;
import ru.crp.lvms.model.entities.tasks.TaskEntity;

import javax.annotation.Nullable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

abstract public class BaseTasksJob {
    private static final Logger LOGGER = LoggerFactory.getLogger(BaseTasksJob.class);

    @Autowired
    protected TaskManager taskManager;

    @Autowired
    protected TaskStatusUpdater taskStatusUpdater;

    @Autowired
    protected TaskDao taskDao;

    @Autowired
    protected NodesManager nodesManager;

    @Autowired
    protected TaskCommandsFactory taskCommandsFactory;

    @Autowired
    protected ClusterDao clusterDao;

    private Lock executionLock = new ReentrantLock();

    private ConcurrentMap<Integer, Integer> processingTasksIds = new ConcurrentHashMap<>();

    public void checkTaskQueueManually() {
        if (!executionLock.tryLock()) {
            return;
        }
        try {
            doCheckTaskQueue();
        } finally {
            executionLock.unlock();
        }
    }

    // Method should be called by asyncExecutor
    public void checkTaskQueue() {
        if (!executionLock.tryLock()) {
            return;
        }
        try {
            doCheckTaskQueue();
        } finally {
            executionLock.unlock();
        }
    }

    protected void doCheckTaskQueue() {
        List<TaskEntity> tasksToCheck = getTasksToCheck();
        Map<Integer, ClusterEntity> clustersCache = new HashMap<>();
        for (TaskEntity taskToCheck : tasksToCheck) {
            Integer taskId = taskToCheck.getId();
            // DO NOT move inside try block because in that case we remove taskId from processingTaskIds map even
            // if we didn't put it there
            if (processingTasksIds.putIfAbsent(taskId, taskId) != null) {
                LOGGER.warn("Task {} is processing by another method", taskId);
                continue;
            }
            try {
                ClusterEntity taskCluster = clustersCache.computeIfAbsent(taskToCheck.getClusterId(), clusterDao::findOne);
                ClusterCommandsManager clusterCommandManager = taskCommandsFactory.getClusterCommandManager(taskCluster.getType());
                checkTask(taskToCheck, taskCluster, clusterCommandManager);
            } catch (Exception ex) {
                LOGGER.error("Error while checking tasks queue", ex);
            } finally {
                processingTasksIds.remove(taskId);
            }
        }
    }

    @Nullable
    public Integer addTaskToProcessing(Integer taskId) {
        return processingTasksIds.putIfAbsent(taskId, taskId);
    }

    public void removeTaskFromProcessing(Integer taskId) {
        processingTasksIds.remove(taskId);
    }

    public abstract List<TaskEntity> getTasksToCheck();

    public abstract void checkTask(TaskEntity taskToCheck, ClusterEntity taskCluster, ClusterCommandsManager clusterCommandManager) throws Exception;
}
