package ru.crp.lvms.utils;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.transaction.support.TransactionSynchronizationAdapter;
import org.springframework.transaction.support.TransactionSynchronizationManager;

import javax.annotation.Nonnull;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.DecimalFormat;
import java.util.List;
import java.util.function.IntFunction;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static org.apache.commons.lang3.StringUtils.isEmpty;

public final class Utils {
    private final static String[] units = new String[]{"B", "KB", "MB", "GB", "TB"};

    private Utils() {
    }

    @Nonnull
    public static String toHex(byte[] bytes) {
        return new String(Hex.encodeHex(bytes));
    }

    @Nonnull
    public static String createMD5Hash(String source) {
        try {
            return toHex(createHash(new ByteArrayInputStream(source.getBytes("UTF-8")), "MD5"));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private static byte[] createHash(InputStream is, String hash) throws IOException {
        byte[] buffer = new byte[1024];
        MessageDigest complete;
        try {
            complete = MessageDigest.getInstance(hash);
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
        int numRead;
        try {
            do {
                numRead = is.read(buffer);
                if (numRead > 0) {
                    complete.update(buffer, 0, numRead);
                }
            } while (numRead != -1);
        } catch (IOException e) {
            is.close();
            throw e;
        } finally {
            is.close();
        }
        return complete.digest();
    }


    @Nonnull
    public static String createSHAHashString(InputStream is) throws IOException {
        return toHex(createHash(is, "SHA-256"));
    }


    public static long memoryInMegabytes(long memoryInBytes) {
        return Math.round(memoryInBytes / 1024 / 1024);
    }

    public static void registerTransactionCallback(Runnable func) {
        TransactionSynchronizationManager.registerSynchronization(
                new TransactionSynchronizationAdapter() {
                    @Override
                    public void afterCommit() {
                        func.run();
                    }
                }
        );
    }

    public static <T, U> Stream<T> mapWithIndex(List<U> elements, Integer startIndex, IntFunction<T> mapper) {
        return IntStream.range(startIndex, elements.size() + startIndex).mapToObj(mapper);
    }

    public static <U> Stream<Pair<Integer, U>> zipWithIndex(List<U> elements) {
        return zipWithIndex(elements, 0);
    }

    public static <U> Stream<Pair<Integer, U>> zipWithIndex(List<U> elements, Integer startIndex) {
        return IntStream.range(0, elements.size())
                .mapToObj(idx -> Pair.of(idx + startIndex, elements.get(idx)));
    }

    public static String humanReadableByteCount(long bytes) {
        if (bytes <= 0) return "0";
        int digitGroups = (int) (Math.log10(bytes) / Math.log10(1000));
        return new DecimalFormat("#,##0.##")
                .format(bytes / Math.pow(1000, digitGroups)) + " " + units[digitGroups];
    }

    public static String valueOrZero(String value) {
        return isEmpty(value) ? "0" : value;
    }
}
