package ru.crp.lvms.validators.task.optimization;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import org.springframework.validation.Errors;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.mf.entities.*;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.mf.enums.MFConstraintsType;
import ru.crp.lvms.optimizations.mf.enums.MFSettingsType;
import ru.crp.lvms.optimizations.mf.enums.MFTargetFuncsType;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.optimus.entities.*;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusConstraintsType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusSettingsType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusTargetFuncsType;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;

import java.util.List;
import java.util.Map;
import java.util.Objects;

import static ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType.*;
//import static ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType.*;


@Component
public class OptimusTaskParamsValidator extends BaseOptimizationTaskParamsValidator {

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Override
    protected void doValidate(TaskEntity task, OptimizationEntity taskParent, Errors errors) {
        if (!Objects.equals(taskParent.getSolverCaoId(), OptimizationCao.Optimus.getId())) {
            throw new IllegalArgumentException(String.format("Wrong validator is used for task with solver %s", OptimizationCao.valueOf(taskParent.getSolverCaoId())));
        }
        super.doValidate(task, taskParent, errors);
        if (errors.hasErrors()) {
            return;
        }
        validateCadFiles(taskParent, errors);
        if (errors.hasErrors()) {
            return;
        }
        Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings = optimusOptimizationSettingsManager
                .findBaseSettings(taskParent.getId(), null);
        validateBaseSettings(taskParent, baseSettings, errors);
        if (errors.hasErrors()) {
            return;
        }
        List<OptimusOptimizationTargetFunc> targetFuncs = (List<OptimusOptimizationTargetFunc>) optimusOptimizationSettingsManager
                .findOptimizationSettings(taskParent.getId(), OptimusSettingsType.TargetFunc, null);
        validateTargetFuncs(targetFuncs, baseSettings, errors);

        List<OptimusOptimizationVaryingParam> varyingParams = (List<OptimusOptimizationVaryingParam>) optimusOptimizationSettingsManager
                .findOptimizationSettings(taskParent.getId(), OptimusSettingsType.VaryingParam, null);
        validateVaryingParams(varyingParams, baseSettings, errors);

        List<? extends AbstractOptimizationSettings> cadVaryingParams = optimusOptimizationSettingsManager
                .findOptimizationSettings(taskParent.getId(), OptimusSettingsType.CADVaryingParam, null);
        if (!cadVaryingParams.isEmpty()) {
            validateCadVaryingParams((List<OptimusOptimizationCADVaryingParam>) cadVaryingParams, baseSettings, errors);
        }

        List<OptimusOptimizationConstraint> constraints = (List<OptimusOptimizationConstraint>) optimusOptimizationSettingsManager
                .findOptimizationSettings(taskParent.getId(), OptimusSettingsType.Constraint, null);
        validateConstraints(constraints, baseSettings, errors);

        validateDoe(taskParent, errors);
        validateAlgorithm(taskParent, errors);
    }

    private void validateCadFiles(OptimizationEntity taskParent, Errors errors) {
        if (taskParent.getSolverCadId() == null) {
            return;
        }
        List<FileEntity> cadFiles = optimizationManager.getOptimizationFiles(taskParent, FileType.CADFile, null);
        if (cadFiles.size() != 1) {
            errors.reject("wrongCADFilesCount");
        }
    }

    private void validateAlgorithm(OptimizationEntity taskParent, Errors errors) {
        Integer selectedAlgorithmId = optimusOptimizationSettingsManager.findSelectedAlgorithmId(taskParent.getId());
        if (selectedAlgorithmId == null) {
            errors.reject("algorithmNotFound");
            return;
        }
        ParametersDictionary selectedAlgorithm = optimusOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.OptimusOptimizationMethod, selectedAlgorithmId);
        if (selectedAlgorithm == null) {
            errors.reject("unexpectedAlgorithmId", new Object[]{selectedAlgorithmId}, null);
        }
    }

    private void validateDoe(OptimizationEntity taskParent, Errors errors) {
        Integer selectedDoeId = optimusOptimizationSettingsManager.findSelectedDoeId(taskParent.getId());
        if (selectedDoeId == null) {
            errors.reject("doeNotFound");
            return;
        }
        ParametersDictionary selectedAlgorithm = optimusOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.OptimusDoEType, selectedDoeId);
        if (selectedAlgorithm == null) {
            errors.reject("unexpectedDoeId", new Object[]{selectedDoeId}, null);
        }
    }

    public void validateConstraints(List<OptimusOptimizationConstraint> constraints, Map<OptimusBaseSettingType,
            OptimusOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedConstraintsCount = baseSettings.get(NUM_CONSTRAINTS).getCurrentValue();
        if (constraints.size() < expectedConstraintsCount) {
            errors.reject("notEnoughConstraints");
            return;
        }
        for (int i = 0; i < expectedConstraintsCount; ++i) {
            OptimusOptimizationConstraint constraintToCheck = constraints.get(i);
            if (constraintToCheck.getOptimizationId() == 0 || StringUtils.isEmpty(constraintToCheck.getName()) ||
                    constraintToCheck.getFuncId() == null || constraintToCheck.getFuncId() == OptimusConstraintsType.NONE.getId() || constraintToCheck.getValue() == null) {
                errors.reject("notEnoughConstraints");
                break;
            }
        }
    }

    private void validateVaryingParams(List<OptimusOptimizationVaryingParam> varyingParams,
                                       Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedVaryingParamsCount = baseSettings.get(NUM_CAO_VARYING_PARAMS).getCurrentValue();
        if (varyingParams.size() < expectedVaryingParamsCount) {
            errors.reject("notEnoughVaryingParams");
            return;
        }
        for (int i = 0; i < expectedVaryingParamsCount; ++i) {
            OptimusOptimizationVaryingParam paramToCheck = varyingParams.get(i);
            if (paramToCheck.getOptimizationId() == 0 || StringUtils.isEmpty(paramToCheck.getName()) ||
                    paramToCheck.getLowerBound() == null || paramToCheck.getUpperBound() == null
                    || paramToCheck.getLowerBound() > paramToCheck.getUpperBound() || paramToCheck.getSampling() == null
                    || paramToCheck.getSampling() < 0 || paramToCheck.getSampling() == 1
                    || !paramToCheck.getIsUsedInScript()) {
                errors.reject("notEnoughVaryingParams");
                break;
            }
        }
    }

    private void validateCadVaryingParams(List<OptimusOptimizationCADVaryingParam> cadVaryingParams,
                                          Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedCadVaryingParamsCount = baseSettings.get(NUM_CAD_VARYING_PARAMS).getCurrentValue();
        if (cadVaryingParams.size() < expectedCadVaryingParamsCount) {
            errors.reject("notEnoughCADVaryingParams");
            return;
        }
        for (int i = 0; i < expectedCadVaryingParamsCount; ++i) {
            OptimusOptimizationCADVaryingParam paramToCheck = cadVaryingParams.get(i);
            if (paramToCheck.getOptimizationId() == 0 || StringUtils.isEmpty(paramToCheck.getName()) ||
                    paramToCheck.getLowerBound() == null || paramToCheck.getUpperBound() == null
                    || paramToCheck.getLowerBound() > paramToCheck.getUpperBound() || paramToCheck.getSampling() == null
                    || paramToCheck.getSampling() < 0 || paramToCheck.getSampling() == 1) {
                errors.reject("notEnoughCADVaryingParams");
                break;
            }
        }
    }

    public void validateTargetFuncs(List<OptimusOptimizationTargetFunc> targetFuncs,
                                    Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedTargetFuncsCount = baseSettings.get(NUM_TARGET_FUNCS).getCurrentValue();
        if (targetFuncs.size() < expectedTargetFuncsCount) {
            errors.reject("notEnoughTargetFuncs");
            return;
        }
        for (int i = 0; i < expectedTargetFuncsCount; ++i) {
            OptimusOptimizationTargetFunc funcToCheck = targetFuncs.get(i);
            if (funcToCheck.getOptimizationId() == 0 || StringUtils.isEmpty(funcToCheck.getName())
                    || funcToCheck.getFuncId() == null || funcToCheck.getFuncId() == OptimusTargetFuncsType.NONE.getId()) {
                errors.reject("notEnoughTargetFuncs");
                break;
            }
        }
    }

    public void validateBaseSettings(OptimizationEntity taskParent, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings, Errors errors) {
//        if ((taskParent.getSolverCadId() == null && baseSettings.size() != MFBaseSettingType.values().length - 1) ||
//                taskParent.getSolverCadId() != null && baseSettings.size() != MFBaseSettingType.values().length) {
//            errors.reject("notEnoughBaseParams");
//            return;
//        }
        baseSettings.forEach((type, settings) -> {
            // if settings with this type is not used
            if (settings == null) {
                return;
            }
            Integer settingsValue = settings.getCurrentValue();
            if (settingsValue == null && ! (taskParent.getSolverCadId() == null && type == OptimusBaseSettingType.NUM_CAD_VARYING_PARAMS)) {
                errors.reject("baseParamValueIsNotSet", new Object[]{settings.getName()}, null);
            } else if (settingsValue > type.getMaxValue() || settingsValue < 0) {
                errors.reject("baseParamValueInvalid", new Object[]{settings.getName(), type.getMaxValue() + 1}, null);
            }
        });
    }
}
