package ru.crp.lvms.web;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.MessageSource;
import org.springframework.http.HttpStatus;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.validation.Errors;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;
import ru.crp.lvms.exceptions.*;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.optimizations.tosca.exceptions.PropertyProcessingException;
import ru.crp.lvms.web.dto.misc.ResponseMessage;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.IOException;

import static java.util.stream.Collectors.joining;


@ControllerAdvice
@ResponseBody
public class ExceptionHandlerAdvice {
    private static final String PATH_TO_DEFAULT_IMAGE = "images/default.png";
    @Value("${storage.root}")
    private String filesFolder;

    @Autowired
    private MessageSource messageSource;

    private static final Logger LOGGER = LoggerFactory.getLogger(ExceptionHandlerAdvice.class);


    @Nonnull
    @ExceptionHandler(NotFoundException.class)
    @ResponseStatus(HttpStatus.NOT_FOUND)
    public ResponseMessage handleNotFoundExceptions(NotFoundException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }


    @Nonnull
    @ExceptionHandler(HttpMediaTypeNotSupportedException.class)
    @ResponseStatus(HttpStatus.UNSUPPORTED_MEDIA_TYPE)
    public ResponseMessage handleNotSupportedMediaType(HttpMediaTypeNotSupportedException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }


    @Nonnull
    @ExceptionHandler(AccessDeniedException.class)
    @ResponseStatus(HttpStatus.FORBIDDEN)
    public ResponseMessage handleAccessDeniedExceptions(AccessDeniedException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError("Недостаточно прав для совершения данного действия");
    }


    @Nonnull
    @ExceptionHandler(HttpRequestMethodNotSupportedException.class)
    @ResponseStatus(HttpStatus.METHOD_NOT_ALLOWED)
    public ResponseMessage handleMethodNotSupportedExceptions(HttpRequestMethodNotSupportedException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }


    // JSR-303 Validation error
    @Nonnull
    @ExceptionHandler(MethodArgumentNotValidException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleMethodArgumentNotValidException(MethodArgumentNotValidException ex) {
        String message = createErrorMessage(ex.getBindingResult(), false);
        LOGGER.info(message, ex);
        return ResponseMessage.createError(message);
    }


    @Nonnull
    @ExceptionHandler(BadRequestException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleBadRequestException(BadRequestException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }


    @Nonnull
    @ExceptionHandler(MethodArgumentTypeMismatchException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleMethodArgumentTypeMismatchException(MethodArgumentTypeMismatchException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError("Неверный путь или недопустимые параметры запроса");
    }


    @Nonnull
    @ExceptionHandler(MissingServletRequestParameterException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleMissingServletRequestParameterException(MissingServletRequestParameterException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError("Недостаточно параметров запроса");
    }

    @Nonnull
    @ExceptionHandler(org.springframework.web.multipart.support.MissingServletRequestPartException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleMissingServletRequestParameterExceptionMultipart(org.springframework.web.multipart.support.MissingServletRequestPartException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError("Недостаточно параметров запроса");
    }


    @Nonnull
    @ExceptionHandler(HttpMessageNotReadableException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleMessageNotReadableException(HttpMessageNotReadableException ex) {
        LOGGER.info(ex.getMessage(), ex);
        return ResponseMessage.createError("Неверный формат запроса");
    }

    @Nonnull
    @ExceptionHandler(EntityDeleteException.class)
    public ResponseMessage handleEntityDeleteException(HttpServletResponse response, EntityDeleteException ex) {
        LOGGER.info(ex.getMessage(), ex);
        if (ex.isHasNotDeletedDependency()) {
            response.setStatus(HttpServletResponse.SC_CONFLICT);
        } else {
            response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        }
        return ResponseMessage.createError(ex.getMessage());
    }


    @Nonnull
    @ExceptionHandler(Exception.class)
    @ResponseStatus(HttpStatus.INTERNAL_SERVER_ERROR)
    public ResponseMessage handleAllExceptions(Exception ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError("Внутренняя ошибка сервера");
    }

    @Nonnull
    @ExceptionHandler(InternalException.class)
    @ResponseStatus(HttpStatus.INTERNAL_SERVER_ERROR)
    public ResponseMessage handleInternalExcepiton(InternalException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError("Внутренняя ошибка сервера: " + ex.getMessage());
    }


    @Nonnull
    @ExceptionHandler(ReportGenerationException.class)
    @ResponseStatus(HttpStatus.INTERNAL_SERVER_ERROR)
    public ResponseMessage handleReportGenerationException(ReportGenerationException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }


    // JSR-303 Validation error
    @Nonnull
    @ExceptionHandler(ValidationException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleValidationException(ValidationException ex) {
        String message = ex.getConstraintViolationMessages().stream()
                .collect(joining("\n", "Недопустимый запрос:\n", ""));
        LOGGER.info(message, ex);
        return ResponseMessage.createError(message);
    }

    @Nonnull
    @ExceptionHandler(TaskStartingException.class)
    @ResponseStatus(HttpStatus.BAD_REQUEST)
    public ResponseMessage handleTaskStartingException(TaskStartingException ex) {
        String message = ex.getMessage();
        if (ex.getErrors() != null) {
            message = createErrorMessage(ex.getErrors(), true);
        }
        LOGGER.info(message);
        return ResponseMessage.createError(message);
    }


    private String createErrorMessage(Errors errors, boolean useMessageCode) {
        return errors.getAllErrors().stream()
                .map(err -> useMessageCode ? messageSource.getMessage(err, null) : err.getDefaultMessage())
                .collect(joining("\n", "Недопустимый запрос:\n", ""));
    }

    @ExceptionHandler(ImageNotFoundException.class)
    @ResponseStatus(HttpStatus.OK)
    public void handleImageNotFoundException(HttpServletResponse response, ImageNotFoundException ex) throws IOException {
        File defaultImage = new File(filesFolder + PATH_TO_DEFAULT_IMAGE);
        if (!defaultImage.exists()) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
        response.setHeader("Content-Type", "image/" + FilenameUtils.getExtension(defaultImage.getAbsolutePath()));
        FileUtils.copyFile(defaultImage, response.getOutputStream());
    }

    @Nonnull
    @ExceptionHandler(PropertyProcessingException.class)
    @ResponseStatus(HttpStatus.INTERNAL_SERVER_ERROR)
    public ResponseMessage handlePropertyProcessingException(PropertyProcessingException ex) {
        LOGGER.error(ex.getMessage(), ex);
        return ResponseMessage.createError(ex.getMessage());
    }
}

