package ru.crp.lvms.web.controllers;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.ReportGenerationException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.StandardManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.TaskManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SimulationType;
import ru.crp.lvms.model.entities.simulations.SimulationTypeSuggest;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.pdf.reports.ReportGenerator;
import ru.crp.lvms.pdf.reports.SimulationReportInput;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.web.controllers.misc.ControllerImageNew;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.controllers.misc.FileHelper;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static ru.crp.lvms.utils.URLs.SIMULATIONS_URL;

@RestController
@RequestMapping(value = SIMULATIONS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class SimulationsController extends CrudControllerSuggest<SimulationEntity, SimulationEntity, SimulationEntity>
        implements ControllerImageNew<SimulationEntity> {

    @Autowired
    private EventManager eventManager;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private StandardManager standardManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ReportGenerator reportGenerator;

    @Autowired
    private TaskManager taskManager;

    @Override
    protected BaseManagerSuggest<SimulationEntity> getManager() {
        return simulationManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(method = RequestMethod.POST)
    public SimulationEntity createSimulation(HttpServletRequest request, @Valid @RequestBody SimulationEntity newSimulation,
                                             @CurrentUser PortalUser currentUser) {
        newSimulation.setStatusId(SolvingStatus.Created.getId());
        newSimulation.setOwnerId(currentUser.getUser().getId());
        if (newSimulation.getTypeId() == 0) {
            newSimulation.setTypeId(SimulationType.Extension.getId());
        }
        return simulationManager.createAndRead(newSimulation, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @Override
    public SimulationEntity update(HttpServletRequest request, @PathVariable("id") int id, @Valid @RequestBody SimulationEntity updateVO,
                                   @CurrentUser PortalUser user) {
        SimulationEntity existed = getManager().findOne(id, user.getUser());
        existed = checkEntityExists(id, existed);

        existed.setName(updateVO.getName());
        existed.setDescription(updateVO.getDescription());
        existed.setTypeId(updateVO.getTypeId());
        existed.setSampleFileId(updateVO.getSampleFileId());
        existed.setStandardId(updateVO.getStandardId());
        existed.setManualSizeH(updateVO.getManualSizeH());
        existed.setManualSizeB(updateVO.getManualSizeB());
        existed.setManualSizeL(updateVO.getManualSizeL());

        return getManager().updateAndRead(existed, user.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public SimulationEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }


    @RequestMapping(value = "/types/suggest", method = RequestMethod.POST)
    public List<SimulationTypeSuggest> suggestType(@RequestBody @Valid SuggestRequest suggestRequest,
                                                   @CurrentUser PortalUser currentUser) {
        return simulationManager.suggestSimulationType(suggestRequest, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/experiment", method = RequestMethod.GET)
    public void getExperimentalFile(HttpServletResponse response,
                                    @PathVariable("id") Integer entityId,
                                    @CurrentUser PortalUser currentUser) throws IOException {
        SimulationEntity entity = getManager().checkEntityExistsAndAccesible(entityId, AccessRights.view, currentUser.getUser());
        Integer fileId = entity.getExperimentFileId();
        if (fileId == null) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
        FileEntity file = fileManager.findOne(fileId, currentUser.getUser());
        if (file == null) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
        }
        writeFileToResponse(file, response, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{id}/experiment", method = RequestMethod.POST)
    public SimulationEntity updateExperimentalFile(HttpServletRequest request,
                                                   @PathVariable("id") Integer entityId,
                                                   @RequestParam("file") MultipartFile newFile,
                                                   @CurrentUser PortalUser currentUser) {
        SimulationEntity sim = simulationManager.checkEntityExistsAndAccesible(entityId, AccessRights.update, currentUser.getUser());
        Integer fileId = sim.getExperimentFileId();
        if (fileId == null) {
            FileEntity newFileEntity = fileManager.createFile(newFile, entityId, FileType.ExperimentalResultFile, currentUser.getUser());
            sim.setExperimentFileId(newFileEntity.getId());
            getManager().update(sim, currentUser.getUser());
        } else {
            fileManager.updateFileContent(fileId, newFile, currentUser.getUser());
        }
        return get(request, entityId, currentUser);
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/experiment", method = RequestMethod.DELETE)
    public SimulationEntity deleteExperimentalFile(HttpServletRequest request,
                                                   @PathVariable("id") Integer entityId,
                                                   @CurrentUser PortalUser currentUser) {
        SimulationEntity entity = getManager().checkEntityExistsAndAccesible(entityId, AccessRights.view, currentUser.getUser());
        Integer fileId = entity.getExperimentFileId();
        if (fileId != null) {
            entity.setExperimentFileId(null);
            getManager().update(entity, currentUser.getUser());
            fileManager.deleteFile(fileId, currentUser.getUser());
        }
        return get(request, entityId, currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/report", method = RequestMethod.GET)
    public void generateReport(@PathVariable("id") Integer id, @CurrentUser PortalUser currentUser,
                               HttpServletResponse response) throws IOException {
        SimulationEntity simulationEntity = getManager().findOne(id, currentUser.getUser());
        checkEntityExists(id, simulationEntity);
        if (simulationEntity.getStatus() != SolvingStatus.Completed) {
            throw new ReportGenerationException("Виртуальное испытание не проведено и нет результатов для отчета");
        }
        SimulationReportInput reportInfo = simulationManager.formSimulationReportInput(simulationEntity);
        Path report = reportGenerator.generateReport(reportInfo);
        FileHelper.writeFileToResponse(report, simulationEntity.getName() + "-report.pdf", response);
        Files.delete(report);
        Files.delete(reportInfo.getFiberPic());
        Files.delete(reportInfo.getMatrixPic());
    }

    @PreAuthorize("hasRole('ROLE_GUEST')")
    @RequestMapping(value = "/{id}/refreshResults", method = RequestMethod.GET)
    public void refreshResults(@PathVariable("id") int id,
                               @CurrentUser PortalUser currentUser) {
        ResultsParent resultsParent = Mapper.map(getManager().findOne(id, currentUser.getUser()), ResultsParent.class);
        taskManager.refreshResultsHandling(resultsParent, currentUser.getUser());
    }


    @Override
    public ImageHandler imageHandler() {
        return simulationManager;
    }

    @Override
    public SimulationEntity getEntity(HttpServletRequest request, Integer entityId, PortalUser currentUser) {
        return get(request, entityId, currentUser);
    }

    @Override
    public FileManager getFileManager() {
        return fileManager;
    }
}
