package ru.crp.lvms.web.controllers.materials;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.CompositeMaterialManager;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.Layer;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.URLs;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.controllers.misc.FileHelper;
import ru.crp.lvms.web.dto.CompositeMaterialIn;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.annotation.PreDestroy;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import java.io.IOException;
import java.util.List;

import static ru.crp.lvms.utils.URLs.COMPOSITE_MATERIALS_URL;


@RestController
@RequestMapping(value = COMPOSITE_MATERIALS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class CompositeMaterialController
        extends CrudControllerSuggest<CompositeMaterialEntity, CompositeMaterialIn, CompositeMaterialEntity>
        implements FileHelper {
    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ECManager ecManager;

    @Autowired
    private SimulationManager simulationManager;

    @Nonnull
    @Override
    public List<CompositeMaterialEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        List<CompositeMaterialEntity> all = super.getAll(request, currentUser);
        all.forEach(material -> extendMaterialIfNeed(request, material, currentUser));
        return all;
    }


    private void extendMaterialIfNeed(HttpServletRequest request, CompositeMaterialEntity material,
                                      PortalUser currentUser) {
//        boolean useExtended = ServletRequestUtils.getBooleanParameter(request, URLs.EXTENDED_PARAM_NAME, false);
//        if (useExtended) {
        compositeMaterialManager.extendMaterial(material, currentUser.getUser());
//        }
    }


    @Nonnull
    @Override
    public PageingResponse<CompositeMaterialEntity> page(HttpServletRequest request,
                                                         @RequestBody DataRequest dataRequest,
                                                         @CurrentUser PortalUser currentUser) {
        PageingResponse<CompositeMaterialEntity> page = super.page(request, dataRequest, currentUser);
        page.getContent().forEach(material -> extendMaterialIfNeed(request, material, currentUser));
        return page;
    }


    @Nonnull
    @Override
    public CompositeMaterialEntity get(HttpServletRequest request, @PathVariable("id") int id,
                                       @CurrentUser PortalUser currentUser) {
        CompositeMaterialEntity material = super.get(request, id, currentUser);
        extendMaterialIfNeed(request, material, currentUser);
        return material;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public CompositeMaterialEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public CompositeMaterialEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                 @Valid @RequestBody CompositeMaterialIn updateVO,
                                 @CurrentUser PortalUser user) {
        return super.update(request,id,updateVO,user);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(method = RequestMethod.POST)
    public CompositeMaterialEntity createCompositeMaterial(@Valid @RequestBody CompositeMaterialIn materialIn,
                                                           @CurrentUser PortalUser currentUser) {
        CompositeMaterialEntity newMaterial = Mapper.map(materialIn, entityClass);
        newMaterial.setOwnerId(currentUser.getUser().getId());
        return compositeMaterialManager.createAndRead(newMaterial, currentUser.getUser());
    }



    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/layers/order", method = RequestMethod.POST)
    public List<Layer> setLayers(@PathVariable("id") int materialId,
                                 @Valid @RequestBody List<Integer> newLayers,
                                 @CurrentUser PortalUser currentUser) {
        compositeMaterialManager.updateLayersOrder(materialId, newLayers, currentUser.getUser());
        return getLayers(materialId, currentUser);
    }


    @Nonnull
    @RequestMapping(value = "/{id}/layers", method = RequestMethod.GET)
    public List<Layer> getLayers(@PathVariable("id") int materialId,
                                 @CurrentUser PortalUser currentUser) {
        return compositeMaterialManager.getLayers(materialId, true, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/layers", method = RequestMethod.POST)
    @ResponseStatus(HttpStatus.OK)
    public void addLayer(@PathVariable("id") int materialId,
                         @Valid @RequestBody Layer newLayer,
                         @CurrentUser PortalUser currentUser) {
        compositeMaterialManager.addLayer(materialId, newLayer, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/layers/{layerId}", method = RequestMethod.DELETE)
    @ResponseStatus(HttpStatus.OK)
    public void deleteLayer(@PathVariable("id") int materialId,
                            @PathVariable("layerId") int layerId,
                            @CurrentUser PortalUser currentUser) {
        compositeMaterialManager.deleteLayer(materialId, layerId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/layers", method = RequestMethod.DELETE)
    public void deleteSeveralLayers(@PathVariable("id") Integer materialId,
                                    @RequestParam("ids") List<Integer> layerIds,
                                    @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(Layer.class, layerIds, (id) -> {
            deleteLayer(materialId, id, currentUser);
        });
    }


    // sample files endpoints
    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles", method = RequestMethod.GET)
    public List<FileEntity> getSampleFiles(@PathVariable("id") int materialId,
                                           @CurrentUser PortalUser currentUser) {
        return fileManager.getSampleFiles(materialId, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles/list", method = RequestMethod.POST)
    public PageImpl<FileEntity> getSampleFilesPage(@PathVariable("id") int materialId,
                                                   @RequestBody DataRequest dataRequest,
                                                   @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        return fileManager
                .getSampleFilesPage(materialId, dataRequest.getFilters(), pageable, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles/suggest", method = RequestMethod.POST)
    public List<FileEntity> suggestSampleFiles(HttpServletRequest request,
                                               @PathVariable("id") int materialId,
                                               @RequestBody @Valid SuggestRequest suggestRequest,
                                               @CurrentUser PortalUser currentUser) {
        return fileManager
                .suggestSampleFiles(materialId, suggestRequest, currentUser.getUser());
    }

    @Nonnull
    @RequestMapping(value = "/{materialId}" + URLs.EFFECTIVE_CHARACTERISTICS_URL + "/suggest", method = RequestMethod.POST)
    public List<ECEntity> suggestEC(@PathVariable("materialId") Integer materialId, @RequestBody @Valid SuggestRequest suggestRequest,
                                    @CurrentUser PortalUser currentUser) {
        suggestRequest.upsertQuery("compositeMaterialId", materialId.toString());
        return ecManager.suggest(suggestRequest, currentUser.getUser());
    }

    @Nonnull
    @RequestMapping(value = "/{materialId}" + URLs.SIMULATIONS_URL + "/suggest", method = RequestMethod.POST)
    public List<SimulationEntity> suggestSimulation(@PathVariable("id") Integer materialId, @RequestBody @Valid SuggestRequest suggestRequest,
                                                    @CurrentUser PortalUser currentUser) {
        suggestRequest.upsertQuery("compositeMaterialId", materialId.toString());
        return simulationManager.suggest(suggestRequest, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles/{sampleFileId}", method = RequestMethod.GET)
    public FileEntity getSampleFile(@PathVariable("id") int materialId,
                                    @PathVariable("sampleFileId") int sampleFileId,
                                    @CurrentUser PortalUser currentUser) {
        return fileManager.getSampleFile(materialId, sampleFileId, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/sampleFiles/{sampleFileId}/download", method = RequestMethod.GET)
    public void downloadSampleFile(HttpServletResponse response,
                                   @PathVariable("id") int materialId,
                                   @PathVariable("sampleFileId") int sampleFileId,
                                   @CurrentUser PortalUser currentUser) throws IOException {
        FileEntity sampleFile = fileManager.getSampleFile(materialId, sampleFileId, currentUser.getUser());
        writeFileToResponse(sampleFile, response, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles", method = RequestMethod.POST,
            consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public FileEntity createSampleFile(@PathVariable("id") int materialId,
                                       @RequestParam("file") MultipartFile sampleFile,
                                       @CurrentUser PortalUser currentUser) {
        return fileManager.createSampleFile(true, materialId, sampleFile, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles/{sampleFileId}", method = RequestMethod.PUT)
    public FileEntity updateSampleFile(@PathVariable("id") int materialId,
                                       @PathVariable("sampleFileId") int sampleFileId,
                                       @Valid @RequestBody FileEntity updatedFile,
                                       @CurrentUser PortalUser currentUser) {
        updatedFile.setId(sampleFileId);
        updatedFile.setFileTypeId(FileType.SampleFile.getId());
        return fileManager.updateMaterialFile(true, materialId, updatedFile, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/sampleFiles/{sampleFileId}", method = RequestMethod.DELETE)
    public FileEntity deleteSampleFile(@PathVariable("id") int materialId,
                                       @PathVariable("sampleFileId") int sampleFileId,
                                       @CurrentUser PortalUser currentUser) {
        return fileManager.deleteSampleFile(materialId, sampleFileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/sampleFiles", method = RequestMethod.DELETE)
    public void deleteSeveralSampleFiles(@PathVariable("id") Integer materialId,
                                         @RequestParam("ids") List<Integer> sampleFileIds,
                                         @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, sampleFileIds, (id) -> {
            deleteSampleFile(materialId, id, currentUser);
        });
    }


    @Override
    protected BaseManagerSuggest<CompositeMaterialEntity> getManager() {
        return compositeMaterialManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }


    @Override
    public FileManager getFileManager() {
        return fileManager;
    }
}
