package ru.crp.lvms.web.controllers.misc;

import com.fasterxml.classmate.TypeResolver;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.modelmapper.TypeToken;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.util.Assert;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.misc.BaseManager;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.MappingHelpers;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import static ru.crp.lvms.exceptions.BadRequestException.createDeleteMultipleException;

public abstract class CrudController<E extends Entity, InVO, OutVO> {

    protected abstract BaseManager<E> getManager();

    protected abstract EventManager getEventsManager();

    protected Class<E> entityClass;
    protected Class<InVO> inVoClass;
    protected Class<OutVO> outVoClass;
    protected TypeToken<List<OutVO>> listOutVOs;

    @PostConstruct
    public void init() {
        entityClass = (Class<E>) new TypeResolver()
                .resolve(getClass()).typeParametersFor(CrudController.class).get(0).getErasedType();
        Assert.notNull(entityClass);
        inVoClass = (Class<InVO>) new TypeResolver()
                .resolve(getClass()).typeParametersFor(CrudController.class).get(1).getErasedType();
        Assert.notNull(inVoClass);
        outVoClass = (Class<OutVO>) new TypeResolver()
                .resolve(getClass()).typeParametersFor(CrudController.class).get(2).getErasedType();
        Assert.notNull(outVoClass);
        listOutVOs = MappingHelpers.createListOfType(outVoClass);
        Assert.notNull(listOutVOs);
    }


    @NotNull
    @RequestMapping(method = RequestMethod.GET)
    public List<OutVO> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        return Mapper.map(getManager().findAll(currentUser.getUser()), listOutVOs);
    }


    @NotNull
    @RequestMapping(value = "/list", method = RequestMethod.POST)
    public PageingResponse<OutVO> page(HttpServletRequest request, @RequestBody DataRequest dataRequest,
                                       @CurrentUser PortalUser currentUser) {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<E> pageImpl = getManager().find(dataRequest.getFilters(), pageable, currentUser.getUser());
        List<OutVO> newContent = Mapper.map(pageImpl.getContent(), listOutVOs);
        return new PageingResponse<>(newContent, dataRequest.getFilters(), pageable, pageImpl.getTotalElements());
    }


    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.GET)
    public OutVO get(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        E entity = getManager().findOne(id, currentUser.getUser());
        if (entity == null) {
            throw new NotFoundException(id, entityClass);
        }
        return Mapper.map(entity, outVoClass);
    }

    /*
     * Batch deletion is supported by following entities:
     *
     * <Entity name>                            <URL>
     * OptimizationFile                         OPTIMIZATION_FILES_URL
     * OptimizationReport                       OPTIMIZATION_REPORTS_URL
     * Result                                   RESULTS_URL
     * TreeNode                                 TREE_URL/{treeType}
     * CADFile                                  CONSTRUCTION_ELEMENTS_URL/{id}/CADFiles
     * CAEFile                                  CONSTRUCTION_ELEMENTS_URL/{id}/CAEFiles
     * CAOFile                                  CONSTRUCTION_ELEMENTS_URL/{id}/CAOFiles
     * DesignSolution                           CONSTRUCTION_ELEMENTS_URL/{id}/designSolutions
     * LoadCondition                            CONSTRUCTION_ELEMENTS_URL/{id}/loadConditions
     * ConstructionElementDoc                   CONSTRUCTION_ELEMENTS_URL/descriptionFiles
     * AdjacentElementsEntityTable              CONSTRUCTION_ELEMENTS_URL/adjacents
     * ECDefect                                 EFFECTIVE_CHARACTERISTICS_URL/{ecId}/defects        (forbidden)
     * ECParam                                  EFFECTIVE_CHARACTERISTICS_URL/{ecId}/parameters     (forbidden)
     * Layer                                    COMPOSITE_MATERIALS_URL/{id}/layers
     * SampleFile                               COMPOSITE_MATERIALS_URL/{id}/sampleFiles
     * MaterialFile                             MATERIAL_FILES_URL
     * MaterialFile                             COMPOSITE_MATERIAL_FILES_URL
     * Equipment                                EQUIPMENT_URL
     * User                                     USERS_URL
     */
    public static void deleteSeveralEntities(Class entityClass, List<Integer> ids, Consumer<Integer> deleteOneEntity) throws BadRequestException {
        Map<Integer, Exception> errors = new HashMap<>();
        for (Integer id : ids) {
            try {
                deleteOneEntity.accept(id);
            } catch (Exception e) {
                errors.put(id, e);
            }
        }
        if (!errors.isEmpty()) {
            throw createDeleteMultipleException(entityClass, errors);
        }
    }


    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public OutVO delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        E entity = getManager().findOne(id, currentUser.getUser());
        if (entity == null) {
            throw new NotFoundException(id, entityClass);
        }
        // use this version of delete method (instead of deleting by id) to be able to return deleted entity to client
        getManager().delete(entity, currentUser.getUser());
        return Mapper.map(entity, outVoClass);
    }

    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(entityClass, ids, (id) -> {
            delete(request, id, currentUser);
        });
    }

    protected E checkEntityExists(int id, @Nullable E entity) {
        if (entity == null) {
            throw new NotFoundException(id, entityClass);
        }
        return entity;
    }

    protected E checkEntityExists(int id, PortalUser currentUser) {
        E entity = getManager().findOne(id, currentUser.getUser());
        if (entity == null) {
            throw new NotFoundException(id, entityClass);
        }
        return entity;
    }


    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public OutVO update(HttpServletRequest request, @PathVariable("id") int id,
                        @Valid @RequestBody InVO updateVO,
                        @CurrentUser PortalUser user) {
        E updateEntity = getManager().findOne(id, user.getUser());
        checkEntityExists(id, updateEntity);
        Mapper.map(updateVO, updateEntity);
        updateEntity.setId(id);
        updateEntity = getManager().updateAndRead(updateEntity, user.getUser());
        return Mapper.map(updateEntity, outVoClass);
    }

    public Class<OutVO> getOutVoClass() {
        return outVoClass;
    }

    @Nonnull
    public TypeToken<List<OutVO>> getListOutVOs() {
        return listOutVOs;
    }
}
