package ru.crp.lvms.web.controllers.technology;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.ValidationException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.TechnologyManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.model.SuggestDTO;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.model.entities.technology.TechnologyType;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.ControllerImageNew;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.controllers.misc.FileHelper;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.ConstraintViolation;
import javax.validation.Valid;
import javax.validation.Validator;
import java.io.IOException;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.URLs.TECHNOLOGY_URL;

@RestController
@RequestMapping(value = TECHNOLOGY_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class TechnologyController
        extends CrudControllerSuggest<TechnologyEntity, TechnologyEntity, TechnologyEntity>
        implements FileHelper, ControllerImageNew {
    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private Validator validator;

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.APPLICATION_JSON_VALUE)
    public TechnologyEntity createTechnology(@RequestBody @Valid TechnologyEntity newTechnology,
                                             @CurrentUser PortalUser currentUser) {
        return technologyManager.createTechnology(newTechnology, null, null, currentUser.getUser());
    }


    @RequestMapping(value = "/types", method = RequestMethod.GET)
    public List<SuggestDTO> getTechnologyTypes(@CurrentUser PortalUser currentUser) {
        return Stream.of(TechnologyType.values())
                .map(SuggestDTO::new)
                .sorted(Comparator.comparing(SuggestDTO::getText))
                .collect(toList());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public TechnologyEntity createTechnology(
            @RequestParam("name") String name,
            @RequestParam("technologyType") String technologyType,
            @RequestParam(value = "videoUrl", required = false) String videoUrl,
            @RequestParam(value = "description", required = false) String description,
            @RequestParam(value = "imageFile", required = false) MultipartFile image,
            @RequestParam(value = "descriptionFile", required = false) MultipartFile descriptionFile,
            @CurrentUser PortalUser currentUser) {
        TechnologyEntity newTechnology = new TechnologyEntity();
        newTechnology.setName(name);
        newTechnology.setVideoUrl(videoUrl);
        newTechnology.setDescription(description);
        newTechnology.setType(TechnologyType.valueOf(technologyType));
        Set<ConstraintViolation<TechnologyEntity>> constraintViolations = validator.validate(newTechnology);
        if (!constraintViolations.isEmpty()) {
            throw new ValidationException(ValidationException.extractMessages(constraintViolations));
        }
        return technologyManager.createTechnology(newTechnology, image, descriptionFile, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public TechnologyEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public TechnologyEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                          @Valid @RequestBody TechnologyEntity updateVO,
                                          @CurrentUser PortalUser user) {
        return super.update(request,id,updateVO,user);
    }

    // working with description file
    @RequestMapping(value = "/{id}/descriptionFile", method = RequestMethod.GET)
    public void getDescriptionFile(HttpServletResponse response,
                                   @PathVariable("id") Integer technologyId,
                                   @CurrentUser PortalUser currentUser) throws IOException {
        FileEntity file = technologyManager.checkAndGetDescriptionFile(technologyId, currentUser.getUser());
        writeFileToResponse(file, response, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/descriptionFile", method = RequestMethod.POST)
    public TechnologyEntity updateDescriptionFile(HttpServletRequest request,
                                                  @PathVariable("id") Integer technologyId,
                                                  @RequestParam("file") MultipartFile descriptionFile,
                                                  @CurrentUser PortalUser currentUser) {
        technologyManager.updateDescriptionFile(technologyId, descriptionFile, currentUser.getUser());
        return get(request, technologyId, currentUser);
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/descriptionFile", method = RequestMethod.DELETE)
    public TechnologyEntity deleteDescriptionFile(HttpServletRequest request,
                                                  @PathVariable("id") Integer technologyId,
                                                  @CurrentUser PortalUser currentUser) {
        technologyManager.deleteDescriptionFile(technologyId, currentUser.getUser());
        return get(request, technologyId, currentUser);
    }


    @RequestMapping(value = "/{id}/simulations", method = RequestMethod.GET)
    public List<SimulationEntity> getOptimizations(@PathVariable("id") Integer technologyId,
                                                   @CurrentUser PortalUser currentUser) throws IOException {
        return technologyManager.getSimulationssForTechnology(technologyId, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/simulations/list", method = RequestMethod.POST)
    public PageingResponse<SimulationEntity> getOptimizationsPage(@PathVariable("id") Integer technologyId,
                                                                  @RequestBody DataRequest dataRequest,
                                                                  @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<SimulationEntity> page = technologyManager.getSimulationsPageForTechnology(technologyId, pageable,
                dataRequest.getFilters(), currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }

    @Override
    protected BaseManagerSuggest<TechnologyEntity> getManager() {
        return technologyManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }


    @Override
    public FileManager getFileManager() {
        return fileManager;
    }


    @Override
    public ImageHandler imageHandler() {
        return technologyManager;
    }


    @Override
    public Object getEntity(HttpServletRequest request, Integer entityId, PortalUser currentUser) {
        return get(request, entityId, currentUser);
    }
}
