package ru.crp.lvms.web.controllers.technology;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.HandlerMapping;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.ValidationException;
import ru.crp.lvms.managers.PossibleDefectsManager;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.model.dao.DefectNameDao;
import ru.crp.lvms.model.entities.PossibleDefectsEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.CrudControllerColumns;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.servlet.http.HttpServletRequest;
import javax.validation.ConstraintViolation;
import javax.validation.Valid;
import javax.validation.Validator;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static ru.crp.lvms.utils.URLs.POSSIBLE_DEFECTS_URL;

@RestController
@RequestMapping(value = POSSIBLE_DEFECTS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class TechnologyPossibleDefectsController
        extends CrudControllerColumns<PossibleDefectsEntity, PossibleDefectsEntity, PossibleDefectsEntity> {
    @Autowired
    private PossibleDefectsManager possibleDefectsManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private DefectNameDao defectNameDao;

    @Autowired
    private Validator validator;


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public PossibleDefectsEntity createPossibleDefect(@PathVariable("technologyId") Integer technologyId,
                                                      @RequestParam("value") String value,
                                                      @Nullable @RequestParam(value = "actName", required = false) String actName,
                                                      @RequestParam("defectNameId") Integer defectNameId,
                                                      @Nullable @RequestParam(required = false) MultipartFile file,
                                                      @CurrentUser PortalUser currentUser) throws UnsupportedEncodingException {
        if (actName == null && file == null) {
            throw new BadRequestException("Имя акта или файл должен быть задан");
        }
        if (actName == null || actName.trim().isEmpty()) {
            actName = (URLDecoder.decode(file.getOriginalFilename(), "UTF-8"));
        }
        PossibleDefectsEntity newDefect = new PossibleDefectsEntity();
        newDefect.setTechnologyId(technologyId);
        newDefect.setActName(actName);
        newDefect.setValue(value);
        newDefect.setDefectNameId(defectNameId);
        Set<ConstraintViolation<PossibleDefectsEntity>> constraints = validator.validate(newDefect);
        if (!constraints.isEmpty()) {
            throw new ValidationException(ValidationException.extractMessages(constraints));
        }
        return getManager().createAndRead(newDefect, file, currentUser.getUser());
    }


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST)
    public PossibleDefectsEntity createPossibleDefect(@PathVariable("technologyId") Integer technologyId,
                                                      @Valid @RequestBody PossibleDefectsEntity in,
                                                      @CurrentUser PortalUser currentUser) {
        in.setTechnologyId(technologyId);
        return getManager().createAndRead(in, null, currentUser.getUser());
    }


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    public PossibleDefectsEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                        @Valid @RequestBody PossibleDefectsEntity updateVO,
                                        @CurrentUser PortalUser user) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        updateVO.setTechnologyId(technologyId);
        return super.update(request, id, updateVO, user);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public PossibleDefectsEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @Nonnull
    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<PossibleDefectsEntity> suggest(@PathVariable("technologyId") Integer technologyId,
                                               @RequestBody @Valid SuggestRequest suggestRequest,
                                               @CurrentUser PortalUser currentUser) {
        suggestRequest.upsertQuery("technologyId", technologyId.toString());
        return getManager().suggest(suggestRequest, currentUser.getUser());
    }


    @Nonnull
    @Override
    public List<PossibleDefectsEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        return getManager().getTechnologyEquipment(technologyId, currentUser.getUser());
    }


    @Nonnull
    @Override
    public PageingResponse<PossibleDefectsEntity> page(HttpServletRequest request,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<PossibleDefectsEntity> page = getManager()
                .getTechnologyDefectsPage(technologyId, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @Override
    protected PossibleDefectsManager getManager() {
        return possibleDefectsManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }
}
