CREATE TABLE optimus_optimization_base_setting_types (
  id SERIAL,
  max_value INTEGER NOT NULL,
  display_name VARCHAR(255),

  PRIMARY KEY (id)
);

INSERT INTO optimus_optimization_base_setting_types(max_value, display_name) VALUES
 (2, 'Количество целевых функций'),
 (8, 'Количество варьируемых параметров'),
 (8, 'Количество выходных параметров'),
 (4, 'Количество ограничений'),
 (8, 'Количество варьируемых CAD параметров');

ALTER TABLE optimus_optimization_base_setting_types ADD COLUMN min_value INTEGER NOT NULL DEFAULT 1;

UPDATE optimus_optimization_base_setting_types SET max_value = 100 WHERE id = 2; -- num CAO params
UPDATE optimus_optimization_base_setting_types SET max_value = 100 WHERE id = 5; -- num CAD params


CREATE TABLE optimus_optimization_base_settings (
  id SERIAL,
  type_id INTEGER NOT NULL,
  optimization_id INTEGER NOT NULL,
  current_value INTEGER,

  PRIMARY KEY(id),
  FOREIGN KEY(optimization_id) REFERENCES optimizations(id) ON DELETE CASCADE,
  FOREIGN KEY(type_id) REFERENCES optimus_optimization_base_setting_types(id)
);
CREATE INDEX ON optimus_optimization_base_settings(optimization_id);

CREATE VIEW optimus_optimization_base_settings_join AS (
  SELECT
    obs.*,
    o.optimization_type_id AS optimization_type_id,
    obst.max_value         AS max_value,
    obst.min_value    AS min_value,
    obst.display_name      AS name

  FROM optimus_optimization_base_settings obs
    JOIN optimizations o ON obs.optimization_id = o.id
    JOIN optimus_optimization_base_setting_types obst ON obs.type_id = obst.id
);

CREATE RULE optimus_optimization_base_settings_join_CREATE AS ON INSERT TO optimus_optimization_base_settings_join DO INSTEAD (
  INSERT INTO optimus_optimization_base_settings (type_id, optimization_id, current_value)
  VALUES (NEW.type_id, NEW.optimization_id, NEW.current_value)
  RETURNING
    optimus_optimization_base_settings.*,
    NULL :: INTEGER AS optimization_type_id,
    NULL :: INTEGER AS max_value,
    NULL :: INTEGER AS min_value,
    NULL :: VARCHAR(255) AS name
);

CREATE RULE optimus_optimization_base_settings_join_UPDATE AS ON UPDATE TO optimus_optimization_base_settings_join DO INSTEAD (
  UPDATE optimus_optimization_base_settings
  SET current_value = NEW.current_value
  WHERE id = NEW.id);

CREATE RULE optimus_optimization_base_settings_join_DELETE AS ON DELETE TO optimus_optimization_base_settings_join DO INSTEAD (
  DELETE FROM optimus_optimization_base_settings
  WHERE id = OLD.id
);

-- Optimus CAO varying params
CREATE TABLE optimus_optimization_varying_params (
  id SERIAL,
  optimization_id INTEGER NOT NULL,
  is_used_in_script BOOLEAN NOT NULL DEFAULT FALSE,
  name VARCHAR(255),
  lower_bound DOUBLE PRECISION,
  upper_bound DOUBLE PRECISION,
  sampling INTEGER,

  PRIMARY KEY(id),
  FOREIGN KEY(optimization_id) REFERENCES optimizations(id) ON DELETE CASCADE
);
CREATE INDEX ON optimus_optimization_varying_params(optimization_id);


CREATE VIEW optimus_optimization_varying_params_join AS (
  SELECT
    ocvp.*,
    o.optimization_type_id AS optimization_type_id
  FROM optimus_optimization_varying_params ocvp
    JOIN optimizations o ON ocvp.optimization_id = o.id
);

CREATE RULE optimus_optimization_varying_params_join_CREATE AS ON INSERT TO optimus_optimization_varying_params_join DO INSTEAD (
  INSERT INTO optimus_optimization_varying_params (optimization_id, name, lower_bound, upper_bound, sampling, is_used_in_script)
  VALUES (NEW.optimization_id, NEW.name, NEW.lower_bound, NEW.upper_bound, NEW.sampling, FALSE)
  RETURNING
    optimus_optimization_varying_params.*,
    NULL :: INTEGER AS optimization_type_id
);


CREATE RULE optimus_optimization_varying_params_join_UPDATE AS ON UPDATE TO optimus_optimization_varying_params_join DO INSTEAD (
  UPDATE optimus_optimization_varying_params
  SET name            = NEW.name,
    lower_bound       = NEW.lower_bound,
    upper_bound       = NEW.upper_bound,
    sampling          = NEW.sampling,
    is_used_in_script = NEW.is_used_in_script
  WHERE id = NEW.id);

CREATE RULE optimus_optimization_varying_params_join_DELETE AS ON DELETE TO optimus_optimization_varying_params_join DO INSTEAD (
  DELETE FROM optimus_optimization_varying_params
  WHERE id = OLD.id
);

-- Optimus CAD varying params
CREATE TABLE optimus_optimization_cad_varying_params (
  id                SERIAL,
  optimization_id   INTEGER NOT NULL,
  name              VARCHAR(255),
  lower_bound       DOUBLE PRECISION,
  upper_bound       DOUBLE PRECISION,
  sampling          INTEGER,

  PRIMARY KEY (id),
  FOREIGN KEY (optimization_id) REFERENCES optimizations (id) ON DELETE CASCADE
);
CREATE INDEX ON optimus_optimization_cad_varying_params (optimization_id);

CREATE VIEW optimus_optimization_cad_varying_params_join AS (
  SELECT
    ocvp.*,
    o.optimization_type_id AS optimization_type_id
  FROM optimus_optimization_cad_varying_params ocvp
    JOIN optimizations o ON ocvp.optimization_id = o.id
);

CREATE RULE optimus_optimization_cad_varying_params_join_CREATE AS ON INSERT TO optimus_optimization_cad_varying_params_join DO INSTEAD (
  INSERT INTO optimus_optimization_cad_varying_params (optimization_id, name, lower_bound, upper_bound, sampling)
  VALUES (NEW.optimization_id, NEW.name, NEW.lower_bound, NEW.upper_bound, NEW.sampling)
  RETURNING
    optimus_optimization_cad_varying_params.*,
    NULL :: INTEGER AS optimization_type_id
);


CREATE RULE optimus_optimization_cad_varying_params_join_UPDATE AS ON UPDATE TO optimus_optimization_cad_varying_params_join DO INSTEAD (
  UPDATE optimus_optimization_cad_varying_params
  SET name      = NEW.name,
    lower_bound = NEW.lower_bound,
    upper_bound = NEW.upper_bound,
    sampling    = NEW.sampling
  WHERE id = NEW.id);

CREATE RULE optimus_optimization_cad_varying_params_join_DELETE AS ON DELETE TO optimus_optimization_cad_varying_params_join DO INSTEAD (
  DELETE FROM optimus_optimization_cad_varying_params
  WHERE id = OLD.id
);

-- Optimus target functions
CREATE TABLE optimus_optimization_target_funcs (
  id SERIAL,
  optimization_id INTEGER NOT NULL,
  func_id INTEGER NOT NULL DEFAULT 1,
  name VARCHAR(255),

  PRIMARY KEY(id),
  FOREIGN KEY(optimization_id) REFERENCES optimizations(id) ON DELETE CASCADE
);
CREATE INDEX ON optimus_optimization_target_funcs(optimization_id);

CREATE VIEW optimus_optimization_target_funcs_join AS (
  SELECT
    otf.*,
    o.optimization_type_id AS optimization_type_id,
    pd.name                AS func_name
  FROM optimus_optimization_target_funcs otf
    JOIN optimizations o ON otf.optimization_id = o.id
    LEFT JOIN parameters_dictionary pd ON pd.id = otf.func_id
);

CREATE RULE optimus_optimization_target_funcs_join_CREATE AS ON INSERT TO optimus_optimization_target_funcs_join DO INSTEAD (
  INSERT INTO optimus_optimization_target_funcs (optimization_id, name, func_id)
  VALUES (NEW.optimization_id, NEW.name, NEW.func_id)
  RETURNING
    optimus_optimization_target_funcs.*,
    NULL :: INTEGER AS optimization_type_id,
    NULL :: VARCHAR(255) AS func_name
);

CREATE RULE optimus_optimization_target_funcs_join_UPDATE AS ON UPDATE TO optimus_optimization_target_funcs_join DO INSTEAD (
  UPDATE optimus_optimization_target_funcs
  SET name  = NEW.name,
    func_id = NEW.func_id
  WHERE id = NEW.id);

CREATE RULE optimus_optimization_target_funcs_join_DELETE AS ON DELETE TO optimus_optimization_target_funcs_join DO INSTEAD (
  DELETE FROM optimus_optimization_target_funcs
  WHERE id = OLD.id
);

-- Optimus constraints
CREATE TABLE optimus_optimization_constrains (
  id SERIAL,
  optimization_id INTEGER NOT NULL,
  name VARCHAR(255),
  func_id INTEGER NOT NULL DEFAULT 4,
  val DOUBLE PRECISION,

  PRIMARY KEY(id),
  FOREIGN KEY(optimization_id) REFERENCES optimizations(id) ON DELETE CASCADE
);
CREATE INDEX ON optimus_optimization_constrains(optimization_id);

CREATE VIEW optimus_optimization_constrains_join AS (
  SELECT
    oc.*,
    o.optimization_type_id AS optimization_type_id,
    pd.name                AS func_name
  FROM optimus_optimization_constrains oc
    JOIN optimizations o ON oc.optimization_id = o.id
    LEFT JOIN parameters_dictionary pd ON pd.id = oc.func_id
);

CREATE RULE optimus_optimization_constrains_join_CREATE AS ON INSERT TO optimus_optimization_constrains_join DO INSTEAD (
  INSERT INTO optimus_optimization_constrains (optimization_id, name, func_id, val)
  VALUES (NEW.optimization_id, NEW.name, NEW.func_id, NEW.val)
  RETURNING
    optimus_optimization_constrains.*,
    NULL :: INTEGER AS optimization_type_id,
    NULL :: VARCHAR(255) AS func_name
);

CREATE RULE optimus_optimization_constrains_join_UPDATE AS ON UPDATE TO optimus_optimization_constrains_join DO INSTEAD (
  UPDATE optimus_optimization_constrains
  SET name  = NEW.name,
    func_id = NEW.func_id,
    val     = NEW.val
  WHERE id = NEW.id);

CREATE RULE optimus_optimization_constrains_join_DELETE AS ON DELETE TO optimus_optimization_constrains_join DO INSTEAD (
  DELETE FROM optimus_optimization_constrains
  WHERE id = OLD.id
);

INSERT INTO caos (id, name, unique_name) VALUES
(4, 'Optimus', 'OPTIMUS');

---------------------------------------------------------------------------------------------

INSERT INTO optimization_type_solver_combinations(optimization_type_id, cao_id, cae_id, cad_id, available) VALUES
-- Параметрическая оптимизации формы
  (4, 4, NULL, NULL, FALSE),
  (4, 4, 1, NULL, TRUE), (4, 4, 3, NULL, FALSE),
  (4, 4, NULL, 1, FALSE),
  (4, 4, 1, 1, FALSE), (4, 4, 3, 1, FALSE),
-- Топологическая оптимизация микроструктуры
  (5, 4, NULL, NULL, FALSE),
  (5, 4, 1, NULL, TRUE), (5, 4, 2, NULL, FALSE), (5, 4, 3, NULL, FALSE),
-- Параметрическая оптимизация микроструктуры
  (6, 4, NULL, NULL, FALSE),
  (6, 4, 1, NULL, TRUE), (6, 4, 2, NULL, TRUE), (6, 4, 3, NULL, FALSE);

INSERT INTO file_types VALUES(16, 'OptimusScript');

INSERT INTO parameters_dictionary_type(id, name) VALUES
(7, 'OptimusOptimizationMethod'),
(8, 'OptimusDoEType');

INSERT INTO parameters_dictionary(id, name, type_id) VALUES
(20,'LATIN',8),
(21,'RANDOM',8),
(22,'EGO',7),
(23,'SIMAN',7),
(24,'DOE',7);


