DROP TABLE optimization_results;

CREATE TABLE results (
  id              SERIAL,
  parent_type VARCHAR(255) NOT NULL,
  ec_id INTEGER,
  sim_id INTEGER,
  optimization_id INTEGER,
  created_by      CHARACTER VARYING(255),
  type            CHARACTER VARYING(255)                    NOT NULL,
  name            CHARACTER VARYING(255)                    NOT NULL,
  path_to_file    CHARACTER VARYING,
  description     CHARACTER VARYING,
  date_created    TIMESTAMP WITHOUT TIME ZONE DEFAULT now() NOT NULL,
  date_modified   TIMESTAMP WITHOUT TIME ZONE DEFAULT now() NOT NULL,
  owner_id        INTEGER                                   NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (owner_id) REFERENCES users (id),
  FOREIGN KEY (ec_id) REFERENCES effective_characteristics(id),
  FOREIGN KEY (sim_id) REFERENCES virtual_simulations(id),
  FOREIGN KEY (optimization_id) REFERENCES optimizations(id),

  CONSTRAINT check_parent_type CHECK (parent_type = 'EffectiveCharacteristic' OR parent_type = 'VirtualSimulation' OR parent_type = 'Optimization'),
  CONSTRAINT check_parent_id CHECK
    ((parent_type = 'EffectiveCharacteristic' AND ec_id IS NOT NULL) OR (parent_type = 'VirtualSimulation' AND sim_id IS NOT NULL)
    OR (parent_type = 'Optimization' AND optimization_id IS NOT NULL))
);

CREATE TRIGGER results_update
  BEFORE UPDATE ON results
    FOR EACH ROW
      WHEN (OLD.* IS DISTINCT FROM NEW.*)
        EXECUTE PROCEDURE update_last_modified();

CREATE VIEW results_join AS (
  SELECT r.*,
  u.name AS owner,

  (SELECT CASE
        WHEN parent_type = 'EffectiveCharacteristic' THEN ec.id
        WHEN parent_type = 'VirtualSimulation' THEN vs.id
        WHEN parent_type = 'Optimization' THEN o.id
        ELSE NULL
     END) AS parent_id,

     (SELECT CASE
           WHEN parent_type = 'EffectiveCharacteristic' THEN ec.name
           WHEN parent_type = 'VirtualSimulation' THEN vs.name
           WHEN parent_type = 'Optimization' THEN o.name
           ELSE NULL
        END) AS parent_name

    FROM results r
    INNER JOIN users u ON u.id = r.owner_id
    LEFT JOIN effective_characteristics ec ON ec.id = r.ec_id
    LEFT JOIN virtual_simulations vs ON vs.id = r.sim_id
    LEFT JOIN optimizations o ON o.id = r.optimization_id
);


CREATE OR REPLACE RULE results_join_CREATE AS ON INSERT TO results_join DO INSTEAD (
    INSERT INTO results (created_by, type, name, path_to_file, description, owner_id, parent_type, ec_id, sim_id, optimization_id)
     VALUES (NEW.created_by, NEW.type, NEW.name, NEW.path_to_file, NEW.description, NEW.owner_id, NEW.parent_type,
           CASE WHEN NEW.parent_type = 'EffectiveCharacteristic' THEN NEW.parent_id ELSE NULL END,
           CASE WHEN NEW.parent_type = 'VirtualSimulation' THEN NEW.parent_id ELSE NULL END,
           CASE WHEN NEW.parent_type = 'Optimization' THEN NEW.parent_id ELSE NULL END)
    RETURNING
      results.*,
      NULL :: VARCHAR AS owner,
      NULL :: INTEGER AS parent_id,
      NULL :: VARCHAR(255) AS parent_name
);

CREATE OR REPLACE RULE results_join_UPDATE AS ON UPDATE TO results_join DO INSTEAD (
  UPDATE results
  SET created_by = NEW.created_by,
      type = NEW.type,
      path_to_file = NEW.path_to_file,
      name = NEW.name,
      description = NEW.description,
      owner_id = NEW.owner_id
  WHERE id = NEW.id);

CREATE OR REPLACE RULE results_join_DELETE AS ON DELETE TO results_join DO INSTEAD (
  DELETE FROM results WHERE id = OLD.id
);

