------------------------------------- clusters ------------------------------------------
CREATE TABLE cluster_types (
  id   SERIAL PRIMARY KEY,
  name VARCHAR(255) NOT NULL
); --#

-- values are used in 'update_start_exec_date' function. So don't change ids
INSERT INTO cluster_types (id, name) VALUES (1, 'Space'); --#
INSERT INTO cluster_types (id, name) VALUES (2, 'Slurm'); --#

CREATE TABLE slurm_tasks_info (
  id            SERIAL,
  task_id       INTEGER NOT NULL,
  slurm_task_id INTEGER,
  nodes         VARCHAR [],
  is_parallel   BOOLEAN,
  num_of_nodes  INTEGER NOT NULL DEFAULT 1,

  PRIMARY KEY (id),
  FOREIGN KEY (task_id) REFERENCES tasks (id) ON DELETE CASCADE,
  UNIQUE (task_id)
);

CREATE INDEX ON slurm_tasks_info (task_id);


CREATE TABLE clusters (
  id      SERIAL,
  type_id INTEGER      NOT NULL,
  name    VARCHAR(255) NOT NULL UNIQUE,
  enabled BOOLEAN      NOT NULL DEFAULT TRUE,

  PRIMARY KEY (id),
  FOREIGN KEY (type_id) REFERENCES cluster_types (id)
);
INSERT INTO clusters (type_id, name) VALUES (1, 'Space Local');


CREATE TABLE slurm_cluster_configs (
  id                  SERIAL,
  cluster_id          INTEGER      NOT NULL,
  address             VARCHAR(255) NOT NULL,
  user_name           VARCHAR(255) NOT NULL,
  path_to_private_key VARCHAR(512) NOT NULL,
  partition_name      VARCHAR(255) NOT NULL,
  all_results_dir     VARCHAR(512) NOT NULL,
  nodes_owner_id      INTEGER      NOT NULL,
  -- in seconds
  connection_timeout  INTEGER      NOT NULL DEFAULT 10,

  PRIMARY KEY (id),
  FOREIGN KEY (cluster_id) REFERENCES clusters (id) ON DELETE CASCADE,
  FOREIGN KEY (nodes_owner_id) REFERENCES users (id)
);

CREATE INDEX ON slurm_cluster_configs (cluster_id);

------------------------------------- tasks ------------------------------------------
ALTER TABLE tasks
  ADD COLUMN auto_select_node BOOLEAN NOT NULL DEFAULT FALSE;
ALTER TABLE tasks
  ADD COLUMN cluster_id INTEGER DEFAULT NULL REFERENCES cluster_types (id); --#

UPDATE tasks
SET cluster_id = 1
WHERE tasks.cluster_id IS NULL; --#

DROP VIEW tasks_join;
CREATE VIEW tasks_join AS (
  SELECT t.*,
    u.name AS owner,
    n.name AS node,

    CASE
    WHEN parent_type = 'effectiveCharacteristic' THEN ec.id
    WHEN parent_type = 'simulation' THEN vs.id
    WHEN parent_type = 'optimization' THEN o.id
    ELSE NULL
    END
           AS parent_id,

    CASE
    WHEN parent_type = 'effectiveCharacteristic' THEN ec.name
    WHEN parent_type = 'simulation' THEN vs.name
    WHEN parent_type = 'optimization' THEN o.name
    ELSE NULL
    END
           AS parent_name

  FROM tasks t
    INNER JOIN users u ON u.id = t.owner_id
    LEFT JOIN nodes n ON n.id = t.node_id
    LEFT JOIN effective_characteristics ec ON ec.id = t.ec_id
    LEFT JOIN virtual_simulations vs ON vs.id = t.sim_id
    LEFT JOIN optimizations o ON o.id = t.optimization_id
);

CREATE OR REPLACE RULE tasks_join_CREATE AS ON INSERT TO tasks_join DO INSTEAD (
  INSERT INTO tasks (type, status_id, queue_priority, solver_name, additional_startup_arguments, node_id,
                     num_cores, memory, progress, auto_select_node, cluster_id, owner_id, parent_type, ec_id, sim_id, optimization_id) VALUES
    (NEW.type, NEW.status_id, NEW.queue_priority, NEW.solver_name, NEW.additional_startup_arguments, NEW.node_id,
               NEW.num_cores, NEW.memory, NEW.progress, NEW.auto_select_node, NEW.cluster_id, NEW.owner_id, NEW.parent_type,
     CASE WHEN NEW.parent_type = 'effectiveCharacteristic' THEN NEW.parent_id ELSE NULL END,
     CASE WHEN NEW.parent_type = 'simulation' THEN NEW.parent_id ELSE NULL END,
     CASE WHEN NEW.parent_type = 'optimization' THEN NEW.parent_id ELSE NULL END)
  RETURNING
    tasks.*,
    NULL :: VARCHAR AS owner,
    NULL :: VARCHAR AS node,
    NULL :: INTEGER AS parent_id,
    NULL :: VARCHAR(255) AS parent_name
);

CREATE OR REPLACE RULE tasks_join_UPDATE AS ON UPDATE TO tasks_join DO INSTEAD (
  UPDATE tasks
  SET type = NEW.type,
    status_id = NEW.status_id,
    queue_priority = NEW.queue_priority,
    task_position = NEW.task_position,
    solver_name = NEW.solver_name,
    additional_startup_arguments = NEW.additional_startup_arguments,
    node_id = NEW.node_id,
    num_cores = NEW.num_cores,
    memory = NEW.memory,
    progress = NEW.progress,
    owner_id = NEW.owner_id,
    auto_select_node = NEW.auto_select_node,
    cluster_id = NEW.cluster_id
  WHERE id = NEW.id);

CREATE OR REPLACE RULE tasks_join_DELETE AS ON DELETE TO tasks_join DO INSTEAD (
  DELETE FROM tasks WHERE id = OLD.id
);

------------------------------------- nodes ------------------------------------------
ALTER TABLE nodes
  ADD COLUMN enabled BOOLEAN NOT NULL DEFAULT TRUE;
ALTER TABLE nodes
  ADD COLUMN cluster_id INTEGER NOT NULL DEFAULT 1; --#

UPDATE nodes
SET cluster_id = 1;

DROP VIEW nodes_alive;
CREATE VIEW nodes_alive AS
  SELECT
    o.*,
    users.name                    AS owner,
    (SELECT ARRAY_AGG(nst.solver_name)
     FROM nodes_solver_types nst) AS available_solvers
  FROM nodes o
    JOIN users ON o.owner_id = users.id; --#
CREATE VIEW space_nodes_alive AS
  SELECT * FROM nodes_alive WHERE cluster_id = 1;


CREATE OR REPLACE RULE nodes_alive_CREATE AS ON INSERT TO nodes_alive DO INSTEAD (
  INSERT INTO nodes (name, num_cores, memory, url, description, alive, enabled, cluster_id, owner_id) VALUES
    (NEW.name, NEW.num_cores, NEW.memory, NEW.url, NEW.description, NEW.alive, NEW.enabled, NEW.cluster_id,
     NEW.owner_id)
  RETURNING nodes.*, NULL :: VARCHAR AS owner, NULL :: VARCHAR [] AS available_solvers);

CREATE OR REPLACE RULE nodes_alive_UPDATE AS ON UPDATE TO nodes_alive DO INSTEAD (
  UPDATE nodes
  SET name      = NEW.name,
    num_cores   = NEW.num_cores,
    memory      = NEW.memory,
    url         = NEW.url,
    description = NEW.description,
    owner_id    = NEW.owner_id,
    alive       = NEW.alive,
    enabled     = NEW.enabled,
    cluster_id  = NEW.cluster_id
  WHERE id = NEW.id);

------------------------------------- triggers ------------------------------------------
-- CREATE FUNCTION check_task_node_cluster()
--   RETURNS TRIGGER
-- LANGUAGE plpgsql
-- AS $$
-- DECLARE
--   node_cluster_id INTEGER;
-- BEGIN
--   IF NEW.node_id IS NULL OR (NEW.cluster_id IS NULL OR NEW.cluster_id = 0)
--   THEN
--     RETURN NEW;
--   END IF;
--
--   SELECT cluster_id
--   INTO node_cluster_id
--   FROM nodes
--   WHERE id = NEW.node_id;
--   IF NEW.cluster_id IS DISTINCT FROM node_cluster_id
--   THEN
--     RAISE EXCEPTION 'Node and task should have same cluster % %', NEW.cluster_id, node_cluster_id;
--   END IF;
--   RETURN new;
-- END;
-- $$; --#
--
-- CREATE TRIGGER check_task_node_cluster_TRIGGER
-- BEFORE INSERT OR UPDATE ON tasks
-- FOR EACH ROW EXECUTE PROCEDURE check_task_node_cluster(); --#
--
--
-- CREATE OR REPLACE FUNCTION update_start_exec_date()
--   RETURNS TRIGGER AS $body$
-- DECLARE
--   cluster_type_id INTEGER;
-- BEGIN
--   SELECT type_id
--   INTO cluster_type_id
--   FROM clusters
--   WHERE id = NEW.cluster_id;
--   -- 102 = 'Sending' state
--   -- 105 = 'Solving' state
--   -- 1 = 'Bench' task manager
--   -- 2 = 'Slurm' task manager
--   IF NEW.status_id <> OLD.status_id AND
--      ((cluster_type_id = 1 AND NEW.status_id = 102) OR (cluster_type_id = 2 AND NEW.status_id = 105))
--   THEN
--     NEW.start_exec_date := now();
--   END IF;
--   RETURN NEW;
-- END;
-- $body$ LANGUAGE plpgsql; --#

