package ru.crp.lvms;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructOptimizationTFDao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructTFDao;
import ru.crp.lvms.optimizations.optistruct.entities.*;
import ru.crp.lvms.optimizations.optistruct.enums.ComponentType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructConvertDirection;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructLoadStepCalculationType;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructFemParser;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructConvertScriptGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructFemScriptGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructOssScriptGenerator;
import ru.crp.lvms.utils.TestUtils;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.function.BiFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static ru.crp.lvms.optimizations.optistruct.enums.ComponentType.PBEAML;
import static ru.crp.lvms.optimizations.optistruct.managers.OptistructAdditionalParametersManager.PLAIN_TYPE_ID;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.*;

public class OptistructIT extends BaseIT {

    @Autowired
    private OptistructOptimizationTFDao optistructOptimizationTFDao;

    @Autowired
    private OptistructOssScriptGenerator optistructOssScriptGenerator;

    @Autowired
    private OptistructConvertScriptGenerator optistructConvertScriptGenerator;

    @Autowired
    private OptistructFemScriptGenerator optistructFemScriptGenerator;

    @Autowired
    private OptistructTFDao optistructTFDao;

    @Autowired
    private OptistructFemParser optistructFemParser;

    @Autowired
    private OptistructOptimizationSettingsManager optistructOptimizationSettingsManager;

    private String optistructSettingsUrl;

    private OptimizationEntity optimizationEntity;

    private OptistructOptimizationTF tfOfOptimization;

    @Before
    public void init() throws Exception {
        optimizationEntity = projectCreator.createOptimization(OptimizationCao.Optistruct);
        optistructSettingsUrl = OPTISTRUCT_SETTINGS_URL.replace("{optimizationId}", String.valueOf(optimizationEntity.getId()));
        createSetAndChooseItAsOptimizationSpace(optimizationEntity.getId());
        tfOfOptimization = getOneOfOptimizationTFs(optimizationEntity.getId());
    }

    @Test
    public void itMustUpdateTfId() throws Exception {
        Integer tfTypeIdForChange = requestDifferentTfType(tfOfOptimization.getTfId());
        tfOfOptimization.setTfId(tfTypeIdForChange);
        OptistructOptimizationTF afterUpdate = updateTF(tfOfOptimization);
        Assert.assertEquals(tfOfOptimization.getTfId(), afterUpdate.getTfId());
    }

    @Test
    public void itMustUpdateOptType() throws Exception {
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions/types");
        List<OptistructTFOptimizationType> optTypesDictionary = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructTFOptimizationType>>() {
        });

        tfOfOptimization.setOptimizationTypeId(optTypesDictionary.get(0).getId());
        OptistructOptimizationTF afterUpdate = updateTF(tfOfOptimization);
        Assert.assertEquals(tfOfOptimization.getTfId(), afterUpdate.getTfId());
    }

    @Test
    public void itMustUpdateAtta() throws Exception {
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions/" + tfOfOptimization.getTfId() + "/attributes");
        OptistructGroupedTFParameters params = jacksonMapper.readValue(jsonNode.toString(), OptistructGroupedTFParameters.class);

        tfOfOptimization.setAttaId(params.getAtta().get(0).getId());
        OptistructOptimizationTF afterUpdate = updateTF(tfOfOptimization);
        Assert.assertEquals(tfOfOptimization.getTfId(), afterUpdate.getTfId());
        Assert.assertEquals(tfOfOptimization.getAttaId(), afterUpdate.getAttaId());
    }

    @Test
    public void itMustDeleteDependentParametersOnTFidChange() throws Exception {
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions/types");
        List<OptistructTFOptimizationType> optTypesDictionary = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructTFOptimizationType>>() {
        });

        tfOfOptimization.setOptimizationTypeId(optTypesDictionary.get(0).getId());
        tfOfOptimization.setAttaId(PLAIN_TYPE_ID);
        tfOfOptimization.setAttaValue("TestAtta");

        OptistructOptimizationTF afterUpdate = updateTF(tfOfOptimization);

        Assert.assertEquals(tfOfOptimization.getOptimizationTypeId(), afterUpdate.getOptimizationTypeId());
        Assert.assertEquals(tfOfOptimization.getAttaId(), afterUpdate.getAttaId());
        Assert.assertEquals(tfOfOptimization.getAttaValue(), afterUpdate.getAttaValue());

        tfOfOptimization.setTfId(requestDifferentTfType(tfOfOptimization.getTfId()));

        afterUpdate = updateTF(tfOfOptimization);

        Assert.assertEquals(tfOfOptimization.getOptimizationTypeId(), afterUpdate.getOptimizationTypeId());
        Assert.assertEquals(null, afterUpdate.getAttaValue());
        Assert.assertNotEquals(null, afterUpdate.getAttaId());
    }

    @Test
    public void itMustUpdateMultipleParameters() throws Exception {
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions/types");
        List<OptistructTFOptimizationType> optTypes = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructTFOptimizationType>>() {
        });

        jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions/" + tfOfOptimization.getTfId() + "/attributes");
        OptistructGroupedTFParameters params = jacksonMapper.readValue(jsonNode.toString(), OptistructGroupedTFParameters.class);

        tfOfOptimization.setOptimizationTypeId(optTypes.get(0).getId());
        tfOfOptimization.setAttaId(PLAIN_TYPE_ID);
        tfOfOptimization.setAttaValue("TestAtta");
        tfOfOptimization.setAttiId(PLAIN_TYPE_ID);
        tfOfOptimization.setAttiValue("TestAtti");

        OptistructOptimizationTF afterUpdate = updateTF(tfOfOptimization);

        Assert.assertEquals(tfOfOptimization.getAttaId(), afterUpdate.getAttaId());
        Assert.assertEquals(tfOfOptimization.getAttaValue(), afterUpdate.getAttaValue());
        Assert.assertEquals(tfOfOptimization.getAttiId(), afterUpdate.getAttiId());
        Assert.assertEquals(tfOfOptimization.getAttiValue(), afterUpdate.getAttiValue());

        tfOfOptimization = afterUpdate;
        tfOfOptimization.setTfId(requestDifferentTfType(tfOfOptimization.getTfId()));
        tfOfOptimization.setOptimizationTypeId(optTypes.get(0).getId());
        tfOfOptimization.setAttaId(params.getAtta().get(0).getId());
        tfOfOptimization.setAttaValue("TestAtta");
        tfOfOptimization.setAttiId(params.getAtti().get(0).getId());
        tfOfOptimization.setAttiValue("TestAtti");

        afterUpdate = updateTF(tfOfOptimization);

        Assert.assertEquals(tfOfOptimization.getAttaId(), afterUpdate.getAttaId());
        Assert.assertEquals(null, afterUpdate.getAttaValue());
        Assert.assertEquals(tfOfOptimization.getAttiId(), afterUpdate.getAttiId());
        Assert.assertEquals(null, afterUpdate.getAttiValue());
    }

    @Test
    public void itMustUpdateOptimizationConstraints() throws Exception {
        OptistructOptimizationConstraint cOfOptimization = getOneOfOptimizationConstraints(optimizationEntity.getId());

        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/constraints/" + cOfOptimization.getTfId() + "/attributes");
        OptistructGroupedTFParameters params = jacksonMapper.readValue(jsonNode.toString(), OptistructGroupedTFParameters.class);

        cOfOptimization.setAttaId(PLAIN_TYPE_ID);
        cOfOptimization.setAttaValue("TestAtta");
        cOfOptimization.setAttiId(PLAIN_TYPE_ID);
        cOfOptimization.setAttiValue("TestAtti");
        cOfOptimization.setLowerBound("TestLowerBound");
        cOfOptimization.setUpperBound("TestUpperBound");

        OptistructOptimizationConstraint afterUpdate = updateC(cOfOptimization);

        Assert.assertEquals(cOfOptimization.getAttaId(), afterUpdate.getAttaId());
        Assert.assertEquals(cOfOptimization.getAttaValue(), afterUpdate.getAttaValue());
        Assert.assertEquals(cOfOptimization.getAttiId(), afterUpdate.getAttiId());
        Assert.assertEquals(cOfOptimization.getAttiValue(), afterUpdate.getAttiValue());
        Assert.assertEquals(cOfOptimization.getLowerBound(), afterUpdate.getLowerBound());
        Assert.assertEquals(cOfOptimization.getUpperBound(), afterUpdate.getUpperBound());

        cOfOptimization = afterUpdate;
        cOfOptimization.setAttaId(params.getAtta().get(0).getId());
        cOfOptimization.setAttaValue("TestAtta");
        cOfOptimization.setAttiId(params.getAtti().get(0).getId());
        cOfOptimization.setAttiValue("TestAtti");
        cOfOptimization.setLowerBound("TestLowerBound");
        cOfOptimization.setUpperBound("TestUpperBound");

        afterUpdate = updateC(cOfOptimization);

        Assert.assertEquals(cOfOptimization.getAttaId(), afterUpdate.getAttaId());
        Assert.assertEquals(null, afterUpdate.getAttaValue());
        Assert.assertEquals(cOfOptimization.getAttiId(), afterUpdate.getAttiId());
        Assert.assertEquals(null, afterUpdate.getAttiValue());
        Assert.assertEquals(cOfOptimization.getLowerBound(), afterUpdate.getLowerBound());
        Assert.assertEquals(cOfOptimization.getUpperBound(), afterUpdate.getUpperBound());
    }

    @Ignore
    @Test
    public void itMustUpdateOptimizationTechConstraints() throws Exception {
        String cUrl = optistructSettingsUrl + "/technologicalConstraints";
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/technologicalConstraints");
        List<OptistructTechnologicalConstraint> tfs = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructTechnologicalConstraint>>() {
        });

        jsonNode = TestUtils.GETobjects(mockMvc, cUrl);
        List<OptistructOptimizationTechnologicalConstraint> tcs = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructOptimizationTechnologicalConstraint>>() {
        });

        OptistructOptimizationTechnologicalConstraint entity = tcs.get(0);
        allowEverythingForTFType(tfs.get(0).getId());
        entity.setTechnologicalConstraintId(tfs.get(0).getId());

        OptistructOptimizationTechnologicalConstraint afterUpdate =
                UPDATEobject(mockMvc, cUrl + "/" + entity.getId(), entity, OptistructOptimizationTechnologicalConstraint.class, TestUtils::toJsonNoProxy);

        Assert.assertEquals(entity.getTechnologicalConstraintId(), afterUpdate.getTechnologicalConstraintId());
    }

    @Test
    public void itMustGenerateOssWithoutExceptions() throws IOException {
        Path path = optistructOssScriptGenerator.generateConfig(optimizationEntity);
        Files.delete(path);
    }

    @Test
    public void itMustGenerateConvertWithoutExceptions() throws IOException {
        for (OptistructConvertDirection direction : OptistructConvertDirection.values()) {
            Path path = optistructConvertScriptGenerator.generateConfig(optimizationEntity, direction);
            Files.delete(path);
        }
    }

    @Test
    @Ignore
    public void itMustGenerateFemWithoutExceptions() throws IOException {
        Path path = optistructFemScriptGenerator.generateConfig(optimizationEntity);
        Files.delete(path);
    }

    @Test
    @Ignore
    public void itMustExportFemWithoutExceptions() throws Exception {
        // Values for constraints
        OptistructOptimizationConstraint cOfOptimization = getOneOfOptimizationConstraints(optimizationEntity.getId());
        cOfOptimization.setUpperBound("test");
        cOfOptimization.setLowerBound("test");
        updateC(cOfOptimization);
        // Values for tech constraints
        String cUrl = optistructSettingsUrl + "/technologicalConstraints";
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, cUrl);
        List<OptistructOptimizationTechnologicalConstraint> tcs = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructOptimizationTechnologicalConstraint>>() {
        });
        OptistructOptimizationTechnologicalConstraint entity = tcs.get(0);
        UPDATEobject(mockMvc, cUrl + "/" + entity.getId(), entity, OptistructOptimizationTechnologicalConstraint.class, TestUtils::toJsonNoProxy);

        Path path = TestUtils.downloadFile(mockMvc, OPTISTRUCT_CONFIG_URL.replace("{optimizationId}", String.valueOf(optimizationEntity.getId())));
        Files.delete(path);
    }

    @Test
    public void itMustParseSets() throws IOException {
        Path tmp = Files.createTempFile(null, null);
        String data = "$\n" +
                "$$\n" +
                "$$  PSHELL Data\n" +
                "$$\n" +
                "$HMNAME PROP                   3\"des2\" 4\n" +
                "$HWCOLOR PROP                  3      53\n" +
                "PSHELL         3       25.0            2               2        0.0     \n" +
                "$$\n" +
                "$$  PCOMP Data\n" +
                "$$\n" +
                "$                                                                              \n" +
                "$                                                                              \n" +
                "$HMNAME PROP                   1\"non_des\" 4\n" +
                "$HWCOLOR PROP                  1      18\n" +
                "PCOMP          1        0.0                     0.0                     \n" +
                "+              15.0     0.0          YES\n" +
                "$HMNAME PROP                   2\"des\" 4\n" +
                "$HWCOLOR PROP                  2      21\n" +
                "PCOMP          2        0.0                     0.0                     \n" +
                "+              25.0     0.0          YES\n" +
                "$$\n" +
                "$$  MAT1 Data\n" +
                "$$\n" +
                "$HMNAME MAT                    1\"MAT1\" \"MAT1\"\n" +
                "$HWCOLOR MAT                   1       3\n" +
                "MAT1           12.0+11          0.3     1200.0                          \n" +
                "$HMNAME MAT                    2\"MAT2\" \"MAT1\"\n" +
                "$HWCOLOR MAT                   2       4\n" +
                "MAT1           22.0+11          0.3     1200.0                          \n" +
                "$$\n" +

                "$HMNAME PROP                  11\"P6;p6_1\" 3\n" +
                "$HWCOLOR PROP                 11      18\n" +
                "$HMBEAMSEC PBARASSOC          11       5\n" +
                "PBAR          11       16.4-5   3.413-103.413-105.768-100.0\n" +

                "$$\n" +
                "$HMNAME PROP                   8\"P3;p3\" 3\n" +
                "$HWCOLOR PROP                  8       6\n" +
                "$HMBEAMSEC PRODASSOC           8      11\n" +
                "PROD           8       13.1416-41.5708-8\n" +

                "$HMNAME PROP                   6\"P1;p1\" 3\n" +
                "$HWCOLOR PROP                  6       4\n" +
                "$HMBEAMSEC PBARLASSOC          6       8\n" +
                "PBARL          6       1             BAR                                +\n" +
                "+       0.0075  0.0075\n" +
                "$$\n" +

                "$HMNAME PROP                  10\"P5;p5\" 3\n" +
                "$HWCOLOR PROP                 10       8\n" +
                "$HMBEAMSEC PBEAMASSOC         10       8\n" +
                "PBEAM         10       15.625-5 2.637-102.637-100.0     4.456-10\n" +

                "$HMNAME PROP                   7\"P2;p2\" 3\n" +
                "$HWCOLOR PROP                  7       5\n" +
                "$HMBEAMSEC PBEAMLASSOC         7       9\n" +
                "PBEAML         7       1             ROD\n";
        Files.write(tmp, data.getBytes());
        List<OptistructSet> sets = optistructFemParser.extractSets(1, tmp);
        Assert.assertEquals(8, sets.size());
        Stream.of(new OptistructSet(1, ComponentType.PSHELL, "des2"),
                new OptistructSet(1, ComponentType.PCOMP, "non_des"),
                new OptistructSet(1, ComponentType.PCOMP, "des"),
                new OptistructSet(1, ComponentType.PBAR, "P6;p6_1"),
                new OptistructSet(1, ComponentType.PROD, "P3;p3"),
                new OptistructSet(1, ComponentType.PBARL, "P1;p1"),
                new OptistructSet(1, ComponentType.PBEAM, "P5;p5"),
                new OptistructSet(1, PBEAML, "P2;p2")
        )
                .forEach(set -> Assert.assertTrue(sets.contains(set)));
        Files.delete(tmp);
    }

    @Test
    public void itMustParseLoadCollectors() throws IOException {
        createCollectors();
    }

    private List<OptistructCollector> createCollectors() throws IOException {
        Path tmp = Files.createTempFile(null, null);
        String data = "$$------------------------------------------------------------------------$\n" +
                "$$ HyperMesh Commands for loadcollectors name and color information $\n" +
                "$$------------------------------------------------------------------------------$\n" +
                "$HMNAME LOADCOL                1\"1\"\n" +
                "$HWCOLOR LOADCOL               1      18\n" +
                "$$\n" +
                "$HMNAME LOADCOL                2\"2\"\n" +
                "$HWCOLOR LOADCOL               2      21\n" +
                "$$\n" +
                "$$\n" +
                "$$  SPC Data\n" +
                "$$\n" +
                "SPC            2       1  1234560.0     \n" +
                "SPC            2      22  1234560.0     \n" +
                "$$\n" +
                "$$\n" +
                "$$\n" +
                "$$  FORCE Data\n" +
                "$$\n" +
                "FORCE          1      14       01.0     0.0     -100.0  0.0     \n" +
                "FORCE          1       2       01.0     0.0     -100.0  0.0     \n" +
                "FORCE          1      13       01.0     0.0     -100.0  0.0     \n" +
                "ENDDATA\n" +
                "$$";
        Files.write(tmp, data.getBytes());
        List<OptistructCollector> collectors = optistructFemParser.extractCollectors(1, tmp);
        Assert.assertEquals(2, collectors.size());
        Stream.of(new OptistructCollector(1, 1, "1", "Load"),
                new OptistructCollector(1, 2, "2", "Fixation"))
                .forEach(step -> Assert.assertTrue(collectors.contains(step)));
        Files.delete(tmp);

        return collectors;
    }

    @Test
    public void itMustParseTestExamples() throws Exception {
        String basicURL = optistructSettingsUrl + "/basic";
        JsonNode basicSettingsJson = GETobjects(mockMvc, basicURL);
        List<OptistructOptimizationBaseSetting> settings = jacksonMapper.readValue(basicSettingsJson.toString(),
                new TypeReference<List<OptistructOptimizationBaseSetting>>() {
                });
        Map<Integer, List<OptistructOptimizationBaseSetting>> tcSettings = settings.stream()
                .collect(Collectors.groupingBy(OptistructOptimizationBaseSetting::getTypeId));
        OptistructOptimizationBaseSetting baseSetting = tcSettings.get(OptistructBaseSettingType.NUM_LOAD_STEPS.getId()).get(0);
        baseSetting.setCurrentValue(2);
        OptistructOptimizationBaseSetting updated = TestUtils.UPDATEobject(mockMvc, basicURL + "/" + baseSetting.getId(),
                baseSetting, OptistructOptimizationBaseSetting.class, TestUtils::toJsonNoProxy);

        checkParseTestExamples("fem-parsing/parsing_fem_1.fem",
                "fem-parsing/parsing_result_1.tcl",
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.STATIC);
                    loadStep.setIsFixed(true);
                    loadStep.setSelectedLoadIds(
                            collectors.stream()
                                    .filter(col -> col.getName().equals("Force") || col.getName().equals("Moment"))
                                    .map(OptistructCollector::getId)
                                    .collect(Collectors.toList())
                    );
                    return loadStep;
                },
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.MODAL);
                    loadStep.setIsFixed(false);
                    loadStep.setSelectedLoadIds(Collections.emptyList());
                    return loadStep;
                });

        checkParseTestExamples("fem-parsing/parsing_fem_2.fem",
                "fem-parsing/parsing_result_2.tcl",
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.STATIC);
                    loadStep.setIsFixed(true);
                    loadStep.setSelectedLoadIds(
                            collectors.stream()
                                    .filter(col -> col.getName().equals("Force") || col.getName().equals("Moment"))
                                    .map(OptistructCollector::getId)
                                    .collect(Collectors.toList())
                    );
                    return loadStep;
                },
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.MODAL);
                    loadStep.setIsFixed(false);
                    loadStep.setSelectedLoadIds(Collections.emptyList());
                    return loadStep;
                });

        checkParseTestExamples("fem-parsing/parsing_fem_3.fem",
                "fem-parsing/parsing_result_3.tcl",
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.STATIC);
                    loadStep.setIsFixed(true);
                    loadStep.setSelectedLoadIds(
                            collectors.stream()
                                    .filter(col -> col.getName().equals("Force") || col.getName().equals("Moment"))
                                    .map(OptistructCollector::getId)
                                    .collect(Collectors.toList())
                    );
                    return loadStep;
                },
                (loadStep, collectors) -> {
                    loadStep.setCalculationType(OptistructLoadStepCalculationType.STATIC);
                    loadStep.setIsFixed(true);
                    loadStep.setSelectedLoadIds(
                            collectors.stream()
                                    .filter(col -> col.getName().equals("Force"))
                                    .map(OptistructCollector::getId)
                                    .collect(Collectors.toList())
                    );
                    return loadStep;
                });
    }

    private void checkParseTestExamples(String fem, String tcl,
                                        BiFunction<OptistructOptimizationLoadStep, List<OptistructCollector>, OptistructOptimizationLoadStep> firstStepUpdater,
                                        BiFunction<OptistructOptimizationLoadStep, List<OptistructCollector>, OptistructOptimizationLoadStep> secondStepUpdater) throws Exception {
        Path pathToTestScript = Paths.get(getClass().getClassLoader().getResource(fem).toURI());
        Path pathToCorrectResult = Paths.get(getClass().getClassLoader().getResource(tcl).toURI());

        List<OptistructCollector> parsedCollectors = optistructFemParser.extractCollectors(1, pathToTestScript);
        optistructOptimizationSettingsManager.resetCollectors(optimizationEntity.getId(), parsedCollectors);

        String lsUrl = optistructSettingsUrl + "/loadSteps";

        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, lsUrl + "/loads");
        List<OptistructCollector> collectors =
                jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructCollector>>() {
                });

        jsonNode = TestUtils.GETobjects(mockMvc, lsUrl + "/calculationTypes");
        List<OptistructLoadStepCalculationType> stepTypes =
                jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructLoadStepCalculationType>>() {
                });
        Assert.assertArrayEquals(stepTypes.toArray(), OptistructLoadStepCalculationType.values());

        jsonNode = TestUtils.GETobjects(mockMvc, lsUrl);
        List<OptistructOptimizationLoadStep> lss =
                jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructOptimizationLoadStep>>() {
                });

        OptistructOptimizationLoadStep entity = firstStepUpdater.apply(lss.get(0), collectors);

        OptistructOptimizationLoadStep afterUpdate =
                UPDATEobject(mockMvc, lsUrl + "/" + entity.getId(), entity,
                        OptistructOptimizationLoadStep.class, TestUtils::toJsonNoProxy);

        Assert.assertArrayEquals(entity.getSelectedLoadIds().toArray(), afterUpdate.getSelectedLoadIds().toArray());
        Assert.assertEquals(entity.getCalculationType(), afterUpdate.getCalculationType());
        Assert.assertEquals(entity.getIsFixed(), afterUpdate.getIsFixed());

        entity = secondStepUpdater.apply(lss.get(1), collectors);

        afterUpdate =
                UPDATEobject(mockMvc, lsUrl + "/" + entity.getId(), entity, OptistructOptimizationLoadStep.class, TestUtils::toJsonNoProxy);

        Assert.assertArrayEquals(entity.getSelectedLoadIds().toArray(), afterUpdate.getSelectedLoadIds().toArray());
        Assert.assertEquals(entity.getCalculationType(), afterUpdate.getCalculationType());
        Assert.assertEquals(entity.getIsFixed(), afterUpdate.getIsFixed());

        Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings =
                optistructOptimizationSettingsManager.findBaseSettings(optimizationEntity.getId(), null);
        String content = optistructFemScriptGenerator.formLoadSteps(optimizationEntity, baseSettings);
        Assert.assertEquals(Files.readAllLines(pathToCorrectResult).stream().collect(Collectors.joining("\n")), content);
    }

    private OptistructOptimizationTF getOneOfOptimizationTFs(Integer optimizationId) throws IOException {
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, OPTISTRUCT_TF_URL.replace("{optimizationId}", String.valueOf(optimizationId)));
        List<OptistructOptimizationTF> otfs = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructOptimizationTF>>() {
        });
        allowEverythingForTFType(otfs.get(0).getTfId());
        return otfs.get(0);
    }

    private OptistructOptimizationConstraint getOneOfOptimizationConstraints(Integer optimizationId) throws IOException {
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, OPTISTRUCT_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(optimizationId)));
        List<OptistructOptimizationConstraint> ocs = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructOptimizationConstraint>>() {
        });
        allowEverythingForTFType(tfOfOptimization.getTfId());
        return ocs.get(0);
    }

    private Integer requestDifferentTfType(Integer tfId) throws IOException {
        JsonNode jsonNode = GETobjects(mockMvc, OPTISTRUCT_SUGGEST_URL + "/targetFunctions");
        List<OptistructTF> tfDictionary = jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<OptistructTF>>() {
        });
        int chosenTfTypeId = tfDictionary.stream().filter(tf -> tf.getId() != tfId).findAny().get().getId();
        allowEverythingForTFType(chosenTfTypeId);
        return chosenTfTypeId;
    }

    private OptistructOptimizationTF updateTF(OptistructOptimizationTF optimizationTf) throws Exception {
        String tfUrl = OPTISTRUCT_TF_URL.replace("{optimizationId}", String.valueOf(optimizationTf.getOptimizationId())) + "/" + optimizationTf.getId();
        return UPDATEobject(mockMvc, tfUrl, optimizationTf, OptistructOptimizationTF.class, TestUtils::toJsonNoProxy);
    }

    private OptistructOptimizationConstraint updateC(OptistructOptimizationConstraint optimizationTf) throws Exception {
        String tfUrl = OPTISTRUCT_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(optimizationTf.getOptimizationId())) + "/" + optimizationTf.getId();
        return UPDATEobject(mockMvc, tfUrl, optimizationTf, OptistructOptimizationConstraint.class, TestUtils::toJsonNoProxy);
    }

    private void createSetAndChooseItAsOptimizationSpace(Integer optimizationId) throws Exception {
        OptistructSet optistructSet = new OptistructSet(optimizationId, ComponentType.values()[0], "testSetName");
        optistructSet.setIsOptimizationSpace(true);
        optistructSet.setIsSelected(true);
        CREATEobject(mockMvc, optistructSettingsUrl + "/sets", optistructSet);
    }

    private void allowEverythingForTFType(Integer tfId) {
        String ptypeSql = "INSERT INTO optistruct_tf_allowed_ptype(tf_id, ptype_id) " +
                "SELECT ?, t.id " +
                "FROM optistruct_tf_ptype AS t";
        optistructTFDao.jdbcTemplate().update(ptypeSql, tfId);
        String attaSql = "INSERT INTO optistruct_tf_allowed_atta(tf_id, atta_id) " +
                "SELECT ?, t.id " +
                "FROM optistruct_tf_atta AS t";
        optistructTFDao.jdbcTemplate().update(attaSql, tfId);
        String attbSql = "INSERT INTO optistruct_tf_allowed_attb(tf_id, attb_id) " +
                "SELECT ?, t.id " +
                "FROM optistruct_tf_attb AS t";
        optistructTFDao.jdbcTemplate().update(attbSql, tfId);
        String attiSql = "INSERT INTO optistruct_tf_allowed_atti(tf_id, atti_id) " +
                "SELECT ?, t.id " +
                "FROM optistruct_tf_atti AS t";
        optistructTFDao.jdbcTemplate().update(attiSql, tfId);
    }
}
