package ru.crp.lvms.material;

import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Assert;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Value;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.BaseIT;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.web.dto.pagination.DataRequest;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static java.util.stream.Collectors.toList;
import static org.junit.Assert.*;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.COMPOSITE_MATERIALS_URL;
import static ru.crp.lvms.utils.URLs.MATERIALS_URL;

public class MaterialFilesEndpointIT extends BaseIT {
    @Value("${storage.root}")
    private String filesFolder;

    private Path pathToFiles;

    @Override
    public void setup() throws Exception {
        super.setup();
        pathToFiles = Paths.get(filesFolder);
    }


    @Test
    public void itShouldBeAbleToCreateMaterialFile() throws Exception {
        MaterialEntity material = projectCreator.createMaterial();
        testCreation(false, material.getId());

        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial();
        testCreation(true, compositeMaterial.getId());
    }


    private void testCreation(boolean isComposite, int materialId) throws Exception {
        FileEntity newMaterialFile = projectCreator.newMaterialFile(isComposite);
        newMaterialFile.setPid(materialId);
        FileEntity createdMaterialFile = projectCreator.createMaterialFile(isComposite, newMaterialFile);
        assertEquals(Integer.valueOf(materialId), createdMaterialFile.getPid());
    }


    @Test
    public void itShouldBeAbleToGetAllMaterialFiles() throws Exception {
        MaterialEntity material = projectCreator.createMaterial();
        testGetAll(false, material.getId());

        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial();
        testGetAll(true, compositeMaterial.getId());
    }

    private void testGetAll(boolean isComposite, int parentId) throws Exception {
        FileEntity materialFile1 = projectCreator.newMaterialFile(isComposite);
        materialFile1.setPid(parentId);
        projectCreator.createMaterialFile(isComposite, materialFile1);

        FileEntity materialFile2 = projectCreator.newMaterialFile(isComposite);
        materialFile2.setPid(parentId);
        projectCreator.createMaterialFile(isComposite, materialFile2);

        String materialUrl = isComposite ? COMPOSITE_MATERIALS_URL : MATERIALS_URL;
        JsonNode response = GETobjects(mockMvc, createObjectUrl(materialUrl, parentId) + "/materialFiles");
        List<FileEntity> allMaterialFiles = filesFromJson(response);
        assertEquals(2, allMaterialFiles.size());
        assertTrue(allMaterialFiles.get(0).getId() == materialFile1.getId()
                || allMaterialFiles.get(0).getId() == materialFile2.getId());
    }


    @Test
    public void itShouldBeAbleToGetMaterialFilesPage() throws Exception {
        MaterialEntity material = projectCreator.createMaterial();
        testGetPage(false, material.getId());

        CompositeMaterialEntity compositeMaterial = projectCreator.getCompositeMaterial();
        testGetPage(true, compositeMaterial.getId());
    }

    private void testGetPage(boolean isComposite, int pid) throws Exception {
        FileEntity materialIn1 = projectCreator.createMaterialFile(isComposite, pid);
        FileEntity materialIn2 = projectCreator.createMaterialFile(isComposite, pid);

        DataRequest dataRequest = projectCreator.newDataRequest();
        String materialUrl = isComposite ? COMPOSITE_MATERIALS_URL : MATERIALS_URL;
        String url = createObjectUrl(materialUrl, materialIn1.getPid()) + "/materialFiles/list";
        JsonNode materialsPageJson = GETPage(mockMvc, url, dataRequest);
        List<Integer> materialsPage = filesPageFromJson(materialsPageJson).stream()
                .map(FileEntity::getId).collect(toList());
        assertEquals(2, materialsPage.size());
        assertTrue(materialsPage.contains(materialIn1.getId()));
        assertTrue(materialsPage.contains(materialIn2.getId()));
    }


    @Test
    public void itShouldBeAbleToUpdateMaterialFile() throws Exception {
        testUpdate(false);
        testUpdate(true);
    }

    private void testUpdate(boolean isComposite) throws Exception {
        FileEntity newMf = projectCreator.newMaterialFile(isComposite);
        FileEntity createdMf = projectCreator.createMaterialFile(isComposite, newMf);
        newMf.setName(unique("updated name"));
        newMf.setId(createdMf.getId());
        FileEntity updatedMf = UPDATEobject(mockMvc, createMFUrl(isComposite, createdMf), newMf, FileEntity.class);
        assertEquals(newMf.getName(), updatedMf.getName());
    }


    @Test
    public void itShouldBeAbleToDeleteMaterialFile() throws Exception {
        FileEntity materialFile = projectCreator.createMaterialFile(false);
        DELETEobject(mockMvc, createMFUrl(false, materialFile));

        materialFile = projectCreator.createMaterialFile(true);
        DELETEobject(mockMvc, createMFUrl(true, materialFile));
    }


    @Test
    public void materialFileMustBeRemovedFromDiskWhenRemovingItFromDb() throws Exception {
        testRemoveFileFromDisk(false);
        testRemoveFileFromDisk(true);
    }

    private void testRemoveFileFromDisk(boolean isComposite) throws Exception {
        FileEntity materialFile = projectCreator.createMaterialFile(isComposite);
        assertTrue(Files.exists(pathToFiles.resolve(materialFile.getPathToFile())));
        DELETEobject(mockMvc, createMFUrl(isComposite, materialFile));
        assertFalse(Files.exists(pathToFiles.resolve(materialFile.getPathToFile())));
    }


    @Test
    public void itShouldSuggestMaterialFiles() throws Exception {
        MaterialEntity material = projectCreator.createMaterial();
        testSuggest(false, material.getId());

        CompositeMaterialEntity compositeMaterial = projectCreator.createCompositeMaterial();
        testSuggest(true, compositeMaterial.getId());
    }

    private void testSuggest(boolean isComposite, int pid) throws Exception {
        List<String> files = new ArrayList<>();
        String materialUrl = isComposite ? COMPOSITE_MATERIALS_URL : MATERIALS_URL;
        for (int i = 0; i < 5; i++) {
            FileEntity newMaterialFile = projectCreator.newMaterialFile(isComposite);
            newMaterialFile.setName(unique("TestName") + i);
            newMaterialFile.setPid(pid);
            FileEntity createdMaterialFile = projectCreator.createMaterialFile(isComposite, newMaterialFile);
            assertEquals(Integer.valueOf(pid), createdMaterialFile.getPid());
            assertEquals(newMaterialFile.getName(), createdMaterialFile.getName());
            files.add(newMaterialFile.getName());
        }
        SuggestRequest suggestRequest = projectCreator.newSuggestRequest("name", "%TestName%");
        FileEntity[] result =
                CREATEobject(mockMvc, createObjectUrl(materialUrl, pid) + "/materialFiles/suggest",
                        suggestRequest, FileEntity[].class);
        Assert.assertEquals(files.size(), result.length);
        List<String> out = Arrays.asList(result).stream().map(FileEntity::getName).collect(Collectors.toList());
        for (String file : files) {
            Assert.assertTrue(out.contains(file));
        }
    }


    private String createMFUrl(boolean isComposite, FileEntity materialFile) {
        return createObjectUrl(isComposite ? COMPOSITE_MATERIALS_URL : MATERIALS_URL,
                materialFile.getPid()) + "/materialFiles/" + materialFile.getId();
    }
}
