package ru.crp.lvms.optimizations;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Assert;
import org.junit.Test;
import ru.crp.lvms.BaseIT;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNode;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.tosca.entities.*;
import ru.crp.lvms.utils.SetterChecker;
import ru.crp.lvms.utils.TestUtils;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType.NUM_TECH_CONSTRAINTS;
import static ru.crp.lvms.utils.TestUtils.GETobjects;
import static ru.crp.lvms.utils.TestUtils.unique;
import static ru.crp.lvms.utils.URLs.*;

public class ToscaTechnologicalConstraintTests extends BaseIT {

    @Test
    public void itMustUpdateTCFieldValues() throws Exception {
        OptimizationEntity toscaOptimization = projectCreator.createOptimization(OptimizationCao.Tosca);

        String suggestUrl = TOSCA_SUGGEST_URL.replace("{optimizationTypeId}",
                String.valueOf(toscaOptimization.getOptimizationTypeId()));

        setTCLimitToMax(toscaOptimization);

        List<ToscaTechnologicalConstraint> possibleTCForTosca =
                getPossibleTCForTosca(suggestUrl);

        Assert.assertFalse(possibleTCForTosca.isEmpty());


        Map<Integer, List<TechnologicalConstraintTreeNode>> leafsByTC = possibleTCForTosca.stream()
                .flatMap(tc -> getParametersWithFields(suggestUrl, tc.getId()))
                .filter(node -> !node.getIsParent())
                .collect(groupingBy(TechnologicalConstraintTreeNode::getTechnologicalConstraintId));
        leafsByTC.forEach((tcId, leafs) -> {
            if (leafs.isEmpty()) {
                return;
            }
            ToscaOptimizationTechnologicalConstraint tcConstraint = setNewTC(tcId, possibleTCForTosca, toscaOptimization);
            createValues(tcConstraint, leafs, toscaOptimization);
            updateValues(tcConstraint, toscaOptimization);
        });
    }

    private void setTCLimitToMax(OptimizationEntity toscaOptimization) throws Exception {
        String basicURL = TOSCA_BASE_SETTINGS_URL.replace("{optimizationId}", String.valueOf(toscaOptimization.getId()));
        JsonNode basicSettingsJson = GETobjects(mockMvc, basicURL);
        List<ToscaOptimizationBaseSetting> settings = jsonToToscaOptimizationBaseSettingList(basicSettingsJson);
        ToscaOptimizationBaseSetting tcSetting = settings.stream()
                .filter(s -> s.getTypeId() == NUM_TECH_CONSTRAINTS.getId()).findAny().get();
        tcSetting.setCurrentValue(4);
        TestUtils.UPDATEobject(mockMvc, basicURL + "/" + tcSetting.getId(), tcSetting,
                ToscaOptimizationBaseSetting.class, TestUtils::toJsonNoProxy);
    }

    private void createValues(ToscaOptimizationTechnologicalConstraint optimizationTC, List<TechnologicalConstraintTreeNode> fields, OptimizationEntity toscaOptimization) {
        fields.forEach(field -> {
            String fieldUrl = TOSCA_TECH_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(toscaOptimization.getId())) + "/" + optimizationTC.getId() + "/fields/" + field.getId();
            TechnologicalConstraintTreeNodeValue newValue = null;
            try {
                newValue = SetterChecker.createProxy(TechnologicalConstraintTreeNodeValue.class);
            } catch (Exception e) {
                Assert.fail("Unable to create SetterChecker proxy");
            }
            // only this 3 fields are really inserted to the table
            newValue.setTechnologicalConstraintTreeId(field.getId());
            newValue.setTechnologicalConstraintId(optimizationTC.getId());
            newValue.setValue(unique("Test"));

            newValue.setDisplayName(null);
            newValue.setId(0);
            newValue.setIsValueEditable(null);
            newValue.setName(null);
            newValue.setParentId(null);
            newValue.setIsParent(false);
            newValue.setGroupName(null);
            newValue.setTooltip(null);
            newValue.setType(null);
            newValue.setChildren(null);
            newValue.setDependencies(null);
            newValue.setIsSelectable(false);

            TechnologicalConstraintTreeNodeValue created = null;
            try {
                created = TestUtils.UPDATEobject(mockMvc, fieldUrl, newValue, TechnologicalConstraintTreeNodeValue.class);
            } catch (Exception e) {
                Assert.fail("Unable to read value from json");
            }

            if (field.getIsValueEditable()) {
                Assert.assertEquals(newValue.getValue(), created.getValue());
            } else {
                Assert.assertNull(created.getValue());
            }
        });
    }

    private void updateValues(ToscaOptimizationTechnologicalConstraint optimizationTC, OptimizationEntity toscaOptimization) {
        String fieldsUrl = TOSCA_TECH_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(toscaOptimization.getId())) + "/" + optimizationTC.getId() + "/fields";
        JsonNode fieldsJson = TestUtils.GETobjects(mockMvc, fieldsUrl);
        List<TechnologicalConstraintTreeNodeValue> currentFields = jsonToToscaTechnologicalConstraintTreeValue(fieldsJson);
        currentFields.forEach(field -> {
            String fieldUrl = TOSCA_TECH_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(toscaOptimization.getId())) + "/" + optimizationTC.getId() + "/fields/" + field.getTechnologicalConstraintTreeId();
            field.setValue(unique("NewTest"));

            TechnologicalConstraintTreeNodeValue created = null;
            try {
                created = TestUtils.UPDATEobject(mockMvc, fieldUrl, field, TechnologicalConstraintTreeNodeValue.class, TestUtils::toJsonNoProxy);
            } catch (Exception e) {
                Assert.fail("Unable to read value from json");
            }

            if (field.getIsValueEditable()) {
                Assert.assertEquals(field.getValue(), created.getValue());
            } else {
                Assert.assertNull(created.getValue());
            }
        });
    }

    private ToscaOptimizationTechnologicalConstraint setNewTC(int tcId, List<ToscaTechnologicalConstraint> possibleTCForTosca, OptimizationEntity toscaOptimization) {
        if (!possibleTCForTosca.stream().map(ToscaTechnologicalConstraint::getId).anyMatch(i -> i.equals(tcId))) {
            Assert.fail("Target TC is not on list of possible TC for Tosca");
        }

        String tcUrl = TOSCA_TECH_CONSTRAINT_URL.replace("{optimizationId}", String.valueOf(toscaOptimization.getId()));
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, tcUrl);
        List<ToscaOptimizationTechnologicalConstraint> currentConstraints = jsonToToscaTechnologicalConstraintList(jsonNode);
        ToscaOptimizationTechnologicalConstraint updated = currentConstraints.get(0);
        updated.setTechnologicalConstraintId(tcId);
        try {
            TestUtils.UPDATEobject(mockMvc, tcUrl + "/" + updated.getId(), updated, ToscaOptimizationTechnologicalConstraint.class, TestUtils::toJsonNoProxy);
        } catch (Exception e) {
            Assert.fail("Unable to read value from json");
        }

        jsonNode = TestUtils.GETobjects(mockMvc, tcUrl);
        currentConstraints = jsonToToscaTechnologicalConstraintList(jsonNode);
        Assert.assertTrue(currentConstraints.get(0).getTechnologicalConstraintId().equals(tcId));

        return currentConstraints.get(0);
    }

    private List<ToscaTechnologicalConstraint> getPossibleTCForTosca(String suggestUrl) throws IOException {
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, suggestUrl + "/technologicalConstraints");
        return jacksonMapper.readValue(jsonNode.toString(), new TypeReference<List<ToscaTechnologicalConstraint>>() {
        });
    }


    private Stream<TechnologicalConstraintTreeNode> getParametersWithFields(String suggestUrl, Integer tcId) {
        JsonNode jsonNode = TestUtils.GETobjects(mockMvc, suggestUrl + "/technologicalConstraints/" + tcId + "/tree");
        return jsonToToscaTechnologicalConstraintTreeList(jsonNode).stream()
                .flatMap(this::flattenChildren);
    }


    private Stream<TechnologicalConstraintTreeNode> flattenChildren(TechnologicalConstraintTreeNode node) {
        List<TechnologicalConstraintTreeNode> children = node.getChildren();
        if (children == null || children.isEmpty()) {
            return Stream.of(node);
        } else {
            return Stream.concat(Stream.of(node), children.stream().flatMap(this::flattenChildren));
        }
    }


    private List<TechnologicalConstraintTreeNode> jsonToToscaTechnologicalConstraintTreeList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<TechnologicalConstraintTreeNode>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }


    private List<ToscaOptimizationTechnologicalConstraint> jsonToToscaTechnologicalConstraintList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ToscaOptimizationTechnologicalConstraint>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }


    private List<TechnologicalConstraintTreeNodeValue> jsonToToscaTechnologicalConstraintTreeValue(JsonNode json) {
        try {
            Map<Integer, TechnologicalConstraintTreeNodeValue> selectedFields = jacksonMapper
                    .readValue(json.toString(), new TypeReference<Map<Integer, TechnologicalConstraintTreeNodeValue>>() {
                    });
            return selectedFields.values().stream().collect(toList());
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }


    private List<ToscaOptimizationBaseSetting> jsonToToscaOptimizationBaseSettingList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ToscaOptimizationBaseSetting>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }
}
