package ru.crp.lvms.optimizations.base.managers;

import org.modelmapper.TypeToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.MaterialManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.TechnologyManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.dao.TreeDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tree.OptimizationsTreeNode;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.dao.OptimizationDao;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.dao.OptimizationReportDao;
import ru.crp.lvms.optimizations.base.entities.*;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.enums.OptimizationTypes;
import ru.crp.lvms.optimizations.mf.managers.MFDefaultParamsCreator;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.optimus.managers.OptimusDefaultParamsCreator;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructOptimizationAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructOptimizationAdditionalParameter;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructAdditionalParametersManager;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructOptimizationSettingsManager;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationBaseSetting;
import ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaJsonParametersParents;
import ru.crp.lvms.optimizations.tosca.enums.ToscaSettingsType;
import ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.ValidationHelper;
import ru.crp.lvms.web.dto.OptimizationsGraphEntry;

import javax.annotation.Nullable;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.Collections.emptyMap;
import static java.util.stream.Collectors.*;
import static ru.crp.lvms.model.mappers.PgArrayPropertyEditor.LOGGER;
import static ru.crp.lvms.tasks.fileGenerators.ToscaInputFileGenerator.ABAQUS_CAE_FILE_EXT;
import static ru.crp.lvms.tasks.fileGenerators.ToscaInputFileGenerator.ANSYS_CAE_FILE_EXT;
import static ru.crp.lvms.utils.MessageUtils.OPTIMIZATION_USED_IN_FOLDERS;

@Service
public class OptimizationManager extends BaseManagerSuggest<OptimizationEntity> implements ImageHandler<OptimizationEntity> {
    public static final String TOSCA_ABAQUS_STARTUP_ARGUMENTS = "--solver abaqus";
    public static final String TOSCA_ANSYS_STARTUP_ARGUMENTS = "--solver ansys";

    @Autowired
    private TreeDao treeDao;

    @Autowired
    private OptimizationDao optimizationDao;

    @Autowired
    private OptimizationReportDao optimizationReportDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ConstructionElementManager constructionElementManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private MFDefaultParamsCreator mfDefaultParamsCreator;

    @Autowired
    private OptimusDefaultParamsCreator optimusDefaultParamsCreator;

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Autowired
    private ToscaDefaultParamsCreator toskaAdditionalParamsCreator;

    @Autowired
    private MFOptimizationSettingsManager MFOptimizationSettingsManager;

    @Autowired
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    @Autowired
    private OptistructOptimizationSettingsManager optistructOptimizationSettingsManager;

    @Autowired
    private OptistructOptimizationAdditionalParametersDao optistructOptimizationAdditionalParametersDao;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    @Autowired
    private OptimizationDependenciesManager optimizationDependenciesManager;

    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private TechnologyManager technologyManager;


    public void deleteOptimizationFolder(int optimizationFolderId, UserEntity currentUser) {
        OptimizationsTreeNode folderToDelete = checkOptimizationFolderExists(optimizationFolderId);
        List<OptimizationEntity> children = optimizationDao.findByPid(folderToDelete.getId());
        if (!children.isEmpty()) {
            throw new BadRequestException("Невозможно удалить директорию, содержащую оптимизации");
        }
        treeDao.deleteOptimizationFolder(optimizationFolderId);
    }


    @Override
    public void checkDelete(OptimizationEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), OPTIMIZATION_USED_IN_FOLDERS,
                treeDao::getOptimizationFoldersByOptimizationId, currentUser);
    }

    @Override
    public void delete(OptimizationEntity optimizationToDelete, @Nullable UserEntity currentUser) {
        checkDelete(optimizationToDelete, currentUser);
        resultsManager.deleteResults(Mapper.map(optimizationToDelete, ResultsParent.class), true, CreatedBy.Auto, CreatedBy.Manual);
        aliveDao().delete(optimizationToDelete.getId());
    }

    private OptimizationsTreeNode checkOptimizationFolderExists(int folderId) {
        OptimizationsTreeNode optimizationFolder = treeDao.getOptimizationFolder(folderId);
        if (optimizationFolder == null) {
            throw new NotFoundException(folderId, OptimizationsTreeNode.class);
        }
        return optimizationFolder;
    }


    public OptimizationEntity createOptimization(OptimizationEntity newEntity, UserEntity user) {
        OptimizationsTreeNode parent = checkOptimizationFolderExists(newEntity.getPid());
        if (parent.getItemType() != ItemType.OptimizationFolder) {
            throw new BadRequestException("Оптимизация должна принадлежать папке");
        }
        checkPathNames(parent,newEntity.getName());
        newEntity.setOptimizationTypeId(parent.getOptimizationTypeId());
        newEntity.setOwnerId(user.getId());
        newEntity.setSolvingStatusId(SolvingStatus.Created.getId());
        OptimizationEntity newOptimization = super.createAndRead(newEntity, user);
        // for engineering calculations only CAE solvers are used
        if (newOptimization.getSolverCaoId() != null) {
            switch (OptimizationCao.valueOf(newOptimization.getSolverCaoId())) {
                case Tosca:
                    toscaOptimizationSettingsManager.addDefaultSettings(newOptimization);
                    toskaAdditionalParamsCreator.addAdditionalOptimizationParameters(newOptimization);
                    break;
                case ModeFrontier:
                    mfDefaultParamsCreator.addDefaultSettings(newOptimization);
                    break;
                case Optistruct:
                    optistructOptimizationSettingsManager.addDefaultSettings(newOptimization);
                    break;
                case Optimus:
                    optimusDefaultParamsCreator.addDefaultSettings(newOptimization);
                    break;
            }
        }
        return newOptimization;
    }

    private void checkPathNames(TreeNode parent, String name) {
        String pathNames = parent.getPathNames() + "/" + name;
        List<OptimizationEntity> existedOptimizations = optimizationDao.findByPathNames(pathNames);
        if (!existedOptimizations.isEmpty()) {
            throw new BadRequestException("В папке уже есть расчет с таким именем");
        }
    }
    public void extendOptimization(OptimizationEntity optimization, UserEntity currentUser) {
        optimization.setFiles(getOptimizationFiles(optimization, currentUser));
        optimization.setResults(resultsManager.getResultsMapByParent(optimization.getId(), TaskParentType.optimization, currentUser));
        optimization.setReports(getOptimizationReports(optimization, currentUser));
    }


    public Map<CreatedBy, List<OptimizationReportEntity>> getOptimizationReports(OptimizationEntity optimization, UserEntity currentUser) {
        Map<CreatedBy, List<OptimizationReportEntity>> reports = optimizationReportDao.findByOptimizationId(optimization.getId()).stream()
                .collect(groupingBy(OptimizationReportEntity::getCreatedBy));

        Map<CreatedBy, List<OptimizationReportEntity>> emptyReports = Stream.of(CreatedBy.values())
                .filter(type -> !reports.containsKey(type))
                .collect(toMap(Function.identity(), (unused) -> emptyList()));
        reports.putAll(emptyReports);
        return reports;
    }


    public PageImpl<OptimizationReportEntity> getOptimizationReportsPage(OptimizationEntity optimization,
                                                                         Map<String, Object> filters, Pageable page,
                                                                         UserEntity currentUser) {
        return optimizationReportDao.findPageByOptimizationId(optimization.getId(), filters, page);
    }

    public List<FileEntity> getOptimizationFiles(OptimizationEntity optimization, FileType fileType, UserEntity currentUser) {
        return fileManager.findOptimizationFiles(optimization.getId(), fileType.getId(), currentUser);
    }

    public List<FileEntity> getOptimizationFiles(OptimizationEntity optimization, UserEntity currentUser) {
        return fileManager.findOptimizationFiles(optimization.getId(), null, currentUser);
    }

    public List<OptimizationEntity> getOptimizationsByConstructionElementId(int constructionElementId, UserEntity currentUser) {
        return optimizationDao.getOptimizationsByConstructionElementId(constructionElementId);
    }

    public PageImpl<FileEntity> getOptimizationFilesPage(OptimizationEntity optimization,
                                                         Map<String, Object> filters, Pageable page,
                                                         UserEntity currentUser) {
        return fileManager.findOptimizationFilesPage(optimization.getId(), filters, page, currentUser);
    }

    public OptimizationReportEntity getReport(Integer reportId) {
        return optimizationReportDao.findOne(reportId);
    }

    public void updateReport(OptimizationReportEntity updatedEntity) {
        optimizationReportDao.update(updatedEntity);
    }

    public void deleteReport(Integer reportId) {
        optimizationReportDao.delete(reportId);
    }


    public List<OptimizationsGraphEntry> getOptimizationsGraph(int optimizationId, UserEntity currentUser) {
        OptimizationEntity initialOptimization = findOne(optimizationId, currentUser);
        if (initialOptimization == null) {
            return null;
        }
        List<OptimizationsTreeNode> optimizationFolders = treeDao.getOptimizationFoldersAscendingBranch(initialOptimization.getPid());
        ConstructionElementEntity initialCE = constructionElementManager.findOne(optimizationFolders.get(0).getPid(), currentUser);
        Objects.requireNonNull(initialCE, "Data integrity violation");
        List<OptimizationEntity> optimizations = optimizationFolders.stream()
                .map(OptimizationsTreeNode::getBaseOptimizationId)
                .filter(Objects::nonNull)
                .map(id -> findOne(id, currentUser))
                .collect(toList());
        List<OptimizationsGraphEntry> graph = Mapper.map(optimizations, new TypeToken<List<OptimizationsGraphEntry>>() {
        });
        graph.add(0, Mapper.map(initialCE, OptimizationsGraphEntry.class));
        graph.add(Mapper.map(initialOptimization, OptimizationsGraphEntry.class));
        return graph;
    }

    public void replaceFiles(OptimizationEntity optimization, List<Integer> fileIds, UserEntity userEntity) {
        fileIds.forEach(fileId -> fileManager.checkEntityExists(fileId, userEntity));
        optimizationDao.replaceFiles(optimization.getId(), fileIds);
    }

    public List<OptimizationEntity> getOptimizationsThatUseFile(Integer fileId, UserEntity currentUser) {
        return optimizationDao.findByFileId(fileId);
    }


    public FileEntity addLinkToCEFile(OptimizationEntity optimization, Integer fileId, FileType fileType, UserEntity currentUser) {
        int parentCEId = optimization.getParentCEId();
        FileEntity fileToAdd = constructionElementManager.findConstructionElementFiles(parentCEId, fileType, currentUser).stream()
                .filter(file -> file.getId() == fileId)
                .findAny().orElseThrow(() -> new BadRequestException(String.format("Файл с номером %d не принадлежит родительскому элементу конструкции", fileId)));
        if (optimizationDao.isFileUsedByOptimization(optimization.getId(), fileToAdd.getId())) {
            throw new BadRequestException("Файл уже добавлен к оптимизации");
        }
        optimizationDao.appendFile(optimization.getId(), fileToAdd.getId());
        return fileToAdd;
    }


    public FileEntity addModel(OptimizationEntity optimization, MultipartFile newModel, FileType fileType, UserEntity currentUser) {
        FileEntity newModelEntity;
        if (asList(FileType.CADFile, FileType.CAEFile, FileType.CAOFile, FileType.DesignSolutionFile).contains(fileType)) {
            int parentCEId = optimization.getParentCEId();
            newModelEntity = constructionElementManager.createNewFile(parentCEId, newModel, fileType, currentUser);
        } else {
            String modelName = null;
            try {
                modelName = URLDecoder.decode(newModel.getOriginalFilename(), StandardCharsets.UTF_8.name());
            } catch (UnsupportedEncodingException e) {
                throw new RuntimeException("Unsupported encoding ", e);
            }
            if (optimizationDao.isFileUsedByOptimization(optimization.getId(), modelName)) {
                throw new BadRequestException(String.format("Файл '%s' уже используется в оптимизации '%s'", modelName, optimization.getName()));
            }
            newModelEntity = fileManager.createOptimizationFile(newModel, optimization.getId(), fileType, currentUser);
        }
        optimizationDao.appendFile(optimization.getId(), newModelEntity.getId());
        return newModelEntity;
    }


    public void deleteModel(OptimizationEntity optimization, FileEntity fileToDelete, UserEntity currentUser) {
        optimizationDao.deleteFile(optimization.getId(), fileToDelete.getId());

        // if file is used only for this optimization
        if (fileToDelete.getPid() == null && !optimizationDao.isFileUsedByAnyOptimization(fileToDelete.getId())) {
            fileManager.deleteFile(fileToDelete.getId(), currentUser);
        }
    }


    public OptimizationFlow getOptimizationFlow(int optimizationId, UserEntity currentUser) throws IOException {
        OptimizationEntity optimization = findOne(optimizationId, currentUser);
        OptimizationFlow flow = aliveDao().getOptimizationFlow(optimizationId);

        List<OptimizationFlow.ResultsInfo> existedResults = aliveDao().getOptimizationResultsInfo(optimizationId);
        Set<ResultType> existedResultTypes = existedResults.stream()
                .map(OptimizationFlow.ResultsInfo::getResultType)
                .collect(toSet());
        List<OptimizationFlow.ResultsInfo> emptyResults = Stream.of(ResultType.values())
                .filter(resultType -> !existedResultTypes.contains(resultType) && resultType != ResultType.logFiles)
                .map(resultType -> new OptimizationFlow.ResultsInfo(resultType, 0, 0))
                .collect(toList());
        existedResults.addAll(emptyResults);
        int totalAuto = existedResults.stream()
                .mapToInt(OptimizationFlow.ResultsInfo::getTotalAuto)
                .sum();
        int totalManual = existedResults.stream()
                .mapToInt(OptimizationFlow.ResultsInfo::getTotalManual)
                .sum();
        flow.setResults(new OptimizationFlow.TotalResultsInfo(existedResults, totalAuto, totalManual));

        Path pathToResult = resultsFilesManager.getResultsDir(Mapper.map(optimization, ResultsParent.class));
        if (!Files.isDirectory(pathToResult)) {
            flow.setFiles(new OptimizationFlow.BaseInfo(0));
        } else {
            flow.setFiles(new OptimizationFlow.BaseInfo((int) resultsFilesManager.getFilesCount(pathToResult)));
        }
        return flow;
    }

    @Override
    public FileManager fileManager() {
        return fileManager;
    }

    @Override
    public OptimizationDao aliveDao() {
        return optimizationDao;
    }

    public String getAdditionalStartupArgumentsForTosca(OptimizationEntity optimization) {
        AtomicBoolean useAbaqus = new AtomicBoolean(false);
        AtomicBoolean useAnsys = new AtomicBoolean(false);
        getOptimizationFiles(optimization, null).stream()
                .map(FileEntity::getName)
                .forEach(fileName -> {
                    if (fileName.endsWith(ANSYS_CAE_FILE_EXT)) {
                        useAnsys.set(true);
                    } else if (fileName.endsWith(ABAQUS_CAE_FILE_EXT)) {
                        useAbaqus.set(true);
                    }
                });
        if (useAbaqus.get() && useAnsys.get()) {
            throw new IllegalStateException("Solvers conflict for optimization " + optimization.getId());
        }
        // ANSYS is used by default
        return !useAbaqus.get() ? TOSCA_ANSYS_STARTUP_ARGUMENTS : TOSCA_ABAQUS_STARTUP_ARGUMENTS;
    }

    public void copyResultsFemToOptimizationFiles(TaskEntity task) {
        OptimizationEntity optimization = findOne(task.getParentId(), null);
        ResultsParent resultsParent = Mapper.map(optimization, ResultsParent.class);
        List<ResultEntity> fems = resultsManager.getResultByParent(resultsParent, ResultType.optistructFEMFile);
        fems.stream().map(resultsFilesManager::getPathToResult)
                .forEach(fem -> {
                    FileEntity created = fileManager.createOptimizationFile(fem, optimization.getId(), FileType.CAEFile, optimization.getOwnerId());
                    optimizationDao.appendFile(optimization.getId(), created.getId());
                });
    }

    public List<FileEntity> getOptimizationFemFiles(OptimizationEntity optimization) {
        return getOptimizationFiles(optimization, FileType.CAEFile, null)
                .stream().filter(fileEntity -> fileEntity.getName().toLowerCase().endsWith("fem"))
                .collect(Collectors.toList());
    }

    // if <fileTypes> is null then files with all types will be returned
    public List<FileData> getInputFilesForOptimization(OptimizationEntity optimization, @Nullable List<FileType> fileTypes) throws IOException {
        Predicate<FileEntity> predicate = (fileEntity) -> fileTypes == null || fileTypes.contains(FileType.valueOf(fileEntity.getFileTypeId()));

        return getOptimizationFiles(optimization, null).stream()
                .filter(predicate)
                .map(fileEntity -> {
                    Path pathToFile = fileManager.resolvePathToFile(fileEntity, null);
                    if (pathToFile == null || !Files.exists(pathToFile)) {
                        LOGGER.error("File {} is present in db but not in file system", fileEntity.getId());
                        return null;
                    }
                    return FileData.fromFile(fileEntity.getName(), DataType.SUBMODEL, pathToFile.toFile());
                })
                .filter(Objects::nonNull)
                .collect(toList());
    }

    public Map<String, String> getCustomParametersForTask(TaskEntity task) {
        OptimizationEntity optimization = findOne(task.getParentId(), null);
        if (!Objects.equals(optimization.getSolverCaoId(), OptimizationCao.ModeFrontier.getId())) {
            return emptyMap();
        }
        return optimizationJsonSettingsDao.findRunSettings(optimization.getId()).stream()
                .filter(s -> s.getXmlName() == null)
                .filter(s -> s.getValue() != null && !s.getValue().toString().isEmpty())
                .collect(toMap(OptimizationJsonSetting::getName, s -> s.getValue().toString()));
    }

    public OptimizationEntity copyOptimization (String newName,OptimizationEntity referenceOptimization,String isChild, UserEntity user) throws CloneNotSupportedException{
        OptimizationEntity newEntity = referenceOptimization.clone();
        newEntity.setName(newName);
        if (!isChild.equals("false")) {
            newEntity.setReferenceId(referenceOptimization.getId());
        }
        newEntity.setSolvingStatusId(1);
        OptimizationEntity newOptimization = super.createAndRead(newEntity,user);
        cloneAllDependencies(referenceOptimization,newOptimization,user);
        return newOptimization;
    }

    public void cloneAllDependencies (OptimizationEntity referenceOptimization, OptimizationEntity newOptimization,UserEntity user) {
        if (newOptimization.getSolverCaoId() != null) {
            switch (OptimizationCao.valueOf(newOptimization.getSolverCaoId())) {
                case Tosca:
                    toscaOptimizationSettingsManager.copySettings(referenceOptimization.getId(),newOptimization.getId());
                    break;
                case ModeFrontier:
                    MFOptimizationSettingsManager.copySettings(referenceOptimization.getId(),newOptimization.getId());
                    break;
                case Optistruct:
                    optistructOptimizationSettingsManager.copySettings(referenceOptimization.getId(),newOptimization.getId());
                    break;
                case Optimus:
                    optimusOptimizationSettingsManager.copySettings(referenceOptimization.getId(),newOptimization.getId());
                    break;
            }
            cloneJsonSettings(referenceOptimization,newOptimization);
            cloneFiles(referenceOptimization,newOptimization,user);
            cloneDependencies(referenceOptimization,newOptimization,user);
        }
        if (newOptimization.getOptimizationTypeId()==OptimizationType.ENGINEERING_CALCULATION_TYPE_ID) {
            cloneFiles(referenceOptimization,newOptimization,user);
            cloneDependencies(referenceOptimization,newOptimization,user);
        }
    }

    public void cloneJsonSettings(OptimizationEntity referenceOptimization, OptimizationEntity newOptimization) {
        List<OptimizationJsonSetting> referenceOptimizationJsonSettings = optimizationJsonSettingsDao.findAllOptimizationSettings(referenceOptimization.getId());
        for (OptimizationJsonSetting jsonSetting : referenceOptimizationJsonSettings) {
            jsonSetting.setOptimizationId(newOptimization.getId());
            optimizationJsonSettingsDao.createSettings(jsonSetting);
        }
        OptimizationTypes optType = OptimizationTypes.valueOf(newOptimization.getOptimizationTypeId());
        if (newOptimization.getSolverCaoId() != null) {
            switch (OptimizationCao.valueOf(newOptimization.getSolverCaoId())) {
                case Tosca:
                    Stream.of(ToscaJsonParametersParents.values())
                            .filter(optMethod -> optMethod.isMethodAvailableForOptimizationType(optType))
                            .findAny()
                            .ifPresent(tom -> optimizationDao.selectOptimizationMethod(newOptimization.getId(), tom.getId()));
                    break;
                case ModeFrontier:
                    Integer selectedMFAlgorithm = MFOptimizationSettingsManager.findSelectedAlgorithmId(referenceOptimization.getId());
                    MFOptimizationSettingsManager.selectAlgorithm(newOptimization.getId(),selectedMFAlgorithm);
                    Integer selectedMFDoe = MFOptimizationSettingsManager.findSelectedDoeId(referenceOptimization.getId());
                    MFOptimizationSettingsManager.selectDoe(newOptimization.getId(),selectedMFDoe);
                    break;
                case Optistruct:

                    break;
                case Optimus:
                    Integer selectedOptimusAlgorithm = optimusOptimizationSettingsManager.findSelectedAlgorithmId(referenceOptimization.getId());
                    optimusOptimizationSettingsManager.selectAlgorithm(newOptimization.getId(),selectedOptimusAlgorithm);
                    Integer selectedOptimusDoe = optimusOptimizationSettingsManager.findSelectedDoeId(referenceOptimization.getId());
                    optimusOptimizationSettingsManager.selectDoe(newOptimization.getId(),selectedOptimusDoe);
                    break;
            }
        }

    }

    public void cloneFiles (OptimizationEntity referenceOptimization, OptimizationEntity newOptimization,UserEntity user){
        List<FileEntity> files = fileManager.findOptimizationFiles(referenceOptimization.getId(),null,null);
        files.forEach(f->{
            if (f.getFileType().equals(FileType.CADFile.toString())||f.getFileType().equals(FileType.CAEFile.toString())) {
                optimizationDao.appendFile(newOptimization.getId(), f.getId());
            }else{
                Path newFilePath = Paths.get(f.getPathToFile());
                FileEntity newLC = fileManager.createOptimizationFile(resultsFilesManager.getStorageDir().resolve(newFilePath),newOptimization.getId(),FileType.valueOf(f.getFileType()),user.getId());
                newLC.setName(f.getName());
                fileManager.updateFileEntity(newLC,user);
                optimizationDao.appendFile(newOptimization.getId(), newLC.getId());
            }
        });
    }

    public void cloneDependencies (OptimizationEntity referenceOptimization, OptimizationEntity newOptimization,UserEntity user){
        List<OptimizationDependency> optimizationMaterials = optimizationDependenciesManager.getMaterials(referenceOptimization,user);
        optimizationMaterials.forEach(om->{
            optimizationDependenciesManager.addMaterial(newOptimization,materialManager.findOne(om.id,user),user);
            });
        List<OptimizationDependency> optimizationTechnologies = optimizationDependenciesManager.getTechnologies(referenceOptimization,user);
        optimizationTechnologies.forEach(ot->{
            optimizationDependenciesManager.addTechnology(newOptimization,technologyManager.findOne(ot.id,user),user);
        });
        }
}
