package ru.crp.lvms.tasks.fileGenerators;

import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.TechnologySimulationsManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;

import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;
import java.util.Optional;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

abstract public class BaseInputFileGenerator {

    private static final Logger LOGGER = LoggerFactory.getLogger(BaseInputFileGenerator.class);
    @Autowired
    protected FileManager fileManager;

    @Autowired
    protected ECManager ecManager;

    @Autowired
    protected TechnologySimulationsManager technologySimulationsManager;

    @Autowired
    protected SimulationManager simulationManager;

    @Autowired
    protected ResultsManager resultsManager;

    @Autowired
    protected ResultsFilesManager resultsFilesManager;

    @Autowired
    protected MFOptimizationSettingsManager MFOptimizationSettingsManager;

    @Autowired
    protected OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Autowired
    protected OptimizationManager optimizationManager;


    protected void addFilesFromDir(Collection<FileData> fileData, String pathToDirStr) throws IOException {
        addFilesFromDir(fileData, pathToDirStr, DataType.SESSION_FILE);
    }

    protected void addFilesFromDir(Collection<FileData> fileData, String pathToDirStr, DataType dataType) throws IOException {
        Path pathToDir = Paths.get(pathToDirStr);
        if (!Files.exists(pathToDir)) {
            LOGGER.warn("Trying to add files from nonexistent dir {}", pathToDirStr);
            return;
        }
        addFilesFromDir(fileData, pathToDir, DataType.SESSION_FILE);
    }

    protected void addFilesFromDir(Collection<FileData> fileData, Path pathToDir, DataType dataType) throws IOException {
        try (Stream<Path> filesStream = Files.walk(pathToDir)) {
            fileData.addAll(filesStream
                    .filter(path -> Files.isRegularFile(path))
                    .map(file -> FileData.fromFile(file.getFileName().toString(), dataType,
                            file, pathToDir.relativize(file.getParent()).toString()))
                    .collect(toList()));
        } catch (IOException e) {
            throw new InternalException("Unable to add directory as file data", e);
        }
    }


    public Pair<FileEntity, Path> getSingleFile(OptimizationEntity optimization, FileType fileType, String errorMessage) {
        Optional<FileEntity> file = optimizationManager.getOptimizationFiles(optimization, fileType, null).stream()
                .findFirst();
        if (!file.isPresent()) {
            throw new TaskStartingException(errorMessage);
        }
        Path pathToFile = fileManager.resolvePathToFile(file.get(), null);
        if (pathToFile == null || !Files.exists(pathToFile)) {
            throw new InternalException(String.format("File %d is present in db but not in file system", file.get().getId()));
        }
        return MutablePair.of(file.get(), pathToFile);
    }

    @Nullable
    protected String commasToDots(@Nullable Object value, @Nullable String defaultValue) {
        if (value == null) {
            return defaultValue;
        }
        return value.toString().replace(",", ".");
    }
}
